/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-canvas implementation.

    DTV-canvas is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-canvas is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-canvas.

    DTV-canvas es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-canvas se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "system.h"
#include "window.h"
#include "webviewer.h"
#include "videooverlay.h"
#include "../cairo/canvas.h"
#include "../../event.h"
#include "generated/config.h"
#include <util/task/dispatcherimpl.h>
#include <util/id/pool.h>
#include <util/log.h>
#include <util/mcr.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

namespace canvas {
namespace gtk {

namespace impl {

static System *sys = NULL;

System *getSystem() {
	assert(impl::sys);
	return impl::sys;
}

static gboolean ev_rcv( GIOChannel * /*chan*/, GIOCondition /*condition*/, System::IOEvent *io ) {
	assert(io);
	getSystem()->onDispatchIO( io );
	return TRUE;
}

static gboolean timeoutCallback( System::Timer *timer ) {
	assert(timer);
	getSystem()->onDispatchTimer( timer );
	return FALSE;
}

static gboolean onTaskCallback( System *sys ) {
	assert(sys);
	sys->onTaskEvent();
	return FALSE;
}

}	//	namespace impl


System::System()
{
	_idleFunc = -1;
	impl::sys = this;
	_pool = new util::id::Pool("gtk");
	_mainLoopRunning = false;

	// Create task dispatcher
	_tasks = new util::task::DispatcherImpl();
	_tasks->onPost( boost::bind(&System::onPost,this) );
}

System::~System()
{
	impl::sys = NULL;
	assert(!_mainLoopRunning);
	delete _pool;
	delete _tasks;
}

//	Initialization
bool System::init() {
	_tasks->registerTarget( this );
	
	//	Initialize glib
	gtk_init(NULL, NULL);
	return true;
}

void System::fin() {
	_tasks->runAll();
	_tasks->unregisterTarget( this );	
}

//	Run/exit system loop
void System::runImpl() {
	LINFO("gtk::System", "Main loop begin");
	
	_mainLoopRunning=true;

	//	Run loop
	gtk_main();

	LINFO("gtk::System", "Main loop end");
}

void System::exit() {
	if (_mainLoopRunning) {
		_mainLoopRunning=false;
		//	Wakeup thread
		_tasks->post( this, boost::bind(&System::doExit,this) );
	}
}

void System::doExit() {
	gtk_main_quit();
}

//	IO events
util::id::Ident System::addIO( util::DESCRIPTOR fd, const EventCallback &callback ) {
#ifdef _WIN32
	UNUSED(fd);
	UNUSED(callback);
	return util::id::Ident();
#else
	IOEvent *ev = new IOEvent(_pool->alloc(),callback);
	GIOChannel *chan = g_io_channel_unix_new(fd);
	int watchID=g_io_add_watch( chan, (GIOCondition) (G_IO_IN | G_IO_PRI), (GIOFunc)impl::ev_rcv, (gpointer)ev );
	g_io_channel_unref( chan );
	return startEvent( ev, watchID != 0, watchID, _ioList );
#endif
}
	
void System::stopIO( util::id::Ident &id ) {
	int watchID;
	if (stopEvent( id, _ioList, watchID )) {
		g_source_remove( watchID );
	}
}

void System::onDispatchIO( IOEvent *io ) {
	dispatchEvent( _ioList, io );
}

//	Socket events
util::id::Ident System::addSocket( util::SOCKET sock, const EventCallback &callback ) {
#ifdef _WIN32
	IOEvent *ev = new IOEvent(_pool->alloc(),callback);
	GIOChannel *chan = g_io_channel_win32_new_socket (sock);
	int watchID=g_io_add_watch( chan, (GIOCondition) (G_IO_IN | G_IO_PRI), (GIOFunc)impl::ev_rcv, (gpointer)ev );
	g_io_channel_unref( chan );
	return startEvent( ev, watchID != 0, watchID, _ioList );
#else	
	return addIO( sock, callback );
#endif
}

void System::stopSocket( util::id::Ident &id ) {
	stopIO( id );
}

//	Timers
util::id::Ident System::registerTimer( util::DWORD ms, const EventCallback &callback ) {
	Timer *ev = new Timer(_pool->alloc(), callback);
	int id=g_timeout_add( ms, (GSourceFunc)impl::timeoutCallback, (gpointer)ev );
	return startEvent( ev, id != 0, id, _timerList );
}

void System::unregisterTimer( util::id::Ident &id ) {
	int timerID;
	if (stopEvent( id, _timerList, timerID )) {
		g_source_remove( timerID );
	}
}

void System::onDispatchTimer( Timer *timer ) {
	//	Two pass, becouse in dispatch callback can unregister timer!
	util::id::Ident id=dispatchEvent( _timerList, timer );
	unregisterTimer( id );
}

//	Instance creation
canvas::WebViewer *System::createWebViewer( Surface *surface ) {
#if SYS_HTML_USE_GTK
	return new WebViewer( (canvas::System *)this, surface );
#else
	return canvas::System::createWebViewer( surface );
#endif
}

canvas::Canvas *System::createCanvas() const {
	return new cairo::Canvas();
}

canvas::Window *System::createWindow() const {
	return new Window();
}
	 
void System::onWindowDestroyed() {
	dispatchShutdown();
}

//	Dispatcher
util::task::Dispatcher *System::dispatcher() const {
	return _tasks;
}
 
void System::onTaskEvent() {
	_tasks->runAll();
}
 
void System::onPost() {
 	g_timeout_add( 1, (GSourceFunc)impl::onTaskCallback, (gpointer)this );
}

}
}
