/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-canvas implementation.

    DTV-canvas is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-canvas is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-canvas.

    DTV-canvas es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-canvas se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "mediaplayer.h"
#include "player.h"
#include "../../point.h"
#include "../../window.h"
#include "../../system.h"
#include <util/task/dispatcher.h>
#include <util/url.h>
#include <util/log.h>
#include <boost/bind.hpp>
#include <boost/lexical_cast.hpp>
#include <vlc/vlc.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

namespace canvas {
namespace vlc {

namespace impl {

static unsigned vlc_format( void **opaque, char *chroma, unsigned *width, unsigned *height, unsigned *pitches, unsigned *lines ) {
	MediaPlayer *mp = (MediaPlayer *)(*opaque);
	assert(mp);
	return mp->getFormat( chroma, width, height, pitches, lines );
}

static void vlc_cleanup( void *opaque ) {
	MediaPlayer *mp = (MediaPlayer *)opaque;
	assert(mp);
	mp->cleanup();
}

static void vlc_display( void *opaque, void *id ) {
	MediaPlayer *mp = (MediaPlayer *)opaque;
	assert(mp);
	mp->renderFrame( id );
}

static void *vlc_lock( void *opaque, void **p_pixels ) {
	MediaPlayer *mp = (MediaPlayer *)opaque;
	assert(mp);
	return mp->allocFrame( p_pixels );
}

static void vlc_unlock( void *opaque, void *id, void *const * /*p_pixels*/ ) {
	MediaPlayer *mp = (MediaPlayer *)opaque;
	assert(mp);
	mp->freeFrame( id );
}

static void vlc_event(const libvlc_event_t *p_event, void *p_user_data) {
	if (p_event->type == libvlc_MediaPlayerEndReached) {
		assert(p_user_data);
		MediaPlayer *mp = (MediaPlayer *)p_user_data;
		mp->onStopped();
	}
}

}	//	namespace impl


MediaPlayer::MediaPlayer( System *sys, Player *player )
	: canvas::MediaPlayerImpl( sys )
{
	_player = player;
	_mp = NULL;
	_needsMute = false;
	_volume = -1;
}

MediaPlayer::~MediaPlayer()
{
	if (_mp) {
		libvlc_media_player_release( _mp );
	}
}

bool MediaPlayer::getParam( util::Url *url, const std::string &param1, const std::string &param2, std::string &result ) {
	bool ret=false;
	int pid;
	int type;
	if (url->getParam( param1, pid ) &&
		url->getParam( param2, type ))
	{
		result  = boost::lexical_cast<std::string>(pid);
		result += ":";
		result += boost::lexical_cast<std::string>(type);
		ret=true;
	}
	return ret;
}

//	Initialization
bool MediaPlayer::initialize( const std::string &data ) {
	util::Url url(data);	
	LDEBUG( "vlc", "Initialize: url=%s", data.c_str() );
	
	//	Create media from url
	libvlc_media_t *m;
	if (url.type() == "srvdtv") {
		m = libvlc_media_new_path( instance(), "/dev/dvb/adapter0/dvr0" );
		if (m) {
			std::string pmt = "ts-extra-pmt=";
			int streams = 0;
			std::string audio, video;		

			//	Audio
			if (getParam( &url, "audioPID", "audioType", audio )) {
				streams++;
			}

			//	Video
			if (getParam( &url, "videoPID", "videoType", video )) {
				streams++;
			}

			if (!streams) {
				LWARN( "vlc", "No stream to play" );
				return false;
			}
	
			pmt += boost::lexical_cast<std::string>(streams);
			pmt += "=";

			if (!video.empty()) {
				pmt += video;
			}
	
			if (!audio.empty()) {
				pmt += ",";
				pmt += audio;
			}

			{
				int pcrPID;
				if (url.getParam( "pcrPID", pcrPID )) {
					pmt += ",";			
					pmt += boost::lexical_cast<std::string>(pcrPID);
					pmt += ":pcr";
				}
			}

			LDEBUG( "vlc", "PMT parameters: pmt=%s", pmt.c_str() );

			//	Add vlc options
			libvlc_media_add_option( m, "demux=ts" );
			libvlc_media_add_option( m, pmt.c_str() );
		}
	}
	else {
		m=libvlc_media_new_path( instance(), data.c_str() );
	}

	if (!m) {
		LERROR( "vlc", "Cannot initialize new media from url: url=%s", data.c_str() );
		return false;
	}

	//	Add common parameters
	libvlc_media_add_option( m, "verbose=1000" );
	libvlc_media_add_option( m, "no-osd" );
	libvlc_media_add_option( m, "ffmpeg-hw" );

	//	Create a media player playing environement
	_mp = libvlc_media_player_new_from_media(m);

	//	No need to keep the media now
	libvlc_media_release (m);

	if (!_mp) {
		LERROR( "vlc", "Cannot create media player from url: url=%s", data.c_str() );
		return false;
	}

	if (supportVideoOverlay()) {
		//	Set WindowID
		VideoDescription desc;
		if (getVideoDescription( desc )) {
#ifndef _WIN32
			LDEBUG( "vlc", "Using WinID=0x%x", desc.winID );
			libvlc_media_player_set_xwindow( _mp, desc.winID );
#else
			libvlc_media_player_set_hwnd( _mp, desc );
#endif
		}
	}
	else {
		//	Setup frame display callback
		_originalResizeCallback = win()->onResizeCallback( boost::bind(&MediaPlayer::onWindowResize,this,_1) );
		libvlc_video_set_callbacks(_mp, impl::vlc_lock, impl::vlc_unlock, impl::vlc_display, this );
		libvlc_video_set_format_callbacks( _mp, impl::vlc_format, impl::vlc_cleanup );
	}

	{	//	Attach to stop asyn events
		system()->dispatcher()->registerTarget( this );
		libvlc_event_manager_t *mgr=libvlc_media_player_event_manager( _mp );
		libvlc_event_attach( mgr, libvlc_MediaPlayerEndReached, impl::vlc_event, this ); 
	}

	{	//Setup audio
		if ( _needsMute ) {
			libvlc_audio_set_mute( _mp, 1 );
		}
		if (_volume != -1 ) {
			libvlc_audio_set_volume( _mp, _volume );
		}
	}

	return true;
}

void MediaPlayer::finalize() {
	LDEBUG( "vlc", "Finalize" );
	assert(_mp);

	if (!supportVideoOverlay()) {
		win()->onResizeCallback( _originalResizeCallback );
	}

	{	//	Dettach to stop asyn events
		system()->dispatcher()->unregisterTarget( this );		
		libvlc_event_manager_t *mgr=libvlc_media_player_event_manager( _mp );
		libvlc_event_detach( mgr, libvlc_MediaPlayerEndReached, impl::vlc_event, this ); 
	}

	//	Release player
	libvlc_media_player_release( _mp );
	_mp = NULL;

	MediaPlayerImpl::finalize();
}

//	Media operations
void MediaPlayer::play() {
	LDEBUG( "vlc", "Play" );
	assert(_mp);	
	//	play the media_player
	libvlc_media_player_play( _mp );
}

void MediaPlayer::stop() {
	LDEBUG( "vlc", "Stop" );
	assert(_mp);	
	//	Stop playing	
	libvlc_media_player_stop(_mp);
}

void MediaPlayer::pause( bool pause ) {
	LDEBUG( "vlc", "Pause: param=%d", pause );
	assert(_mp);	
	//	Pause playing
	libvlc_media_player_set_pause(_mp, pause ? 1 : 0 );
}

//	Volume
void MediaPlayer::mute( bool needMute ) {
	LDEBUG( "vlc", "Mute: param=%d", needMute );
	if (_mp) {
		libvlc_audio_set_mute( _mp, needMute ? 1 : 0 );
	} else {
		_needsMute = needMute;
	}
}

void MediaPlayer::setVolume( Volume vol ) {
	LDEBUG( "vlc", "Volume: param=%d", vol );
	if (_mp) {	
		libvlc_audio_set_volume( _mp, vol );
	} else {
		_volume = vol;
	}
}

Volume MediaPlayer::maxVol() {
	return 100;
}

Volume MediaPlayer::minVol() {
	return 40;
}

//	Aux
libvlc_instance_t *MediaPlayer::instance() const {
	return _player->instance();
}

canvas::Window *MediaPlayer::win() const {
	return system()->window();
}

void MediaPlayer::onWindowResize( bool start ) {
	assert(_mp);
	LDEBUG( "vlc", "Window resize" );
	if (start) {
		libvlc_media_player_stop(_mp);
	}
	else {
		libvlc_media_player_play(_mp);
	}
}

void MediaPlayer::onStopped() {
	system()->dispatcher()->post( this, boost::bind(&MediaPlayer::onMediaStopped,this) );
}

//	No overlay methods
void MediaPlayer::onMove( const Point &canvasPoint ) {
	Point point = win()->translateToWindowCoordenate( canvasPoint );	
	win()->moveVideo( point );
}

void MediaPlayer::onResize( const Size &canvasSize ) {
	Size size = win()->translateToWindowSize( canvasSize );	
	win()->resizeVideo( size );	
}

//	Render
void MediaPlayer::renderFrame( void *id ) {
	win()->renderFrame( id );
}

void *MediaPlayer::allocFrame( void **pixels ) {
	// TODO: // returns 0 on success (ie: media loaded and playing... we now know the real size)
	// unsigned int w, h;
	// if (libvlc_video_get_size( _mp, 0, &w, &h) != 0 || (!w && !h)) {
	// 	w=720;
	// 	h=576;
	// }
	// printf( "[rako] Alloc frame: size=(%d,%d)\n", w, h );
	return win()->allocFrame( pixels );	
}

void MediaPlayer::freeFrame( void *id ) {
	win()->freeFrame( id );
}

unsigned MediaPlayer::getFormat( char *chroma, unsigned *width, unsigned *height, unsigned *pitches, unsigned *lines ) {
	int nFrames=win()->getFormat( chroma, width, height, pitches, lines );
	LDEBUG( "vlc", "Format: nFrames=%d, chroma=%s, size=(%d,%d), pitches=(%d,%d,%d), lines=(%d,%d,%d)",
		nFrames, chroma, *width, *height,
		pitches[0], pitches[1], pitches[2],
		lines[0], lines[1], lines[2] );
	return nFrames;
}

void MediaPlayer::cleanup() {
	win()->cleanup();
}

}
}

