/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-canvas implementation.

    DTV-canvas is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-canvas is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-canvas.

    DTV-canvas es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-canvas se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "system.h"
#include "point.h"
#include "canvas.h"
#include "window.h"
#include "input.h"
#include "player.h"
#include "webviewer.h"
#include "generated/config.h"
#if SYS_SYSTEM_USE_GTK
#       include "impl/gtk/system.h"
#endif
#include <util/task/dispatcher.h>
#include <util/cfg/configregistrator.h>
#include <util/log.h>
#include <util/mcr.h>
#include <boost/bind.hpp>

namespace canvas {

REGISTER_INIT_CONFIG( gui ) {
	root().addNode( "gui" )
#if SYS_SYSTEM_USE_GTK
		.addValue( "use", "System to initialize", std::string("gtk") );
#else
		.addValue( "use", "System to initialize", std::string("dummy") );
#endif
}

System *System::create( const std::string &useParam/*=""*/ ) {
	System *sys=NULL;
	const std::string &use = useParam.empty() ? util::cfg::getValue<std::string> ("gui.use") : useParam;

	LINFO("System", "Using system: use=%s", use.c_str() );

#if SYS_SYSTEM_USE_GTK
	if (use == "gtk") {
		sys = new gtk::System();
	}
#endif

	assert(sys);
	return sys;
}

System::System() {
	_win = NULL;
	_canvas = NULL;
	_player = NULL;
}

System::~System() {
	CLEAN_ALL( Input *, _inputs );
	assert(!_win);
	assert(!_canvas);
	assert(!_player);
}

void System::dispatchShutdown() {
	LDEBUG("System", "Shutdown event received")
	event::Notify notify;
	notify.type = event::type::shutdown;
	dispatchNotify( &notify );
}

void System::setNotifyCallback( const event::Callback &callback ) {
	_onNotify = callback;
}

//	Initialization
bool System::initialize() {
	LDEBUG("System", "Initialize");

	//	Create all inputs
	Input::create(this);
	
	if (!init()) {
		return false;
	}

	if (_onNotify.empty()) {
		LWARN("System", "Notifycation event callback empty");
	}

	//	Create and initialize window
	if (!setupWindow()) {
		cleanup();
		return false;
	}

	//	Create and initialize canvas
	if (!setupCanvas()) {
		cleanup();
		return false;
	}

	//	Create and initialize player
	if (!setupPlayer()) {
		cleanup();
		return false;
	}

	//	Initialize inputs
	BOOST_FOREACH( Input *input, _inputs ) {
		if (!input->initialize()) {
			LWARN("System", "Cannot initialize input: name=%s\n", input->name().c_str() );
		}
	}

	return true;
}

void System::cleanup() {
	if (_canvas) {
		_canvas->finalize();
		DEL(_canvas);
	}

	if (_win) {
		_win->finalize();
		DEL(_win);
	}

	if (_player) {
		_player->finalize();
		DEL(_player);
	}
}

void System::finalize() {
	LDEBUG("System", "Finalize");
	//	Finalize inputs
	BOOST_FOREACH( Input *input, _inputs ) {
		input->finalize();
	}
	cleanup();
	fin();
}

void System::run() {
	_threadId = boost::this_thread::get_id();
	window()->mainLoopThreadId( _threadId );
	runImpl();
}

bool System::init() {
	return true;
}

void System::fin() {
}

//	Dispatcher
void System::post( util::task::Target target, const util::task::Type &task ) {
	dispatcher()->post( target, task );
}

//	HTML viewer
WebViewer *System::createWebViewer( Surface *surface ) {
	return WebViewer::create( this, surface );
}

//	Canvas
bool System::setupCanvas() {
	_canvas = createCanvas();
	if (!_canvas) {
		LERROR("System", "cannot create canvas");
		return false;
	}

	if (!_canvas->initialize( _win )) {
		LERROR("System", "cannot initialize canvas");		
		DEL(_canvas);
		return false;
	}

	return true;
}

Canvas *System::createCanvas() const {
	return Canvas::create();
}

Canvas *System::canvas() const {
	assert(_canvas);
	return _canvas;
}

//	Window
Window *System::createWindow() const {
	return Window::create();
}

Window *System::window() const {
	assert(_win);
	return _win;
}

bool System::setupWindow() {
	//	Create and initialize window
	_win = createWindow();
	if (!_win) {
		LWARN("System", "cannot create display");
		return false;
	}

	if (!_win->initialize()) {
		LWARN("System", "cannot initialize display");		
		DEL(_win);
		return false;
	}

	return true;
}

//	Player
Player *System::createPlayer() const {
	return Player::createPlayer();	
}

bool System::setupPlayer() {
	_player = createPlayer();
	if (!_player) {
		LWARN("System", "cannot create player");
		return false;
	}

	if (!_player->initialize()) {
		LWARN("System", "cannot initialize player");		
		DEL(_player);
		return false;
	}

	return true;
}

Player *System::player() const {
	return _player;
}

//	Input
void System::addInput( Input *in ) {
	LINFO( "System", "Using input: %s", in->name().c_str() );
	_inputs.push_back( in );
}

void System::dispatchKey( util::key::type keyValue, bool isUp ) {
	LDEBUG( "System", "Dispatch key: key=%d, isUp=%d", keyValue, isUp );
	event::Notify notify;
	notify.keyEvt.type = event::type::key;
	notify.keyEvt.key = keyValue;
	notify.keyEvt.isUp = isUp;
	dispatchNotify( &notify );
}

void System::dispatchNotify( event::Notify *evt ) {
	if (!_onNotify.empty()) {
		LDEBUG( "System", "Dispatch notify: type=%d", evt->type );
		_onNotify( evt );
	}
	else {
		LWARN( "System", "Notify method empty" );		
	}
}

void System::dispatchMouseNotify( event::Notify *notify ) {
	fixMouseCoordinates( notify );
	dispatchNotify( notify );
}

//	Aux mouse
void System::fixMouseCoordinates( event::Notify *notify ) {
	Point fixedCoord = _win->translateToCanvasCoordenate( Point(notify->buttonEvt.x, notify->buttonEvt.y) );
	notify->buttonEvt.x = fixedCoord.x;
	notify->buttonEvt.y = fixedCoord.y;
}

}
