/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-canvas implementation.

    DTV-canvas is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-canvas is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-canvas.

    DTV-canvas es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-canvas se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "window.h"
#include "system.h"
#include "videooverlay.h"
#include "point.h"
#include "surface.h"
#include "dummy/window.h"
#include "generated/config.h"
#include <util/log.h>
#include <util/cfg/configregistrator.h>
#include <boost/math/special_functions/round.hpp>

namespace canvas {

REGISTER_INIT_CONFIG( gui_window ) {
	root().addNode( "window" )
#if SYS_SYSTEM_USE_GTK
	.addValue( "use", "Window to instance", std::string("gtk") )
#else
	.addValue( "use", "Window to instance", std::string("dummy") )
#endif
	.addValue( "title", "Window title", std::string("Dummy") )
	.addValue( "icon", "Window icon", std::string("") )	
	.addValue( "fullscreen", "Window fullscreen option", false )
	.addValue( "winID", "Window id", 0UL )
	.addNode( "size" )
		.addValue( "width", "Window width", 720 )
		.addValue( "height", "Window height", 576 );
}

Window *Window::create() {
	const std::string &use = util::cfg::getValue<std::string>("gui.window.use");
	LDEBUG("Window", "Creating window: use=%s", use.c_str() );
	return new dummy::Window();
}

Window::Window()
{
	util::cfg::PropertyNode &root = util::cfg::get("gui.window");
	_title = root.get<std::string>("title");
	_icon = root.get<std::string>("icon");
	_isFullScreen = false;
	_isIconified = false;
}

Window::~Window()
{
	if (_overlays.size()) {
		LWARN("Window", "not all video overlays destroyed: size=%d", _overlays.size() );
	}
}

//	Initialization
bool Window::initialize() {
	if (winID() && !supportEmbedded()) {
		LERROR( "Window", "Window object not support embedded" );
		return false;
	}

	if (!init()) {
		LERROR( "Window", "Could not initialize window" );
		return false;
	}

	return true;
}

bool Window::init() {
	initSize();
	return true;
}

void Window::finalize() {
}

const std::string &Window::title() const {
	return _title;
}

const std::string &Window::icon() const {
	return _icon;
}

//	Size methods
const Size &Window::size() const {
	return _size;
}

OnResizeCallback Window::onResizeCallback( const OnResizeCallback &callback ) {
	OnResizeCallback tmp = _resizeCallback;
	_resizeCallback = callback;
	return tmp;
}

Size Window::windowSize() const {
	Size tmp;
	tmp.h = util::cfg::getValue<int>("gui.window.size.height");
	tmp.w = util::cfg::getValue<int>("gui.window.size.width");
	return tmp;
}

Size Window::screenSize() const {
	return Size(0,0);
}

void Window::initSize() {
	const bool &inFullScreen = util::cfg::getValue<bool>("gui.window.fullscreen");
	if (inFullScreen && supportFullScreen()) {
		setFullScreen( true );
	}

	setSize();
}

void Window::resize( const Size &newSize ) {
	util::cfg::setValue<int>("gui.window.size.height", newSize.h );
	util::cfg::setValue<int>("gui.window.size.width", newSize.w );

	if (isFullScreen()) {
		setFullScreen( false );
	}

	setSize();
}

void Window::setSize() {
	Size size;

	if (isFullScreen()) {
		size = screenSize();
		assert( size != Size(0,0) );
	}
	else {
		size = windowSize();
	}

	LDEBUG( "Window", "Set size: old=(%d,%d), new=(%d,%d)",
		_size.w, _size.h, size.w, size.h );

	if (size != _size) {
		//	Notify to depends
		if (!_resizeCallback.empty()) {
			_resizeCallback( true );
		}

		updateVideoBounds( size );

		//	Set new size in the config
		_size = size;
		setSizeImpl( size );


		//	Update overlays bounds
		updateBoundsOnOverlays();

		//	Notify to depends
		if (!_resizeCallback.empty()) {
			_resizeCallback( false );
		}
	}
}

void Window::setSizeImpl( const Size & /*size*/ ) {
}

bool Window::supportFullScreen() const {
	return false;
}

bool Window::isFullScreen() const {
	return _isFullScreen;
}

void Window::fullScreen( bool enable ) {
	if (supportFullScreen()) {
		LDEBUG( "Window", "fullScreen. oldState=%d newState=%d", _isFullScreen, enable );
		if (enable != _isFullScreen) {
			setFullScreen( enable );
			setSize();
		}
	} else {
		LWARN( "Window", "Window doesn't support fullScreen");
	}
}

void Window::setFullScreen( bool enable ) {
	fullScreenImpl( enable );
	_isFullScreen = enable;
}

void Window::fullScreenImpl( bool /*enable*/ ) {
}

bool Window::isIconified() const {
	return _isIconified;
}

void Window::iconify( bool enable ) {
	LDEBUG( "Window", "iconify. oldState=%d newState=%d", _isIconified, enable );
	if (enable != _isIconified) {
		//	Iconify/deicofiny the overlays
		for(size_t i=0; i < _overlays.size(); i++) {
			_overlays[i]->iconify( enable );
		}

		//	Implementation
		iconifyImpl( enable );
		_isIconified = enable;

		//	Raise to restore overlay stack
		if (!enable) {
			restackOverlays();
		}
	}
}

void Window::iconifyImpl( bool /*enable*/ ) {
}

//	Implementation
void Window::mainLoopThreadId( const boost::thread::id &id ) {
	_threadId = id;
}

const boost::thread::id &Window::mainLoopThreadId() const {
	return _threadId;
}

//	Embedded
const unsigned long &Window::winID() const {
	return util::cfg::getValue<unsigned long>("gui.window.winID");
}

bool Window::supportEmbedded() const {
	return false;
}

//	Overlays
bool Window::supportVideoOverlay() const {
	return false;
}

const std::vector<VideoOverlayPtr> &Window::overlays() const {
	return _overlays;
}

VideoOverlay *Window::createOverlayInstance() const {
	return NULL;
}

struct SortOverlays {
	bool operator()( VideoOverlay *vo1, VideoOverlay *vo2 ) const {
		return vo1->zIndex() < vo2->zIndex();
	}
};

void Window::restackOverlays() {
	if (supportVideoOverlay()) {
		LDEBUG( "Window", "Restack overlays: overlays=%d", _overlays.size() );
		//	Sort overlays by zIndex
		std::stable_sort( _overlays.begin(), _overlays.end(), SortOverlays() );
	}
}

void Window::updateBoundsOnOverlays() {
	LDEBUG( "Window", "Update bounds on overlays: overlays=%d", _overlays.size() );
	for (size_t i=0; i<_overlays.size(); i++) {
		_overlays[i]->updateBounds();
	}
}

VideoOverlay *Window::createVideoOverlay() {
	VideoOverlay *tmp=createOverlayInstance();
	if (tmp) {
		_overlays.push_back( tmp );
	}
	return tmp;
}

void Window::destroyVideoOverlay( VideoOverlayPtr &ptr ) {
	LDEBUG("Canvas", "Destroy overlay: overlay=%p", ptr );
	assert(ptr);
	std::vector<VideoOverlay *>::iterator it=std::find(
		_overlays.begin(), _overlays.end(), ptr );
	if (it != _overlays.end()) {
		delete (*it);
		_overlays.erase( it );
	}
	ptr = NULL;
}

//	Coordenates
Size Window::getCanvasSize() {
	Surface *sur = lockLayer();
	const Size &canvasSize = sur->getSize();
	unlockLayer(sur);
	return canvasSize;
}

Size Window::translateToWindowSize( const Size &size ) {
	Size tmp;
	const Size &canvasSize = getCanvasSize();

	tmp.w = boost::math::iround( (double)(size.w * _size.w) / (double)canvasSize.w );
	tmp.h = boost::math::iround( (double)(size.h * _size.h) / (double)canvasSize.h );

	return tmp;
}

Point Window::translateToWindowCoordenate( const Point &point ) {
	return translateCoordenate( point, getCanvasSize(), _size);
}

Point Window::translateToCanvasCoordenate( const Point &point ) {
	return translateCoordenate( point, _size, getCanvasSize());
}

Point Window::translateCoordenate( const Point &point, const Size &from, const Size &to ) const {
	Point tmp;

	tmp.x = boost::math::iround( (double)(point.x * to.w) / (double)from.w );
	tmp.y = boost::math::iround( (double)(point.y * to.h) / (double)from.h );

	return tmp;
}

//	Window methods
void Window::redraw() {
	canvas::Rect r(0,0,_size.w,_size.h);
	redraw( r );
}

void Window::redraw( const Rect &r ) {
	std::vector<Rect> rects;
	rects.push_back( r );
	Surface *s = lockLayer();
	if (s) {
		renderLayer( s, rects );
		unlockLayer( s );
	}
}

//	Layer methods
bool Window::initLayer( Canvas * /*canvas*/ ) {
	return false;
}

void Window::unlockLayer( Surface * /*surface*/ ) {
}

void Window::renderLayer( Surface *surface, const std::vector<Rect> &dirtyRegions ) {
	assert( mainLoopThreadId() == boost::thread::id() || mainLoopThreadId() == boost::this_thread::get_id() );
	renderLayerImpl( surface, dirtyRegions );
}

//	Video methods
const Rect &Window::videoBounds() const {
	return _videoBounds;
}

void Window::updateVideoBounds( const Size &newSize ) {
	if (_size.w>0 && _size.h>0) {
		//	Reset video bounds
		Point tmpPos;
		tmpPos.x = boost::math::iround( (double)(_videoBounds.x * newSize.w) / (double) _size.w );
		tmpPos.y = boost::math::iround( (double)(_videoBounds.y * newSize.h) / (double) _size.h );

		Size tmpSize;
		tmpSize.w = boost::math::iround( (double)(_videoBounds.w * newSize.w) / (double) _size.w );
		tmpSize.h = boost::math::iround( (double)(_videoBounds.h * newSize.h) / (double) _size.h );

		_videoBounds = tmpPos;
		_videoBounds = tmpSize;
	} else {
		_videoBounds = newSize;
	}
}

void Window::moveVideo( const Point &point ) {
	_videoBounds = point;
}

void Window::resizeVideo( const Size &size ) {
	_videoBounds = size;
}

int Window::getFormat( char * /*chroma*/, unsigned * /*width*/, unsigned * /*height*/, unsigned * /*pitches*/, unsigned * /*lines*/ ) {
	return 0;
}

void Window::cleanup() {
}

void *Window::allocFrame( void ** /*pixels*/ ) {
	return NULL;
}

void Window::freeFrame( void * /*frame*/ ) {
}

void Window::renderFrame( void * /*frame*/ ) {
}

}
