/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-luaz implementation.

    DTV-luaz is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-luaz is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-luaz.

    DTV-luaz es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-luaz se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "canvas.h"
#include "../ui.h"
#include "../lua.h"
#include "generated/config.h"
#include <zapper/plugin/pluginmanager.h>
#include <zapper/display/displayservice.h>
#include <zapper/display/zapperlayer.h>
#include <canvas/system.h>
#include <canvas/canvas.h>
#include <canvas/surface.h>
#include <canvas/font.h>
#include <util/log.h>
#include <boost/foreach.hpp>
#include <boost/filesystem.hpp>
#include <boost/math/special_functions/round.hpp>
#include <sstream>
#include <algorithm>
#include <stdlib.h>

namespace luaz {
namespace canvas {

namespace fs = boost::filesystem;

namespace impl {

typedef std::map<char, int> FontCharSizeType;
typedef std::map<int,FontCharSizeType> FontSizeType;
typedef std::map<std::string,FontSizeType> FontCache;
static zapper::display::DisplayService *display = NULL;
static FontCache fontSizes;
static std::map<std::string, ::canvas::Surface *> images;
static ::canvas::Surface *surface;
static ::canvas::Canvas *canvas;

}

//	Aux methods
static ::canvas::Surface* loadImage(const std::string& path) {
	std::string tmp;
	fs::path imgPath(path);
	if (imgPath.has_parent_path()) {
		tmp = path;
	}
	else {
		fs::path tmpImg = lua::imagesPath();
		tmpImg /= path;
		tmp = tmpImg.string();
	}

	::canvas::Surface *img;
	std::map<std::string, ::canvas::Surface *>::const_iterator it=impl::images.find(tmp);
	if (it == impl::images.end()) {
		img=impl::canvas->createSurfaceFromPath(tmp);
		if (img) {
			impl::images[tmp] = img;
		} else {
			LWARN( "modules::Canvas", "Failed to load image file: path=%s", tmp.c_str() );
		}
	}
	else {
		img = (*it).second;
	}
	return img;
}

static void setColor(const ::canvas::Color& color){
	impl::surface->setColor(color);
}

// static void readColor(lua_State* L, int ix, ::canvas::Color& color) {
// 	int rgb[3];
// 	lua_pushnil(L);

// 	for (int i=0; i < 3; ++i) {
// 		lua_next(L, ix);
// 		rgb[i] = luaL_checkint(L, -1);
// 		lua_pop(L, 1);
// 	}

// 	color.r = (util::BYTE) rgb[0];
// 	color.g = (util::BYTE) rgb[1];
// 	color.b = (util::BYTE) rgb[2];
// }

static void boundedText(int x, int y, int w, int /*h*/, const std::string& t) {
	if (t.size()) {
		int hh = 0;
		int linew = 0;
		std::string line = t;
		textSize(line, linew, hh);
		while (linew >= w) {
			line = line.substr(0,line.size()-1);
			textSize(line, linew, hh);
		}
		text(x,y, (char*)line.c_str());
	}
}

static void drawListCells(
	int x, int y,
	int rowH,
	int cell_padding,
	const ::canvas::Color& rowColor,
	const std::vector<int>& widths )
{
	for (size_t index=0; index<widths.size(); ++index) {
		setColor(rowColor);
		fillRect(x, y, widths[index], rowH);

		x += widths[index] + cell_padding;
	}
}

static void drawListItems(
	int x, int y,
	int rowH,
	int text_padding,
	int cell_padding,
	const ::canvas::Color& textColor,
	const std::vector<std::string>& values,
	const std::vector<int>& widths,
	ColumnMapping& images)
{
	int r_x = x;
	for (size_t index=0; index<values.size(); ++index) {
		setColor(textColor);

		std::string value = values[index];
		if (images.find(index) != images.end()) {
			if (images[index].find(value) != images[index].end()) {
				value = images[index][value];
			} else {
				value = "";
			}
		}

		if (value.find("img:", 0,4) == 0) { // if it starts with "img:"
			std::string path = value.substr(4,std::string::npos);
			drawImage(path.c_str(), r_x + text_padding, y);
		} else {
			boundedText(r_x + text_padding, (int)(y+ rowH*0.75), widths[index], 0, (char*)value.c_str());
		}
		r_x += widths[index] + cell_padding;
	}
}

static void drawListItems(
	int x, int y,
	int rowH,
	int text_padding,
	int cell_padding,
	const ::canvas::Color& textColor,
	const std::vector<std::string>& values,
	const std::vector<int>& widths)
{
	ColumnMapping images;
	drawListItems(x, y, rowH, text_padding, cell_padding, textColor, values, widths, images);
}

static void loadImages() {
	fs::directory_iterator end_it;
	for (fs::directory_iterator it( lua::imagesPath() ); it != end_it; ++it) {
		if( !fs::is_directory(it->status()) && fs::extension(it->path()) == ".png" ) {
			loadImage( it->path().string() );
		}
	}
}

static void cacheFonts( const std::string &fontName ) {
	char charset[]  = " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~ÁÀÉÈÍÌÓÒÚÙÜÑáàéèíìóòúùüñ";
	char charset2[] = " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~AAEEIIOOUUUNaaeeiioouuun";
	int  sizes[] = {14, 17, 22};
	
	int sizes_length = sizeof sizes/sizeof(int);
	int w,h;
	
	//	These widths belong to Tiresias font
	for (int s = 0; s < sizes_length; ++s) {
		int size = sizes[s];
		font(fontName.c_str(), size);
		for (int i = 0; charset[i] != '\0'; ++i) {
			std::string str = std::string("") + charset2[i];
			textSize(str, w, h);
			impl::fontSizes[fontName][size][charset[i]] = w;
		}
	}
}

//	API
void clear( int x, int y, int w, int h ) {
	impl::surface->clear(::canvas::Rect(x,y, w,h));
}

void fillRect( int x, int y, int w, int h ){
	impl::surface->fillRect(::canvas::Rect(x,y,w,h));
}

void drawRect(int x, int y, int w, int h){
	impl::surface->drawRect(::canvas::Rect(x,y, w,h));
}

void setColor(int r, int g, int b, int a /*=255*/) {
	setColor(::canvas::Color((util::BYTE)r,(util::BYTE)g,(util::BYTE)b,(util::BYTE)a) );
}

void font(const char* face, int size) {
	impl::surface->setFont(::canvas::Font(face, size));
}

void text(int x, int y, char* t) {	//	TODO: Check this!!
	int descent = impl::surface->fontDescent();
	y -= descent/3;
	impl::surface->drawText(::canvas::Point(x,y), t);
}

void textSize(const std::string& text, int &w, int &base) {
	::canvas::Size sz;
	impl::surface->measureText(text, sz);
	w = sz.w;
	base = sz.h;
}

int textWidth( const std::string &fontName, const std::string& text, int size) {
	int width = 0;
	for (size_t i = 0; i < text.size(); ++i) {
		width += impl::fontSizes[fontName][size][text[i]];
	}
	return width;
}

int textWidth( const std::string &fontName, char c, int size) {
	return impl::fontSizes[fontName][size][c];
}

void fillPolygon(const std::vector<std::pair<int,int> >& vertices) {
	std::vector< ::canvas::Point > points;
	for(size_t i=0; i<vertices.size(); ++i) {
		points.push_back(::canvas::Point(vertices[i].first, vertices[i].second));
	}
	impl::surface->fillPolygon(points);
}

void drawImage(const char* path, int x, int y, int /*w*/, int /*h*/) {
	if (strlen(path) > 0) {
		::canvas::Surface* img = loadImage(path);
		if (img) {
			::canvas::Point target(x,y);
			impl::surface->blit( target, img );
		}
	}
}

void drawImageFullScreen(const char* path) {
	impl::surface->drawImage( std::string(path) );
}

void drawList(
	int x, int y
	, const std::vector<int>& cell_widths
	, int rowHeight
	, int rowSep
	, const std::vector<std::string>& headers
	, const std::vector<std::vector<std::string> >& values
	, const ::canvas::Color& bgColor
	, const ::canvas::Color& cellColor
	, const ::canvas::Color& selectedCellColor
	, const ::canvas::Color& textColor
	, const ::canvas::Color& selectedTextColor
	, const ::canvas::Color& headerColor
	, const ::canvas::Color& textHeaderColor
	, int selected
	, bool with_scrollbar
	, int from
	, int to
	, int rowCount
	, const int scrollx
	, const int scrolly
	, const int scrollw
	, const int scrollh
	, const int scrollhandlex
	, const int scrollhandley
	, const int scrollhandlew
	, const int scrollhandleh
	, ColumnMapping& images_normal
	, ColumnMapping& images_selected
)
{
	const int text_padding = 5;

	int listW = 0;
	BOOST_FOREACH(int width, cell_widths) {
		listW += width + rowSep;
	}

	// background
	const int listH = (rowCount+1)*(rowHeight+rowSep) + 2; // header + rows
	setColor(bgColor);
	fillRect(x,y, listW+scrollw,listH);

	// header
	drawListCells(x, y, rowHeight, rowSep, headerColor, cell_widths);
	drawListItems(x, y, rowHeight, text_padding, rowSep, textHeaderColor, headers, cell_widths);
	y += rowHeight + rowSep + 2;

	int r_y = y;
	int r_x = x;
	// fondo de las filas
	//for (int ix=from; ix<to; ++ix) {
	for (int ix=0; ix<rowCount; ++ix) {
		drawListCells(r_x, r_y, rowHeight, rowSep, (selected==ix+from)? selectedCellColor : cellColor, cell_widths);
		r_y += rowHeight + rowSep;
	}

	if (values.size()) {
		r_y = y;
		r_x = x;

		// texto de las filas
		for (int ix=from; ix<=to; ++ix) {
			const std::vector<std::string>& row = values[ix];
			drawListItems(r_x, r_y, rowHeight, text_padding, rowSep, (selected==ix)? selectedTextColor : textColor, row, cell_widths, (selected==ix)? images_selected : images_normal);
			r_y += rowHeight + rowSep;
		}
	}
	if (with_scrollbar) {
		setColor(cellColor);
		fillRect(scrollx, scrolly, scrollw, scrollh);

		setColor(headerColor);
		fillRect(scrollhandlex, scrollhandley, scrollhandlew, scrollhandleh);
	}

}

std::pair<int,int> screenSize() {
	const ::canvas::Size &size=impl::canvas->size();
	return std::make_pair(size.w,size.h);
}

void flush() {
	impl::canvas->flush();
}

//	LUA functions
static int l_size (lua_State* L) {
	const ::canvas::Size &size=impl::canvas->size();
	lua_pushnumber(L, size.w );
	lua_pushnumber(L, size.h );
	return 2; // -> width, height
}

static int l_clear (lua_State* L) {
	int x = luaL_checkint(L,2);
	int y = luaL_checkint(L,3);
	int w = luaL_checkint(L,4);
	int h = luaL_checkint(L,5);
	clear(x,y,w,h);
	return 0;
}

static int l_and(lua_State* L) {
	int narg = 2;
	const int bitset1 = luaL_checkint(L, narg++);
	const int bitset2 = luaL_checkint(L, narg++);

	lua_pushnumber(L, bitset1 & bitset2);

	return 1;
}

static int l_flush(lua_State* /*L*/) {
	impl::canvas->flush();
	return 0;
}

static int l_setFont( lua_State *L ) {
	int narg=2;
	const char *face = luaL_checkstring(L,narg++);
	int size = luaL_checkint(L,narg++);
	impl::surface->setFont(::canvas::Font(face, size));
	return 0;
}

static int l_setColor( lua_State *L ) {
	int narg=2;
	int r = luaL_checkint(L,narg++);
	int g = luaL_checkint(L,narg++);
	int b = luaL_checkint(L,narg++);
	int a = luaL_checkint(L,narg++);
	setColor(r,g,b,a);
	return 0;
}

static int l_drawText( lua_State *L ) {
	int x, y, w, h, narg=2;
	bool boundedText = (lua_gettop(L) == 6);
	const char *txt = luaL_checkstring(L,narg++);

	x = luaL_checkint(L,narg++);
	y = luaL_checkint(L,narg++);

	if (boundedText) {
		w = luaL_checkint(L,narg++);
		h = luaL_checkint(L,narg++);
		impl::surface->drawText(::canvas::Rect(x,y,w,h), txt);
	} else {
		text(x, y, (char *) txt);
	}

	return 0;
}

static int l_measureText( lua_State *L ) {
	int narg=2;
	const char *text = luaL_checkstring(L,narg++);
	if (text) {
		::canvas::Size sz;
		impl::surface->measureText(text, sz);
		lua_pushnumber(L, sz.w );
		lua_pushnumber(L, sz.h );
		return 2;
	}
	return 0;
}

static int l_drawImage( lua_State *L ) {
	int narg=2;
	const char *img = luaL_checkstring(L,narg++);
	int x = luaL_checkint(L,narg++);
	int y = luaL_checkint(L,narg++);
	drawImage( img, x, y, -1, -1 );
	return 0;
}

static int l_imageSize( lua_State *L ) {
	const char *path = luaL_checkstring(L,2);
	::canvas::Surface* img = loadImage(path);
	::canvas::Size tmp = img->getSize();
	lua_pushnumber(L, tmp.w );
	lua_pushnumber(L, tmp.h );
	return 2;
}

static int l_fillRect( lua_State *L ) {
	int narg=2;
	int x = luaL_checkint(L,narg++);
	int y = luaL_checkint(L,narg++);
	int w = luaL_checkint(L,narg++);
	int h = luaL_checkint(L,narg++);
	fillRect( x, y, w, h );
	return 0;
}

static int l_drawRect( lua_State *L ) {
	int narg=2;
	int x = luaL_checkint(L,narg++);
	int y = luaL_checkint(L,narg++);
	int w = luaL_checkint(L,narg++);
	int h = luaL_checkint(L,narg++);
	drawRect( x, y, w, h );
	return 0;
}

static int l_cacheFonts( lua_State *L ) {
	const char *fontName = luaL_checkstring(L,2);
	assert(fontName);
	cacheFonts( fontName );
	return 0;
}

static const struct luaL_Reg canvas_methods[] = {
	{ "clear",            l_clear            },
	{ "size",             l_size             },
	{ "flush",            l_flush            },	
	{ "binary_and",       l_and              },	//	TODO: Sacar de aca!!!!!
	{ "drawImage",        l_drawImage        },
	{ "imageSize",        l_imageSize        },
	{ "drawText",         l_drawText         },
	{ "setColor",         l_setColor         },
	{ "setFont",          l_setFont          },
	{ "fillRect",         l_fillRect         },
	{ "drawRect",         l_drawRect         },
	{ "measureText",      l_measureText      },
	{ "cacheFonts",       l_cacheFonts       },	
	{ NULL,               NULL               }
};

//	Module methods
bool init( zapper::plugin::PluginManager *mgr ) {
	impl::display = mgr->addRef<zapper::display::DisplayService>( "Display" );
	impl::surface = impl::display->getLayer( zapper::display::layer::zapper )->getSurface();
	if (!impl::surface) {
		RELEASE_SERVICE(impl::display);
		return false;
	}
	impl::canvas = impl::surface->canvas();

	loadImages();
	return true;
}

void fin( void ) {
	typedef std::map<std::string, ::canvas::Surface *>::value_type vt;
	BOOST_FOREACH(vt pair, impl::images) {
		impl::canvas->destroy(pair.second);
	}

	RELEASE_SERVICE(impl::display);
}

void start( lua_State *lua ) {
	luaL_register( lua, "canvas", canvas_methods );
}



}
}

