/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-luaz implementation.

    DTV-luaz is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-luaz is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-luaz.

    DTV-luaz es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-luaz se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "channel.h"
#include "../lua.h"
#include <mpegparser/language.h>
#include <zapper/plugin/pluginmanager.h>
#include <zapper/channel/channelmanager.h>
#include <zapper/channel/channel.h>
#include <zapper/channel/show.h>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/date_time/gregorian/gregorian_types.hpp>

namespace luaz {
namespace channels {

using namespace zapper::channel;

namespace impl {
	static ChannelManager *channel = NULL;
	static const char *parentalAge[] = { "Invalid", "ATP", "+10", "+12", "+14", "+16", "+18", NULL };
}

static inline std::string formatTime( const boost::posix_time::ptime& time, const std::string& fmt ) {
	boost::posix_time::time_facet *outputFacet = new boost::posix_time::time_facet();
	outputFacet->format(fmt.c_str());
	std::stringstream ss;
	ss.imbue(std::locale(std::locale::classic(), outputFacet));
	ss << time;
	return ss.str();
}

static inline Channel *get( lua_State *L ) {
	int chID = luaL_checkint(L, 1);
	assert(chID >= 0);
	Channel *ch=impl::channel->get( chID );
	assert(ch);
	//	ch->show();
	return ch;
}

static inline util::BYTE getContent( Channel *ch, zapper::channel::Show *prg ) {
	return zapper::parental::getContent(ch->parentalContent(),prg->parentalContent());
}

static inline zapper::parental::age::type getAge( Channel *ch, zapper::channel::Show *prg ) {
	return zapper::parental::getAge( ch->parentalAge(), prg->parentalAge() );
}

static std::string getContentName( util::BYTE p ) {
	std::string tmp;

	if (p & zapper::parental::content::drugs) {
		tmp += "D";
	}

	if (p & zapper::parental::content::violence) {
		tmp += "V";
	}

	if (p & zapper::parental::content::sex) {
		tmp += "S";
	}

	return tmp;
}

static std::string getAgeName( zapper::parental::age::type age ) {
	return impl::parentalAge[age];
}

static void pushProgramGuide( lua_State *L, Channel *ch, zapper::channel::Show *program ) {
	lua_newtable(L);
	lua::setField( L, "name",            tuner::iso8859ToUTF8(program->name()) );
	lua::setField( L, "description",     tuner::iso8859ToUTF8(program->description()) );
	lua::setField( L, "start",           program->start() );
	lua::setField( L, "duration",        program->duration() );
	lua::setField( L, "parentalAge",     getAgeName( getAge(ch,program) ) );
	lua::setField( L, "parentalContent", getContentName( getContent(ch,program) ) );

	lua::setField( L, "length", program->timeRange().length().minutes() );

	lua_pushstring(L, "startDate" );
	lua_newtable(L);
	lua::setField( L, "time", formatTime(program->timeRange().begin(), "%H:%M"));
	lua::setField( L, "day",  formatTime(program->timeRange().begin(), "%Y-%m-%d"));
	lua_settable(L, -3);

	lua_pushstring(L, "stopDate" );
	lua_newtable(L);
	lua::setField( L, "time", formatTime(program->timeRange().end(), "%H:%M"));
	lua::setField( L, "day",  formatTime(program->timeRange().end(), "%Y-%m-%d"));
	lua_settable(L, -3);
}

static void pushShowInfo( lua_State *L, Channel *ch, zapper::channel::Show *program ) {
	lua::setField( L, "logo",            "img:INFO_logoCanal.png" );	//	TODO: Parsearlo
	lua::setField( L, "parentalAge",     getAgeName( getAge(ch,program) ) );
	lua::setField( L, "parentalContent", getContentName( getContent(ch,program) ) );
	lua::setField( L, "showName",        tuner::iso8859ToUTF8(program->name()) );
	lua::setField( L, "showDescription", tuner::iso8859ToUTF8(program->description()) );
}

static void pushChannel( lua_State *L, Channel *ch ) {
	lua_newtable(L);
	lua::setField( L, "channelID", ch->channelID() );
	lua::setField( L, "channel",   ch->channel() );
	lua::setField( L, "name",      ch->name() );
	lua::setField( L, "oneSeg",    ch->isOneSeg() ? 1 : 0 );
	lua::setField( L, "blocked",   ch->isBlocked() ? 1 : 0 );
	lua::setField( L, "favorite",  ch->isFavorite() ? 1 : 0 );
}

static void onNewChannel( lua_State *L, Channel *ch ) {
	lua_getfield( L, LUA_GLOBALSINDEX, "zapperProgramFound" );
	pushChannel( L, ch );
	lua_call( L, 1, 0 );
}

static void onChannelChanged( lua_State *L, Channel *ch ) {
	assert(ch);
	lua_getfield( L, LUA_GLOBALSINDEX, "zapperChannelChanged" );
	lua_pushnumber( L, ch->channelID() );
	lua_call( L, 1, 0 );
}

static int l_haveChannels( lua_State *L ) {
	bool have=impl::channel->haveChannels();
	lua_pushnumber( L, have ? 1 : 0 );
	return 1;
}

static int l_channelCount( lua_State *L ) {
	const int n = impl::channel->channelCount();
	lua_pushnumber( L, n);
	return 1;
}

static int l_getAll( lua_State *L ) {
	Channels chs;
	impl::channel->getAll( chs );
	lua_newtable(L);
	for (size_t ch=0; ch<chs.size(); ch++) {
		//	Key
		lua_pushnumber( L, ch +1);

		//	Value
		pushChannel( L, chs[ch] );

		//	Set to Services
		lua_settable( L, -3 );
	}
	return 1;
}

static int l_get( lua_State *L ) {
	pushChannel( L, get(L) );
	return 1;
}

static int l_getInfo( lua_State *L ) {
	//	Get channel
	Channel *ch=get(L);

	//	Get EPG program
	zapper::channel::Show *prg=ch->getCurrentShow();
	assert(prg);
	//prg->show();

	pushChannel( L, ch );
	pushShowInfo( L, ch, prg );
	return 1;
}

static int l_getShowsBetween( lua_State *L ) {
	zapper::channel::Shows progs;
	Channel *ch;

	{	//	Get parameters
		int narg = 1;
		const int chID = luaL_checkint(L, narg++);
		const char *time1 = luaL_checkstring(L, narg++);
		const char *time2 = luaL_checkstring(L, narg++);

		//	Get channel
		ch=impl::channel->get( chID );
		assert(ch);

		//	Get shows
		using boost::posix_time::ptime;
		using boost::posix_time::time_period;
		using boost::posix_time::time_from_string;
		time_period timeRange(time_from_string(time1), time_from_string(time2));
		ch->getShowsBetween( progs, timeRange );
	}

	int index=1;
	lua_newtable(L);		
	BOOST_FOREACH( zapper::channel::Show *program, progs ) {
		lua_pushnumber( L, index++ ); // Key
		pushProgramGuide( L, ch, program );
		lua_settable( L, -3 ); // Set table
	}
	return 1;
}

static int l_toggleOneSeg( lua_State *L ) {
	bool show=impl::channel->toggleOneSeg();
	lua_pushnumber( L, show ? 1 : 0 );
	return 1;
}

static int l_showOneSeg( lua_State *L ) {
	bool show=impl::channel->showOneSeg();
	lua_pushnumber( L, show ? 1 : 0 );
	return 1;
}

static int l_remove( lua_State *L ) {
	impl::channel->remove( luaL_checkint(L, 1) );
	return 0;
}

static int l_isProtected( lua_State *L ) {
	Channel *ch=get(L);
	lua_pushnumber( L, ch->isProtected() ? 1 : 0 );
	return 1;
}

static int l_isBlocked( lua_State *L ) {
	Channel *ch=get(L);	
	lua_pushnumber( L, ch->isBlocked() ? 1 : 0 );
	return 1;
}

static int l_toggleBlocked( lua_State *L ) {
	Channel *ch=get(L);	
	ch->toggleBlocked();
	return 0;
}

static int l_unblockAll( lua_State * /*L*/ ) {
	impl::channel->unblockAll();
	return 0;
}

static int l_toggleFavorite( lua_State *L ) {
	Channel *ch=get(L);		
	ch->toggleFavorite();
	return 0;
}

static int l_next( lua_State *L ) {
	int first  = luaL_checkint(L, 1);
	int factor = luaL_checkint(L, 2);
	int ch=impl::channel->nextChannel( first, factor );
	lua_pushnumber( L, ch );
	return 1;
}

static int l_nextFavorite( lua_State *L ) {
	int first  = luaL_checkint(L, 1);
	int factor = luaL_checkint(L, 2);
	int ch=impl::channel->nextFavorite( first, factor );
	lua_pushnumber( L, ch );
	return 1;
}

static const struct luaL_Reg channels_methods[] = {
	{ "count",           l_channelCount    },
	{ "haveChannels",    l_haveChannels    },
	{ "get",             l_get             },
	{ "getShowsBetween", l_getShowsBetween },
	{ "getInfo",         l_getInfo         },
	{ "getAll",          l_getAll          },
	{ "next",            l_next            },
	{ "nextFavorite",    l_nextFavorite    },
	{ "toggleFavorite",  l_toggleFavorite  },
	{ "remove",          l_remove          },
	{ "isBlocked",       l_isBlocked       },
	{ "unblockAll",      l_unblockAll      },
	{ "toggleBlocked",   l_toggleBlocked   },
	{ "toggleOneSeg",    l_toggleOneSeg    },
	{ "showOneSeg",      l_showOneSeg      },
	{ "isProtected",     l_isProtected     },	
	{ NULL,              NULL              }
};

//	Public methods
void attach( std::vector<std::string> &deps ) {
	deps.push_back( "ChannelManager" );
}

void initialize( zapper::plugin::PluginManager *mgr ) {
	impl::channel = mgr->addRef<ChannelManager>( "ChannelManager" );
}

void finalize( void ) {
	RELEASE_SERVICE(impl::channel);
}

void start( lua_State *st ) {
	impl::channel->onNewChannel().connect( boost::bind(&onNewChannel,st,_1) );
	impl::channel->onChannelChanged().connect( boost::bind(&onChannelChanged,st,_1) );
	luaL_register( st, "channels", channels_methods );
}

void stop( void ) {
}

}
}

