/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-luaz implementation.

    DTV-luaz is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-luaz is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-luaz.

    DTV-luaz es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-luaz se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "display.h"
#include "../lua.h"
#include <zapper/plugin/pluginmanager.h>
#include <zapper/display/displayservice.h>

namespace luaz {
namespace display {

using namespace zapper::display;

namespace impl {
static DisplayService *display = NULL;
typedef std::pair<connector::type,const char *> ConnItem;
typedef std::pair<mode::type,const char *> VMItem;
typedef std::pair<aspect::type,const char *> VAItem;
typedef std::pair<modulator::type,const char *> ModulatorItem;
typedef std::vector<ConnItem> ConnCol;
typedef std::vector<VMItem> VMCol;
typedef std::vector<VAItem> VACol;
typedef std::vector<ModulatorItem> ModulatorCol;

static ConnCol connectors;
static VMCol vm[connector::LAST_CONNECTOR];
static VACol va;
static ModulatorCol modulators;
}

#define BUILD_NAMES(Type,vValues,getValuesMethod,getNameMethod)	  \
	{ \
		vValues.clear(); \
		std::vector<Type> values = impl::display->getValuesMethod; \
		for (size_t i=0; i<values.size(); i++) { \
			Type val = values[i]; \
			vValues.push_back( std::make_pair(val,getNameMethod) ); \
		}\
	}

#define BUILD_VM_NAMES(conn) \
	BUILD_NAMES( mode::type, impl::vm[conn], supportedModes( conn ), getMode(val).name );

//	Connectors
static int l_getConnector( lua_State *L ) {
	return lua::get( L, impl::display->getActiveConnector(), impl::connectors );	
}

static int l_getConnectors( lua_State *L ) {
	return lua::getModes( L, impl::display->getActiveConnector(), impl::connectors );	
}

//	Video modes
static int getVideoMode( lua_State *L, connector::type conn ) {
	return lua::get( L, impl::display->getCurrentVideoMode( conn ), impl::vm[conn] );
}

static int getVideoModes( lua_State *L, connector::type conn ) {
	return lua::getModes( L, impl::display->getCurrentVideoMode( conn ), impl::vm[conn] );
}
	
static int l_getVideoMode( lua_State *L ) {
	connector::type conn=lua::getIndex(L,impl::connectors);
	return getVideoMode( L, conn );
}

static int l_getVideoModes( lua_State *L ) {
	connector::type conn=lua::getIndex(L,impl::connectors);	
	return getVideoModes( L, conn );
}

static int l_getCurVideoMode( lua_State *L ) {
	return getVideoMode( L, impl::display->getActiveConnector() );
}

static int l_getCurVideoModes( lua_State *L ) {
	return getVideoModes( L, impl::display->getActiveConnector() );
}

static int l_setVideoMode( lua_State *L ) {
	connector::type conn=lua::getIndex(L,impl::connectors);
	mode::type mode=lua::getIndex(L,impl::vm[conn],2);
	impl::display->setMode( conn, mode );
	return 0;
}

//	Transparency
static int l_setTransparency( lua_State *L ) {
	impl::display->setTransparency( luaL_checkint(L, 1) );
	return 0;
}

static int l_getTransparency( lua_State *L ) {
	int value=impl::display->getTransparency();
	lua_pushnumber( L, value );
	return 1;
}

//	Aspect ratio
static int l_getAspectMode( lua_State *L ) {
	return lua::get( L, impl::display->getCurrentAspect(), impl::va );
}

static int l_getAspectModes( lua_State *L ) {
	return lua::getModes( L, impl::display->getCurrentAspect(), impl::va );
}

static int l_setAspectMode( lua_State *L ) {
	impl::display->setAspect( lua::getIndex(L,impl::va) );
	return 0;
}

//	Modulators
static int l_getModulator( lua_State *L ) {
	return lua::get( L, impl::display->getCurrentModulator(), impl::modulators );	
}

static int l_getModulators( lua_State *L ) {
	return lua::getModes( L, impl::display->getCurrentModulator(), impl::modulators );	
}

static int l_setModulator( lua_State *L ) {
	impl::display->setModulator( lua::getIndex(L,impl::modulators) );	
	return 0;
}

static const struct luaL_Reg display_methods[] = {
	{ "getConnectors",       l_getConnectors       },
	{ "getConnector",        l_getConnector        },	
	{ "getTransparency",     l_getTransparency     },
	{ "setTransparency",     l_setTransparency     },
	{ "getCurVideoMode",     l_getCurVideoMode     },		
	{ "getCurVideoModes",    l_getCurVideoModes    },	
	{ "getVideoMode",        l_getVideoMode        },
	{ "getVideoModes",       l_getVideoModes       },
	{ "setVideoMode",        l_setVideoMode        },
	{ "getAspectModes",      l_getAspectModes      },
	{ "setAspectMode",       l_setAspectMode       },
	{ "getAspectMode",       l_getAspectMode       },
	{ "getModulator",        l_getModulator        },	
	{ "getModulators",       l_getModulators       },
	{ "setModulator",        l_setModulator        },
	{ NULL,                  NULL                  }
};

//	Public methods
void attach( std::vector<std::string> &deps ) {
	deps.push_back( "Display" );
}

void initialize( zapper::plugin::PluginManager *mgr ) {
	impl::display = mgr->addRef<DisplayService>( "Display" );
}

void finalize( void ) {
	RELEASE_SERVICE(impl::display);
}

void start( lua_State *st ) {
	{	//	Build display resolutions
		std::vector<connector::type> supported=impl::display->supportedConnectors();
		connector::type conn;
		for (size_t i=0; i<supported.size(); i++) {
			conn=supported[i];
			BUILD_VM_NAMES( conn );
		}
	}
	
	BUILD_NAMES( connector::type, impl::connectors, supportedConnectors(), getConnectorName(val) );	
	BUILD_NAMES( aspect::type, impl::va, supportedAspects(), getAspect(val).name );	
	BUILD_NAMES( modulator::type, impl::modulators, supportedModulators(), getModulatorName(val) );	
	luaL_register( st, "display", display_methods );
}

void stop( void ) {
}

}
}

