/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-luaz implementation.

    DTV-luaz is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-luaz is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-luaz.

    DTV-luaz es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-luaz se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "mainwindow.h"
#include "../lua.h"
#include "../view/view.h"
#include "canvas.h"
#include "module.h"
#include "generated/config.h"
#include <zapper/plugin/pluginmanager.h>
#include <zapper/input/inputservice.h>
#include <zapper/zapper.h>
#include <canvas/rect.h>
#include <canvas/surface.h>
#include <canvas/canvas.h>
#include <canvas/system.h>
#include <canvas/window.h>
#include <mpegparser/time.h>
#include <util/mcr.h>
#include <boost/foreach.hpp>
#include <boost/filesystem.hpp>
#include <assert.h>

#define BACKGROUND_LAYER 1

namespace luaz {

namespace impl {
	static MainWindow *win = NULL;
}

//	Lua events methods
static int l_run( lua_State * /*L*/ ) {
	impl::win->run();
	return 0;
}

static int l_stop( lua_State *L ) {
	int retCode=luaL_checkint(L, 1);
	impl::win->stop( retCode );
	return 0;
}

static int l_lockKeys( lua_State *L ) {
	impl::win->lockKeys( luaL_checkint(L, 1) == 1 );
	return 0;
}

static int l_resetConfig( lua_State * /*L*/ ) {
	impl::win->reset();
	return 0;
}

static int l_getTime( lua_State *L ) {
	const char *tmp = luaL_checkstring(L, 1);
	assert(tmp);
	return impl::win->getTime( tmp );
}

static int l_getDate( lua_State *L ) {
	const char *tmp = luaL_checkstring(L, 1);
	const int daysOffset = luaL_checkinteger(L, 2);
	assert(tmp);
	return impl::win->getDate( tmp , daysOffset );
}

static int l_cancelTimer( lua_State *L ){
	impl::win->cancelTimer( luaL_checkint(L, 1) );
	return 0;
}

static int l_registerTimer( lua_State *L ) {
	int tid = luaL_checkint(L, 1);
	int ms  = luaL_checkint(L, 2);
	impl::win->registerTimer( tid, ms );
	return 0;
}

static int l_showBackground( lua_State *L ) {
	int needShow = luaL_checkint(L, 1);	
	impl::win->showBackground( needShow == 1 );
	return 0;
}

static int l_launchBrowser( lua_State * /*L*/ ) {
	impl::win->launchBrowser();
#ifndef _WIN32
	return 0;
#endif
}

static int l_fullscreen( lua_State *L ) {
	impl::win->fullscreen( lua_toboolean(L, 1) != 0 );
	return 0;
}

static int l_isFullscreen( lua_State *L ) {
	int tmp = impl::win->isFullscreen() ? 1 : 0;
	lua_pushboolean( L, tmp );
	return 1;
}

static int l_iconify( lua_State *L ) {
	impl::win->iconify( lua_toboolean(L, 1) != 0 );
	return 0;
}

static const struct luaL_Reg mainwindow_methods[] = {
	{ "run",              l_run              },
	{ "stop",             l_stop             },
	{ "resetConfig",      l_resetConfig      },
	{ "lockKeys",         l_lockKeys         },
	{ "getTime",          l_getTime          },
	{ "getDatePlus",      l_getDate          },
	{ "registerTimer",    l_registerTimer    },
	{ "cancelTimer",      l_cancelTimer      },
	{ "launchBrowser",    l_launchBrowser    },
	{ "showBackground",   l_showBackground   },
	{ "fullscreen",       l_fullscreen       },
	{ "isFullscreen",     l_isFullscreen     },
	{ "iconify",          l_iconify          },
	{ NULL,               NULL               }
};


MainWindow::MainWindow( zapper::Zapper *zapper, lua_State *lua, bool supportWidgets )
	: zapper::plugin::ServiceImpl<MainWindow>( "MainWindow" )
{
	_lua = lua;
	_input = NULL;
	_zapper = zapper;
	_supportWidgets = supportWidgets;
	_bgSurface = NULL;

	//	Setup MainWindow
	impl::win = this;
}

MainWindow::~MainWindow()
{
	impl::win = NULL;
}

//	Run main lua loop
void MainWindow::run() {
	_zapper->run();
}

void MainWindow::stop( int ret ) {
	_zapper->exit( ret );
}

//	Services virtual methods
void MainWindow::onAttach() {
	addSimpleDependency( "Input" );

	std::vector<std::string> deps;
	module::attach( deps );
	BOOST_FOREACH( const std::string &dep, deps ) {
		addSimpleDependency( dep );
	}
}

bool MainWindow::onOnline() {
	//	Initialize modules
	if (!canvas::init( mgr() )) {
		LWARN("MainWindow", "Cannot init canvas module");
		return false;
	}
	module::initialize( mgr() );
	return true;
}

void MainWindow::onOffline() {
	module::finalize();
	canvas::fin();
}

bool MainWindow::onStart() {
	//	Register lua modules
	luaL_register( _lua, "mainWindow", mainwindow_methods );
	canvas::start( _lua );
	if (_supportWidgets) {
		wgt::View::init(_lua);
	}
	module::start( _lua );

	//	Get input manager
	_input = mgr()->addRef<zapper::input::InputService>( "Input" );
	assert(_input);

	//	Register all keys into for zapper
	registerKeys();

	initBgSurface();

	return true;
}

void MainWindow::onStop() {
	//	destroy background surface
	sys()->canvas()->destroy( _bgSurface );

	//	Un-register lua modules
	module::stop();
	if (_supportWidgets) {
		wgt::View::fin();
	}
	RELEASE_SERVICE(_input);
}

void MainWindow::launchBrowser() {
	exit(163);
}

//	Reset config
void MainWindow::reset() {
	mgr()->resetConfig();
}

void MainWindow::onKey( util::key::type key, bool isUp ) {
	if (key != util::key::null) {
		lua::call( _lua, "OnKeyEvent", key, isUp );
	}
}

void MainWindow::lockKeys( bool needKeys ) {
	_input->renice( "lua_zapper", needKeys ? HI_KEY_PRIORITY : LOW_KEY_PRIORITY );
}

void MainWindow::registerKeys() {
	zapper::input::Keys keys;
#define SAVE_KEY( c, n, u ) keys.push_back( static_cast<util::key::type>(util::key::c) );
	UTIL_KEY_LIST(SAVE_KEY);
#undef SAVE_KEY
	_input->reserveKeys( "lua_zapper", boost::bind(&MainWindow::onKey,this,_1,_2), keys, LOW_KEY_PRIORITY );
	_input->reserveButtons( boost::bind(&MainWindow::onButtonEvent,this,_1) );

	//	Export keys to zapper
#define DO_STR(s) #s
#define EXPORT_KEY( c, n, u )	  \
	lua_pushnumber( _lua, util::key::c ); \
	lua_setglobal ( _lua, DO_STR(KEY_##u) );
	UTIL_KEY_LIST(EXPORT_KEY);
#undef EXPORT_KEY
#undef DO_STR
}

void MainWindow::onButtonEvent( ::canvas::event::ButtonEvent *evt ) {
	lua::call( _lua, "OnButtonEvent", evt->button, evt->isPress ? 1 : 0, evt->x, evt->y );
}

void MainWindow::initBgSurface() {
	::canvas::Canvas *canv= sys()->canvas();

	const ::canvas::Rect bgRect( ::canvas::Point(0,0), sys()->canvas()->size() );
	_bgSurface = canv->createSurface( bgRect );
	_bgSurface->setZIndex( BACKGROUND_LAYER );
	_bgSurface->setVisible(false);
	_bgSurface->autoFlush(true);

	boost::filesystem::path tmp = lua::imagesPath();
	tmp /= "inicio.png";

	_bgSurface->drawImage( tmp.string() );
}

void MainWindow::showBackground( bool show ) {
	_bgSurface->setVisible( show );
	sys()->canvas()->flush();
}

void MainWindow::registerTimer( int tID, int ms ) {
	cancelTimer( tID );
	_timers[tID] = sys()->registerTimer(ms, boost::bind(&MainWindow::processTimer, this, _1, tID));
}

void MainWindow::cancelTimer( int tID ) {
	MapOfTimers::iterator it=std::find_if(
		_timers.begin(),
		_timers.end(),
		boost::bind(&MapOfTimers::value_type::first,_1) == tID
	);
	if(it != _timers.end()) {
		sys()->unregisterTimer( (*it).second );
		_timers.erase( it );		
	}
}

void MainWindow::processTimer( util::id::Ident & /*id*/, int tID ) {
	//	Find timer
	MapOfTimers::iterator it=std::find_if(
		_timers.begin(), _timers.end(), boost::bind(&MapOfTimers::value_type::first,_1) == tID);
	if(it != _timers.end()) {
		//	Timer found, destoy timer
		_timers.erase( it );

		lua::call( _lua, "OnTimerEvent", tID );
	}
}

//	Time
int MainWindow::getTime( const std::string &format ) {
	tuner::pt::ptime now;
	tuner::time::clock.get( now );
	char buffer[256];
	struct tm curtime = tuner::pt::to_tm(now);
	strftime( buffer, 256, format.c_str(), &curtime );
	lua_pushstring( _lua, buffer );
	return 1;
}

// Date
int MainWindow::getDate( const std::string &format, int daysOffset ) {
	tuner::pt::ptime now;
	tuner::time::clock.get( now );
	char buffer[256];
	now += boost::gregorian::date_duration(daysOffset);
	struct tm dd = tuner::pt::to_tm(now);
	strftime( buffer, 256, format.c_str(), &dd );
	lua_pushstring( _lua, buffer );
	return 1;
}

void MainWindow::fullscreen( bool enable ) {
	sys()->window()->fullScreen( enable );
}

bool MainWindow::isFullscreen() const {
	return sys()->window()->isFullScreen();
}

void MainWindow::iconify( bool enable ) {
	sys()->window()->iconify( enable );
}

::canvas::System *MainWindow::sys() const {
	return _zapper->sys();
}

}
