/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-luaz implementation.

    DTV-luaz is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-luaz is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-luaz.

    DTV-luaz es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-luaz se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "player.h"
#include "canvas.h"
#include "../lua.h"
#include <zapper/mount/mountservice.h>
#include <zapper/audio/service.h>
#include <zapper/player/playerservice.h>
#include <zapper/player/mediaplayer.h>
#include <zapper/plugin/pluginmanager.h>
#include <boost/filesystem.hpp>

namespace luaz {
namespace player {

namespace fs = boost::filesystem;

namespace impl {
	typedef std::pair<zapper::audio::channel::type,const char *> ChannelItem;
	typedef std::vector<ChannelItem> ChannelCol;

	static zapper::audio::Service *mixer = NULL;
	static zapper::player::PlayerService *player = NULL;
	static zapper::player::MediaPlayer *audioPlayer = NULL;
	static zapper::player::MediaPlayer *splashPlayer = NULL;
	static ChannelCol aChannels;

	static zapper::MountService *mount = NULL;
	static std::vector<std::string> audioFiles;
	static std::vector<std::string> imageFiles;
}

static void buildChannels( void ) {
	impl::aChannels.clear();
	std::vector<zapper::audio::channel::type> channels = impl::mixer->supportedAudioChannels();
	for (size_t i=0; i<channels.size(); i++) {
		zapper::audio::channel::type id = channels[i];
		zapper::audio::Channel &ch = zapper::audio::getChannelInfo( id );
		impl::aChannels.push_back( std::make_pair(id,ch.name) );
	}
}

static void onAudioStop( lua_State *L ) {
	lua::call( L, "audioFileEnd" );	
}

static void pushFileInfo( lua_State *L, int id, const std::string &file ) {
	fs::path p(file);
	lua_newtable(L);
	lua::setField( L, "id", id );
	lua::setField( L, "name", p.filename().string() );
}

static int buildFiles( lua_State *L, const std::vector<std::string> &files ) {
	lua_newtable(L);
	for (size_t i=0; i<files.size(); i++) {
		//	Key
		lua_pushnumber(L, i+1 );
		//	Value
		pushFileInfo( L, i, files[i] );
		//	Set to table
		lua_settable(L, -3);
	}
	return 1;
}

static void pushImageFileInfo( lua_State *L, int id, const std::string &file ) {
	fs::path p(file);
	lua_newtable(L);
	lua::setField( L, "id", id );
	lua::setField( L, "name", p.filename().string() );
	lua::setField( L, "path", file );
}

static int buildImageFiles( lua_State *L, const std::vector<std::string> &files ) {
	lua_newtable(L);
	for (size_t i=0; i<files.size(); i++) {
		//	Key
		lua_pushnumber(L, i+1 );
		//	Value
		pushImageFileInfo( L, i, files[i] );
		//	Set to table
		lua_settable(L, -3);
	}
	return 1;
}

static int l_toggleMute( lua_State *L ) {
	bool muted=impl::mixer->toggleMute();
	lua_pushnumber( L, muted ? 1 : 0 );		
	return 1;
}

static int l_isMuted( lua_State *L ) {
	bool muted=impl::mixer->isMuted();
	lua_pushnumber( L, muted ? 1 : 0 );	
	return 1;	
}

static int l_setVolume( lua_State *L ) {
	int vol = luaL_checkint(L, 1);
	int res=impl::mixer->volume( vol );
	lua_pushnumber( L, res );	
	return 1;	
}

static int l_volumeUp( lua_State *L ) {
	int vol=impl::mixer->volumeUp();
	lua_pushnumber( L, vol );	
	return 1;	
}

static int l_volumeDown( lua_State *L ) {
	int vol=impl::mixer->volumeDown();
	lua_pushnumber( L, vol );
	return 1;
}

static int l_maxVolume( lua_State *L ) {
	lua_pushnumber( L, 20 );
	return 1;
}

static int l_getVolume( lua_State *L ) {
	zapper::audio::Volume vol=impl::mixer->volume();
	lua_pushnumber( L, vol );	
	return 1;
}

static int l_getAudioChannels( lua_State *L ) {
	return lua::getModes( L, impl::mixer->getCurrentAudioChannel(), impl::aChannels );
}

static int l_getAudioChannel( lua_State *L ) {
	return lua::get( L, impl::mixer->getCurrentAudioChannel(), impl::aChannels );
}

static int l_setAudioChannel( lua_State *L ) {
	impl::mixer->setAudioChannel( lua::getIndex(L,impl::aChannels) );
	return 0;
}

static int l_getImageFiles( lua_State *L ) {
	impl::imageFiles.clear();
	impl::mount->getFiles( ".jpg", impl::imageFiles );
	impl::mount->getFiles( ".png", impl::imageFiles );
	return buildImageFiles( L, impl::imageFiles );
}

static int l_getAudioFiles( lua_State *L ) {
	impl::audioFiles.clear();
	impl::mount->getFiles( ".mp3", impl::audioFiles );
	return buildFiles( L, impl::audioFiles );
}

static int l_playAudioFile( lua_State *L ) {
	size_t id = luaL_checkint(L, 1);
	if (id < impl::audioFiles.size()) {
		impl::audioPlayer->play( impl::audioFiles[id] );
	}
	return 0;
}

static int l_pauseAudioFile( lua_State *L ) {
	int pause = luaL_checkint(L, 1);
	impl::audioPlayer->pause(pause == 1);
	return 0;
}

static int l_stopAudioFile( lua_State * /*L*/ ) {
	impl::audioPlayer->stop();
	return 0;
}

static const struct luaL_Reg player_methods[] = {
	{ "toggleMute",       l_toggleMute        },
	{ "isMuted",          l_isMuted           },	
	{ "setVolume",        l_setVolume         },
	{ "getVolume",        l_getVolume         },
	{ "volumeUp",         l_volumeUp          },
	{ "volumeDown",       l_volumeDown        },
	{ "maxVolume",        l_maxVolume         },
	{ "getAudioChannel",  l_getAudioChannel   },
	{ "setAudioChannel",  l_setAudioChannel   },
	{ "getAudioChannels", l_getAudioChannels  },
	{ "getImageFiles",    l_getImageFiles     },	
	{ "getAudioFiles",    l_getAudioFiles     },
	{ "playAudioFile",    l_playAudioFile     },
	{ "pauseAudioFile",   l_pauseAudioFile    },
	{ "stopAudioFile",    l_stopAudioFile     },
	{ NULL,               NULL                }
};

//	Public methods
void attach( std::vector<std::string> &deps ) {
	deps.push_back( "Mixer" );
	deps.push_back( "Player" );
	deps.push_back( "Mount" );
}

void initialize( zapper::plugin::PluginManager *mgr ) {
	impl::mixer = mgr->addRef<zapper::audio::Service>( "Mixer" );
	assert(impl::mixer);
	
	impl::player = mgr->addRef<zapper::player::PlayerService>( "Player" );
	assert(impl::player);
	
	impl::mount = mgr->addRef<zapper::MountService>( "Mount" );
	assert(impl::mount);

	//	Register extensions
	impl::mount->registerExtension( ".mp3" );
	impl::mount->registerExtension( ".jpg" );
	impl::mount->registerExtension( ".png" );
}

void finalize( void ) {
	impl::player->destroy(impl::audioPlayer);
	impl::player->destroy(impl::splashPlayer);
	RELEASE_SERVICE( impl::player );
	RELEASE_SERVICE( impl::mixer );		
	RELEASE_SERVICE( impl::mount );	
}

void start( lua_State *st ) {
	buildChannels();
	luaL_register( st, "player", player_methods );
	impl::audioPlayer = impl::player->create( zapper::player::type::audio );
	assert(impl::audioPlayer);
	impl::audioPlayer->onPlayStop().connect( boost::bind(&onAudioStop,st) );
	
	impl::splashPlayer = impl::player->create( zapper::player::type::video );
	assert(impl::splashPlayer);
}

void stop( void ) {
}

}
}

