/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-luaz implementation.

    DTV-luaz is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-luaz is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-luaz.

    DTV-luaz es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-luaz se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "view.h"
#include "widget/window.h"
#include "widget/borderedwindow.h"
#include "widget/label.h"
#include "widget/maskedlabel.h"
#include "widget/image.h"
#include "widget/imageviewer.h"
#include "widget/list.h"
#include "widget/textarea.h"
#include "widget/page.h"
#include "widget/tabbedwindow.h"
#include "widget/button.h"
#include "widget/bar.h"
#include "widget/combobox.h"
#include "widget/frame.h"
#include "widget/schedule.h"
#include "widget/schedulerow.h"
#include "modules/wgt.h"
#include "../modules/canvas.h"
#include "../lua.h"
#include <util/log.h>
#include <util/mcr.h>
#include <boost/make_shared.hpp>
#include <boost/foreach.hpp>
#include <boost/bind.hpp>
#include <vector>
#include <iostream>
#include <sstream>
#include <cmath>
#include <algorithm>
#include <stdio.h>

#define EMPTY_CC_W 225
#define EMPTY_CC_H  35

namespace wgt{
	namespace impl {
		View* instance = NULL;
	}

	void View::init(lua_State *L)
	{
		impl::instance = new View();
		wgt::Wgt::init(impl::instance, L);
	}
	void View::fin()
	{
		LDEBUG( "View", "Finalizing" );
		wgt::Wgt::fin();
		DEL(impl::instance);
	}

	View* View::instance() {
		return impl::instance;
	}

	View::View()
		: _cId(0)
		, _wId(0)
		, _fontFace("Tiresias")
		, _needsFullRedraw(true)
	{}

	View::~View(){}

	ColorId View::addColor(const Color& color){
		ColorId id = getNextColorId();
		_colors[id] = color;
		return id;
	}

	WidgetId View::addWidget(WidgetPtr widget){
		WidgetId id = getNextWidgetId();
		_keyHandlingWidgets[id] = widget;
		widget->id(id);
		return id;
	}
	
	WidgetId View::getNextWidgetId(){
		return _wId++;
	}

	ColorId View::getNextColorId(){
		return _cId++;
	}
	
	Color  View::getColorById(ColorId id){
		return _colors[id];
	}
	
	WidgetPtr View::getWidgetById(WidgetId id){
		return _keyHandlingWidgets[id];
	}

	WidgetPtr View::getTopWidget() {

		if (_widgets.size()) {
			return _widgets.back();
		} else {
			return WidgetPtr();
		}
	}

	void View::push(WidgetPtr widget){
		_needsFullRedraw = true;
		_widgets.push_back(widget);
		std::sort(_widgets.begin(), _widgets.end(), Widget::compareWidgets);
	}
	
	void View::pop(WidgetPtr widget){
		_needsFullRedraw = true;

		std::vector<WidgetPtr>::iterator it = std::find(_widgets.begin(), _widgets.end(), widget);
		_widgetsToClear.push_back(*it);
		_widgets.erase(it);
		std::sort(_widgets.begin(), _widgets.end(), Widget::compareWidgets);
	}

	// void View::setInteractivityIndicator(WidgetPtr w)
	// {
	// 	//_i = w;
	// }

	// void View::showInteractivity(bool b)
	// {
	// 	_i->setVisible(b);
	// 	if (!b) {
	// 		_i->needsClear(true);
	// 	}
	// }

	void View::drawWidget(WidgetPtr widget) {
		if (widget->needsClear()) {
			LDEBUG( "view", "Clear widget" );
			widget->clear(this);
		}
		if (widget->isVisible()) {
			widget->draw(this);
		}
	}

	void View::refresh() {
		if (_widgets.size()) {
			//drawWidget(_i);

			BOOST_FOREACH(WidgetPtr widget, _widgetsToClear) {
				clear(widget.get());
			}
			if (_needsFullRedraw) {
				LDEBUG( "view", "draw, full redraw" );
				BOOST_FOREACH(WidgetPtr widget, _widgets) {
					_desaturateColors = widget != getTopWidget();
					drawWidget(widget);
				}
				_needsFullRedraw = false;
				_desaturateColors = false;
			} else {
				LDEBUG( "view", "draw, draw top widget" );
				_desaturateColors = false;
				drawWidget(getTopWidget());
			}
		} else {
			LDEBUG( "view", "draw, clear whole screen" );
			BOOST_FOREACH(WidgetPtr widget, _widgetsToClear) {
				widget->needsClear(true);
				widget->clear(this);
			}
			//drawWidget(_i);
		}
		_widgetsToClear.clear();

		luaz::canvas::flush();
	}
	
	void View::clear(Widget* widget){
		luaz::canvas::clear(widget->x(), widget->y(), widget->w(), widget->h());
	}
	
	Color View::adjust(const Color& c)
	{
		if (c!=Color::None && _desaturateColors) {
			return c.toGrayscale();
		}
		return c;
	}

	void View::drawChildren(Window* window){
		WidgetPtr active;
		BOOST_FOREACH(WidgetPtr widget, window->getChildren()) {
			if (widget && widget->isVisible()) {
				if (widget->isActive()) {
					active = widget;
				} else {
					widget->draw(this);
				}
			}
		}
		if (active) {
			active->draw(this);
		}
	}
	
	void View::draw(Frame* frame) {
		if (frame->needsRepaint()||_needsFullRedraw) {
			Color c = adjust(frame->bgColor());
			if (c != Color::None) {
				luaz::canvas::setColor(c.r(), c.g(), c.b());
				luaz::canvas::fillPolygon(frame->points());
			}
			if (_needsFullRedraw) {
				frame->needsRepaint(false);
			}
		}
		drawChildren(frame);
		frame->needsRepaint(false);
	}

	void View::draw(Window* window){
		if (window->needsRepaint()||_needsFullRedraw) {
			Color c = adjust(window->bgColor());
			if (c != Color::None) {
				luaz::canvas::setColor(c.r(), c.g(), c.b());
				luaz::canvas::fillRect(window->x(), window->y(), window->w(), window->h());
			}
			if (_needsFullRedraw) {
				window->needsRepaint(false);
			}
		}
		drawChildren(window);
		window->needsRepaint(false);
	}
	
	void View::draw(BorderedWindow* borderedWindow){
		if (borderedWindow->needsRepaint()||_needsFullRedraw) {
			int x = borderedWindow->x();
			int y = borderedWindow->y();
			int w = borderedWindow->w();
			int h = borderedWindow->h();
			int border = borderedWindow->borderWidth();
			Color c =adjust(borderedWindow->borderColor());
			if (c != Color::None && border != 0) {
				luaz::canvas::setColor(c.r(), c.g(), c.b());
				//	BORDER
				//top
				luaz::canvas::fillRect(x, y, w, border);
				//bottom
				luaz::canvas::fillRect(x, y+h-border, w, border);
				//left
				luaz::canvas::fillRect(x, y, border, h);
				//right
				luaz::canvas::fillRect(x+w-border, y, border, h);
			}
			//	BACKGROUND
			c = adjust(borderedWindow->bgColor());
			if (c != Color::None) {
				luaz::canvas::setColor(c.r(), c.g(), c.b());
				luaz::canvas::fillRect(x + border, y + border, w - border*2, h - border*2);
			}
			if (_needsFullRedraw) {
				borderedWindow->needsRepaint(false);
			}
		}
		drawChildren(borderedWindow);
		borderedWindow->needsRepaint(false);
	}
	
	void View::draw(Label* label){
		if (label->needsRepaint()||_needsFullRedraw) {
			Color c = adjust(label->bgColor());

			if (c != Color::None) {
				luaz::canvas::setColor(c.r(), c.g(), c.b());
				luaz::canvas::fillRect(label->x(), label->y(), label->w(), label->h());
			}
			
			c = adjust(label->color());
			luaz::canvas::setColor(c.r(), c.g(), c.b());
			
			int size = label->fontSize();
			luaz::canvas::font( _fontFace.c_str(), size );
			
			std::string text = label->text();
			int w,h;
			luaz::canvas::textSize(text, w, h);
			std::pair<int,int> xy = align(label->alignment(), label->x(),label->y()+h, w,h, label->w(), label->h(), label->margin());
//			std::pair<int,int> xy = align(label->alignment(), label->x(),label->y(), w,h, label->w(), label->h(), label->margin());

			luaz::canvas::text(xy.first, xy.second, (char*)(text.c_str()));
		}
		label->needsRepaint(false);
	}

	void View::draw(MaskedLabel* label){
		if (label->needsRepaint()||_needsFullRedraw) {
			Color c = adjust(label->bgColor());

			if (c != Color::None) {
				luaz::canvas::setColor(c.r(), c.g(), c.b());
				luaz::canvas::fillRect(label->x(), label->y(), label->w(), label->h());
			}

			c = adjust(label->color());
			luaz::canvas::setColor(c.r(), c.g(), c.b());

			int size = label->fontSize();
			luaz::canvas::font( _fontFace.c_str(), size );

			std::string text = label->text();
			if (text[0] == '_' && text[1] == '_') {
				text.replace(0,2,"_ _");
			}
			int w,h;
			luaz::canvas::textSize(text, w, h);
			std::pair<int,int> xy = align(label->alignment(), label->x(),label->y()+h, w,h, label->w(), label->h(), label->margin());
//			std::pair<int,int> xy = align(label->alignment(), label->x(),label->y(), w,h, label->w(), label->h(), label->margin());

			luaz::canvas::text(xy.first, xy.second, (char*)(text.c_str()));
		}
		label->needsRepaint(false);
	}

	const std::string& View::adjustedPath(Image* img)
	{
		return (_desaturateColors)? img->dimmedPath() :  img->path();;
	}

	void View::draw(Image* image){
		if (image != NULL && !image->path().empty() && (image->needsRepaint()||_needsFullRedraw)) {
			luaz::canvas::drawImage(adjustedPath(image).c_str(), image->x(), image->y(), image->w(), image->h());
		}
		image->needsRepaint(false);
	}
	
	void View::draw(Bar* bar)
	{
		if (bar->needsRepaint()||_needsFullRedraw) {
			Color c = adjust(bar->bgColor());
			luaz::canvas::setColor(c.r(), c.g(), c.b());
			int value = bar->value();
			int minValue = bar->minValue();
			int maxValue = bar->maxValue();
			int p = (100*(value-minValue))/(maxValue-minValue);
			int w = bar->w() * p / 100;
			luaz::canvas::fillRect(bar->x(), bar->y(), w, bar->h());
		}
		bar->needsRepaint(false);
	}

	::canvas::Color View::toLuaColor(const Color& color)
	{
		Color c = adjust(color);
		return ::canvas::Color((util::BYTE)c.r(), (util::BYTE)c.g(), (util::BYTE)c.b());
	}

	void View::draw(List* list)
	{
		if (list->needsRepaint()||_needsFullRedraw) {
			luaz::canvas::font(_fontFace.c_str(), list->fontSize());
			::canvas::Color bgColor           = toLuaColor(Color::Black);
			::canvas::Color cellColor         = toLuaColor(list->cellColor());
			::canvas::Color selectedCellColor = toLuaColor(list->selectedCellColor());
			::canvas::Color textColor         = toLuaColor(list->textColor());
			::canvas::Color selectedTextColor = toLuaColor(list->selectedTextColor());
			::canvas::Color headerColor       = toLuaColor(list->headerColor());
			::canvas::Color textHeaderColor   = toLuaColor(list->textHeaderColor());
			luaz::canvas::drawList(
					list->x(), list->y()
					, list->cellWidths()
					, list->rowHeight()
					, list->rowSep()
					, list->cellHeaders()
					, list->contents()
					, bgColor, cellColor, selectedCellColor, textColor, selectedTextColor, headerColor, textHeaderColor
					, list->current()
					, true // with scrollbar
					, list->from(), list->to(), list->visible()
					, list->scroll.x
					, list->scroll.y
					, list->scroll.w
					, list->scroll.h
					, list->handle.x
					, list->handle.y
					, list->handle.w
					, list->handle.h
					, list->images_normal()
					, list->images_selected());
		}
		list->needsRepaint(false);
	}
	
	void View::fitToSize(int text_w, int text_h, int margin, int &x, int &w, int &h)
	{
		if (text_w == 0) {
			text_w = EMPTY_CC_W;
			text_h = EMPTY_CC_H;
		}
		
		std::pair<int,int> size = luaz::canvas::screenSize();
		x = (size.first  - text_w)/2;
		w = text_w + 2*margin;
		h = text_h;
	}
	
	static void splitBy(const std::string &str, std::vector<std::string> &vector, char criteria) {
		unsigned int pos = 0;
		while (pos < str.size()) {
			std::string tmp;

			while((pos < str.size()) && (str.at(pos) != criteria)){
				tmp.append(str, pos, 1);
				pos++;
			}
			if( pos < str.size() ) {
				pos++; //skip the criteria character

				while((pos < str.size()) && (str.at(pos) == criteria)){
					tmp.append(str, pos, 1);
					pos++;
				}
			}
			vector.push_back(tmp);
		}
	}
	
#define APPENDSPACE(line, words, word) \
	if ( word != words.back() ) { \
		line.append(" "); \
	}
	
	void View::splitLines(TextArea* textArea, std::vector<std::string>& v_lines)
	{
		std::vector<std::string> explicitLines;
		// Split the explicit lines
		{
			splitBy(textArea->text(), explicitLines, '\n');
		}

		int maxSize = textArea->w() - (2*textArea->borderWidth());
		if ((textArea->alignment() & alignment::hLeft) || (textArea->alignment() & alignment::hCenter)) {
			maxSize -= textArea->margin();
		}

		unsigned int maxLines = textArea->lines();
		int fontSize = textArea->fontSize();

		for (std::vector<std::string>::iterator it=explicitLines.begin(); it != explicitLines.end() ; ++it) {
			std::vector<std::string> words;
			std::string line;
			int textWidth, h;

			luaz::canvas::font(_fontFace.c_str(), fontSize);
			splitBy(*it, words, ' ');
			for (std::vector<std::string>::iterator it=words.begin(); it != words.end() && v_lines.size() < maxLines; ++it) {
				std::string word(*it);

				std::string tmp(line);
				tmp.append(word);
				APPENDSPACE(tmp, words, *it);

				luaz::canvas::textSize(tmp, textWidth, h);
				if (maxSize <= textWidth) {
					if ( v_lines.size() + 1 ==  maxLines ) {
						luaz::canvas::textSize("...", textWidth, h);
						if (textWidth <= maxSize) {
							if (line.size()) {
								std::string tmp(line);
								tmp.append("...");
								luaz::canvas::textSize(tmp, textWidth, h);
								if ( textWidth <= maxSize ) {
									line.append("...");
								} else {
									line.replace(line.size() - 3,  3, "...");
								}
							} else  {
								line.assign("...");
							}
						}
					}

					v_lines.push_back( line );

					line.assign(word);
					APPENDSPACE(line, words, *it);
				} else {
					line.assign(tmp);
				}
			}

			if (v_lines.size() < maxLines) {
				v_lines.push_back( line );
			}
			line.clear();
		}
	}
	
	void View::draw(TextArea* textArea) {
		if (textArea->needsRepaint()||_needsFullRedraw) {
			Color c = Color::None;
			int size = textArea->fontSize();
			luaz::canvas::font(_fontFace.c_str(), size);
						
			std::vector<std::string> v_lines;

			splitLines(textArea, v_lines);

			int margin = textArea->margin();
			int leading = textArea->leading();
			int text_h = v_lines.size()*(5+size + leading);
			int alignment = textArea->alignment();
			int x = 0;
			int y = aligny(alignment, textArea->y(), text_h, textArea->h(), textArea->margin());
			std::vector<std::pair<int, int> > positions;
			int maxW = 0;
			
			for (size_t i = 0; i < v_lines.size(); ++i) {
				int h,text_w = 0;
				luaz::canvas::textSize(v_lines[i], text_w, h);
				maxW = std::max(text_w, maxW);
				x = alignx(alignment, textArea->x(), text_w, textArea->w(), margin);
				positions.push_back(std::make_pair(x, y + i*(5+size + leading) + size));
//				positions.push_back(std::make_pair(x, y + i*(5+size + leading)));
			}

			int tx, tw, th;
			if (textArea->fitToText()) {
				fitToSize(maxW, v_lines.size()*(5+size + leading) + 2*margin/*+ size*/, margin, tx, tw, th);
			}else{
				tx = textArea->x();
				tw = textArea->w();
				th = textArea->h();
			}
			
			int borderWidth = textArea->borderWidth();
			if (borderWidth && textArea->borderColor() != Color::None) {
				Color c = adjust(textArea->borderColor());
				luaz::canvas::setColor(c.r(), c.g(), c.b());
				luaz::canvas::fillRect(tx, textArea->y(), tw, th);
			}
			
			c = adjust(textArea->bgColor());
			if (c != Color::None) {
				luaz::canvas::setColor(c.r(), c.g(), c.b());
				luaz::canvas::fillRect(tx + borderWidth, textArea->y() + borderWidth, tw - borderWidth*2, th - borderWidth*2);
			}

			c = adjust(textArea->textColor());
			luaz::canvas::setColor(c.r(), c.g(), c.b());

			for (size_t i = 0; i < v_lines.size(); ++i) {
				std::pair<int, int> point = positions[i];
				luaz::canvas::text(point.first, point.second, (char*)(v_lines[i].c_str()));
			}
		}
		textArea->needsRepaint(false);
	}

	/* No se está usando
	void View::drawActiveTab(PageTab* tab)
	{
		if (tab->needsRepaint()||_needsFullRedraw) {
			int border = tab->borderWidth();

			Color c = adjust(tab->borderColor());
			canvas::setColor(c.r(), c.g(), c.b());
			luaz::canvas::fillRect(tab->x(),tab->y(), tab->w() - border, tab->h());

			c = adjust(tab->bgColor());
			luaz::canvas::setColor(c.r(), c.g(), c.b());
			luaz::canvas::fillRect(tab->x() + border, tab->y() + border, tab->w() - border, tab->h() - border*2);
			if (_needsFullRedraw) {
				tab->needsRepaint(true);
			}
		}
		drawChildren(tab);
		tab->needsRepaint(false);
	}
	*/

	void View::draw(Page* page)
	{
		if (page->isActive()) {
			if (page->needsRepaint()||_needsFullRedraw) {

				Color c = adjust(page->page()->bgColor());
				luaz::canvas::setColor(c.r(), c.g(), c.b());
				page->clear();

				c = adjust(page->page()->borderColor());
				luaz::canvas::setColor(c.r(), c.g(), c.b());
				page->drawBorder();

				if (_needsFullRedraw) {
					page->needsRepaint(true);
				}
			}

			drawChildren(page->page().get());
			drawChildren(page->tab());
		} else { // is inactive
			draw(page->page().get());
			if (page->needsRepaint()||_needsFullRedraw) {
				luaz::canvas::clear(page->tab()->x(), page->tab()->y(), page->tab()->w()+5, page->tab()->h());
			}
			draw(page->tab());
		}
		if (page->helpVisible()) {
			draw(page->helpBox().get());
		} else {
			page->helpBox()->clear(this);
		}
		page->needsRepaint(false);
	}

	void View::draw(TabbedWindow* menu)
	{
		menu->currentPage()->draw(this);
		BOOST_FOREACH(PagePtr page, menu->pages()) {
			if (!page->isActive()) {
				draw(page->tab());
				page->needsRepaint(false);
			}
		}

		drawChildren(menu);
		menu->needsRepaint(false);
	}
	
	void View::draw(ComboBox* combo)
	{
		if (combo->Widget::needsRepaint()||_needsFullRedraw){
			draw(static_cast<Window*>(combo));

			if (combo->isOpen() && combo->maxLines() && combo->maxLines() < (int)combo->items().size()) {
				Color c = adjust(combo->comboBgColor());
				luaz::canvas::setColor(c.r(), c.g(), c.b());
				luaz::canvas::fillRect(combo->scroll.x - combo->scroll.w - combo->scroll.margin, combo->scroll.y, combo->scroll.w, combo->scroll.h - 12);

				c = adjust(combo->selectedColor());
				luaz::canvas::setColor(c.r(), c.g(), c.b());
				luaz::canvas::fillRect(combo->handle.x - combo->scroll.w - combo->scroll.margin, combo->handle.y , combo->handle.w, combo->handle.h);
			}
		}
		combo->needsRepaint(false);
	}
	
	void View::draw(ImageViewer* iv)
	{
		ImagePtr image = iv->currentImage();
		luaz::canvas::drawImageFullScreen(image->path().c_str());
	}

	void View::draw(ScheduleRow* row)
	{
		if (row->needsRepaint() || _needsFullRedraw) {
			luaz::canvas::clear(row->x(), row->y(), row->w()-4, row->h()+2);

			row->title()->y(row->y());
			row->title()->draw(this);
			BOOST_FOREACH(const ScheduleRow::Cell& cell, row->cells()) {
				if (cell._widget->isVisible()) {
					cell._widget->needsRepaint(true);
					cell._widget->y(row->y());
					cell._widget->draw(this);
				}
			}
		}
		row->needsRepaint(false);
	}

	void View::draw(Schedule* schedule)
	{
		if (schedule->needsRepaint() || _needsFullRedraw) {
			Color c = adjust(schedule->bgColor());
			if (c != Color::None) {
				luaz::canvas::setColor(c.r(), c.g(), c.b());
				luaz::canvas::fillRect(schedule->x(), schedule->y(), schedule->w(), schedule->h());
			}
			schedule->header()->draw(this);

			// scrollbar
			c = adjust(schedule->cellColor());
			luaz::canvas::setColor(c.r(), c.g(), c.b());
			luaz::canvas::fillRect(schedule->scroll.x - schedule->scroll.w - schedule->scroll.margin, schedule->scroll.y, schedule->scroll.w, schedule->scroll.h);

			c = adjust(schedule->headerColor());
			luaz::canvas::setColor(c.r(), c.g(), c.b());
			luaz::canvas::fillRect(schedule->handle.x - schedule->scroll.w - schedule->scroll.margin, schedule->handle.y, schedule->handle.w, schedule->handle.h);
		} // if

		for (int i=schedule->from(); i <= schedule->to(); ++i) {
			ScheduleRowPtr row = schedule->getRow(i);
			row->draw(this);
		}

		schedule->needsRepaint(false);
	} // draw

	void View::invalidate(WidgetPtr widget) {
		if (widget != getTopWidget()) {
			_needsFullRedraw = true;
		}
	}

} // wgt
