/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-luaz implementation.

    DTV-luaz is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-luaz is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-luaz.

    DTV-luaz es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-luaz se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

/*
 * schedule.cpp
 *
 *  Created on: Jul 13, 2011
 *      Author: gonzalo
 */

#include "schedule.h"

#include "../view.h"
#include "../modules/wgt.h"

#include <boost/make_shared.hpp>
#include <boost/foreach.hpp>
#include <boost/date_time/date_facet.hpp>

namespace wgt
{

Schedule::Schedule(int x, int y, int w, int h)
	: Window(x,y, w,h, Color::None)
	, _cell_width(90)
	, _title_width(100)
	, _row_sep(4)
	, _row_height(20)
	, _visibleCount(6)
	, _rowCount(0)
	, _visibleSlots(3)
	, _currentSlot(boost::posix_time::time_period(boost::posix_time::ptime(), boost::posix_time::ptime()))

	, _headerColor(Color::Blue)
	, _headerTextColor(Color::White)

	, _cellTextColor(Color::White)
	, _cellColor(Color::DarkBlue)
	, _selectedCellColor(Color::Yellow)
	, _selectedCellTextColor(Color::Black)

	, _cellTitleTextColor(Color::White)
	, _cellTitleColor(Color::Blue)
	, _selectedCellTitleColor(Color::Yellow)
	, _selectedCellTitleTextColor(Color::Black)
{
	_marginsNumber = 2;
	_timeSlots.visibleSlots(_visibleSlots);
}

Schedule::~Schedule()
{}

void Schedule::widths(int title, int slot)
{
	_title_width = title;
	_cell_width  = slot;
}

void Schedule::sep(int sep)
{
	_row_sep = sep;
}

int Schedule::cellWidth()
{
	return _cell_width;
}

int Schedule::titleWidth()
{
	return _title_width;
}

int Schedule::rowSep()
{
	return _row_sep;
}

int Schedule::rowHeight()
{
	return _row_height;
}

int Schedule::rowWidth()
{
	return _title_width + _row_sep + _visibleSlots*(_cell_width + _row_sep);
}

void Schedule::rowHeight(int n)
{
	_row_height = n;
}

void Schedule::showCount(int n)
{
	_visibleCount = n;
}

int Schedule::visibleCount()
{
	return _visibleCount;
}

ScheduleRowPtr Schedule::getRow(size_t ix)
{
	return _rows.at(ix);
}

const ScheduleRowPtr& Schedule::header()
{
	return _header;
}

const std::vector<ScheduleRowPtr>& Schedule::rows()
{
	return _rows;
}

size_t Schedule::rowCount()
{
	return _rows.size();
}

void Schedule::draw(View* view)
{
	view->draw(this);
}

void Schedule::rows(const std::vector<std::string>& titles, int visible)
{
	_visibleCount = visible;
	BOOST_FOREACH(std::string title, titles) {
		ScheduleRowPtr row = boost::make_shared<ScheduleRow>(this, 0,0);
		row->title(title);
		row->cellColors(false,  _cellTextColor, _cellColor);
		row->cellColors(true,   _selectedCellTextColor, _selectedCellColor);
		row->titleColors(false, _cellTitleTextColor, _cellTitleColor);
		row->titleColors(true,  _selectedCellTitleTextColor, _selectedCellTitleColor);
		_rows.push_back(row);
		addChild(row, true);
	}
	for (int i=_rows.size(); i<_visibleCount; ++i) {
		ScheduleRowPtr row = boost::make_shared<ScheduleRow>(this, 0,0);
		row->title("");
		row->cellColors(false,  _cellTextColor, _cellColor);
		row->cellColors(true,   _selectedCellTextColor, _selectedCellColor);
		row->titleColors(false, _cellTitleTextColor, _cellTitleColor);
		row->titleColors(true,  _selectedCellTitleTextColor, _selectedCellTitleColor);
		_rows.push_back(row);
		addChild(row, false);
	}
	_rowCount = _rows.size();

	initFromTo(_activeChild, _visibleCount, _rowCount);
	initScrollbar(this, _visibleCount, _rowCount, _row_height, _row_sep, 2 + _row_height + _row_sep);

	updateFromTo(   _activeChild, _visibleCount, _rowCount);
	updateScrollbar(_activeChild, _visibleCount, _rowCount);
	int r_y = rowSep() + rowHeight();
	for (int i=from(); i <= to(); ++i) {
		ScheduleRowPtr row = getRow(i);
		row->y(r_y);
		row->arrangeLabels();

		r_y += rowSep() + rowHeight();
	}

}

void Schedule::init(const std::vector<TimeSlots::time_slot>& slots, const std::vector<std::string>& headers)
{
	_timeSlots.setSlots(slots);

	removeChildren();

	_rows.clear();

	_header.reset();
	_header= boost::make_shared<ScheduleRow>(this, 0,0);
	_header->title("");
	_header->cellColors(true,  _headerTextColor, _headerColor);
	_header->cellColors(false, _headerTextColor, _headerColor);
	for (size_t i=0; i<headers.size(); ++i) {
		_header->addCell(slots[i], headers[i], -1);
	}
	_header->activate(false);
	_header->moveLabels();

	addChild(_header, false);
	updateDate();
}

void Schedule::init(const boost::posix_time::ptime& currentTime, int periodLength, int slotCount)
{
	std::vector<TimeSlots::time_slot> slots;
	std::vector<std::string> headers;
	int m = (currentTime.time_of_day().minutes()/periodLength) * periodLength;
	boost::posix_time::ptime initialTime(currentTime.date(), boost::posix_time::time_duration(currentTime.time_of_day().hours(), m, 0,0));
	boost::posix_time::time_duration halfhour = boost::posix_time::minutes(periodLength);

	for(int i=0;i<slotCount;++i) {
		std::ostringstream os;
		os << std::setw(2) << std::setfill('0') << initialTime.time_of_day().hours()
		   << ":"
		   << std::setw(2) << std::setfill('0') << initialTime.time_of_day().minutes();

		headers.push_back(os.str());

		boost::posix_time::time_period range(initialTime, halfhour + boost::posix_time::seconds(1));
		slots.push_back(range);
		initialTime += halfhour;
	}

	init(slots, headers);
}

TimeSlots& Schedule::timeSlots()
{
	return _timeSlots;
}

void Schedule::updateDate()
{
	boost::gregorian::date::ymd_type first(_timeSlots.visibleFrom().date().year_month_day());
	std::stringstream str;
	str << first.day << "/" << int(first.month) << "/" << first.year;
	_header->title()->text(str.str());
}

void Schedule::shiftLeft()
{
	if (_timeSlots.shiftLeft()) {
		_header->synch();
		updateDate();
		BOOST_FOREACH(ScheduleRowPtr row, _rows) {
			row->arrangeLabels();
		}
		needsRepaint(true);
	}
}

void Schedule::shiftRight()
{
	if (_timeSlots.shiftRight()) {
		_header->synch();
		updateDate();
		BOOST_FOREACH(ScheduleRowPtr row, _rows) {
			row->arrangeLabels();
		}
		needsRepaint(true);
	}
}

void Schedule::moveSlotsTo(boost::posix_time::ptime time)
{
	_timeSlots.moveSlotsTo(time);
	_header->synch();
	BOOST_FOREACH(ScheduleRowPtr row, _rows) {
		row->synch();
	}
}

void Schedule::setShowSelected(const std::string& showCallback)
{
	_lshowselected = showCallback;
}

void Schedule::setChannelSelected(const std::string& channelCallback)
{
	_lchannelselected = channelCallback;
}

void Schedule::setMoreShows(const std::string& moreShowsCallback)
{
	_lmoreshows = moreShowsCallback;
}

void Schedule::showSelected(int n)
{
	if (!_lshowselected.empty()) {
		Wgt::call(_lshowselected, n);
	}
}

void Schedule::channelSelected(int n)
{
	if (!_lchannelselected.empty()) {
		Wgt::call(_lchannelselected, n);
	}
}

void Schedule::needMoreShows()
{
	if (!_lmoreshows.empty()) {
		Wgt::call(_lmoreshows, _activeChild);
	}
}

void Schedule::setSelected(int row)
{
	//std::cout << "[lua] Schedule::setSelected()" << std::endl;
	const size_t oldChild = _activeChild;

	selectChild(row);
	if (oldChild != _activeChild) {
		selectedRowChanged(from());
	}
}

void Schedule::selectedRowChanged(int oldFrom)
{
	updateFromTo(   _activeChild, _visibleCount, _rowCount);
	updateScrollbar(_activeChild, _visibleCount, _rowCount);

	if (from() != oldFrom) {
		needMoreShows();
	}

	channelSelected(_activeChild);
	showSelected(_rows[_activeChild]->currentCellId());

	int r_y = y() + rowSep() + rowHeight();
	for (int i=from(); i <= to(); ++i) {
		ScheduleRowPtr row = getRow(i);
		row->y(r_y);
		row->arrangeLabels();

		r_y += rowSep() + rowHeight();
	}

	needsRepaint(true);
}

bool Schedule::onUpArrow()
{
	const size_t oldChild = _activeChild;

	selectPreviousChild(false);
	if (oldChild != _activeChild) {
		selectedRowChanged(from());
	}

	return true;
}

bool Schedule::onDownArrow()
{
	const size_t oldChild = _activeChild;

	selectNextChild(false);
	if (oldChild != _activeChild) {
		selectedRowChanged(from());
	}

	return true;
}

void Schedule::headerColors(const Color& color, const Color& textColor)
{
	_headerColor = color;
	_headerTextColor = textColor;
	if (_header) {
		_header->titleColors(false, textColor, color);
		_header->titleColors(true,  textColor, color);
		_header->cellColors(false,  textColor, color);
		_header->cellColors(true,   textColor, color);
	}
}

void Schedule::cellColors(bool selected, const Color& color, const Color& textColor)
{
	if (selected) {
		_selectedCellColor = color;
		_selectedCellTextColor = textColor;
		BOOST_FOREACH(ScheduleRowPtr row, _rows) {
			row->cellColors(true, _selectedCellTextColor, _selectedCellColor);
		}
	} else {
		_cellColor = color;
		_cellTextColor = textColor;
		BOOST_FOREACH(ScheduleRowPtr row, _rows) {
			row->cellColors(false, _cellTextColor, _cellColor);
		}
	}
}

void Schedule::rowTitleColors(bool selected, const Color& color, const Color& textColor)
{
	if (selected) {
		_selectedCellTitleTextColor = textColor;
		_selectedCellTitleColor     = color;
		BOOST_FOREACH(ScheduleRowPtr row, _rows) {
			row->titleColors(true, textColor, color);
		}
	} else {
		_cellTitleTextColor = textColor;
		_cellTitleColor     = color;
		BOOST_FOREACH(ScheduleRowPtr row, _rows) {
			row->titleColors(false, textColor, color);
		}
	}
}

Color& Schedule::headerColor()
{
	return _headerColor;
}

Color& Schedule::cellColor()
{
	return _cellColor;
}

Color& Schedule::selectedCellColor()
{
	return _selectedCellColor;
}

void Schedule::dumpShows()
{
	BOOST_FOREACH(ScheduleRowPtr row, _rows) {
		if (!row->title()->text().empty()) {
			std::cout << "[lua] # channel: " << row->title()->text() << std::endl;
			row->dumpShows();
			std::cout << "[lua] # " << std::endl;
		}
	}
}

const TimeSlots::time_slot& Schedule::currentSlot()
{
	return _currentSlot;
}

void Schedule::currentSlot(const TimeSlots::time_slot& slot)
{
	_currentSlot = _timeSlots.visibleRange().intersection(slot);
}

void Schedule::resetSlots()
{
	_timeSlots.reset();
	_currentSlot = _timeSlots.visibleRange();
	updateDate();
}

} //wgt
