/*******************************************************************************

  Copyright (C) 2010, 2011 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-connector implementation.

    DTV-connector is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-connector is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-connector.

    DTV-connector es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-connector se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/
#include "pesdemuxer.h"
#include "../text.h"
#include <util/buffer.h>
#include <util/log.h>

#define MAX_BUFFER_SIZE    1024
#define SET_STATE_OFF(st,off)	\
	_dataOff = off; _state = st;
#define SET_STATE(st) SET_STATE_OFF( st, 0 )
#define SET_STATE_SYN(st) \
	_sync = 0xFFFFFFFF; \
	SET_STATE_OFF( st, 0 )

namespace tuner {

namespace impl {

static QWORD getValue( BYTE *ptr ) {
	QWORD value;

	value = (ptr[0] & 0x0e);
	value <<= 7;

	value |= ptr[1];
	value <<= 7;
	value |= BYTE(ptr[2] >> 1);
	value <<= 8;

	value |= ptr[3];
	value <<= 7;
	value |= BYTE(ptr[4] >> 1);

	return value;
}

}

PESDemuxer::PESDemuxer( pes::type pesType )
{
	_pesType = pesType;
	_data = (BYTE *)malloc( MAX_BUFFER_SIZE );
	SET_STATE_SYN(dmxSync);
	_dataOff = 0;
}

PESDemuxer::~PESDemuxer( void )
{
	free(_data);
}

//	Getters
pes::type PESDemuxer::type() const {
	return _pesType;
}

bool PESDemuxer::fillData( int minSize, BYTE *data, int &dataOff, BYTE *ptr, size_t len, size_t &ptrOff ) {
	assert( minSize >= dataOff );
	assert( ptrOff <= len );

	//	How many bytes we need to complete header?
	size_t copy = minSize-dataOff;
	if (copy > (len-ptrOff)) {
		copy = (len-ptrOff);
	}
	if (copy) {	//	Copy bytes
		if (data) {	//	Skip data only?
			memcpy( data+dataOff, (ptr+ptrOff), copy );
		}
		dataOff += copy;
		ptrOff += copy;
	}

	//	Data is complete?
	if (dataOff < minSize) {
		//	Wait for more data
		return false;
	}

	_payloadSize += minSize;
	return true;
}

bool PESDemuxer::fillData( int minSize, bool needFill, BYTE *ptr, size_t len, size_t &ptrOff ) {
	return fillData( minSize, needFill ? _data : NULL, _dataOff, ptr, len, ptrOff );
}

bool PESDemuxer::fillPayload( BYTE *ptr, size_t len, size_t &ptrOff ) {
	assert( _packetLen < MAX_BUFFER_SIZE );
	return fillData( _packetLen-_payloadSize, true, ptr, len, ptrOff );
}

//  Parser operations
void PESDemuxer::parse( BYTE *ptr, size_t len ) {
	size_t off=0;

	while (off < len) {
		LDEBUG( "PESDemuxer", "Parse: state=%d, len=%ld, off=%ld, bytes=%ld",
			_state, len, off, len-off );

		if (_state == dmxSync) {
			while (off < len && ((_sync & 0x00FFFFFF) != 0x00000001)) {
				_sync <<= 8;
				_sync |= *(ptr+off);
				off++;
			}

			if (off < len) {
				LDEBUG( "PESDemuxer", "Find sync: sync=%08x, off=%ld", _sync, off );
				//	Reset header and change state
				SET_STATE_OFF( dmxHeader, 3 );
			}
		}
		else if (_state == dmxHeader) {
			//	Copy header
			if (!fillData( PES_BASIC_HEADER, true, ptr, len, off )) {
				return;
			}

			//	Setup
			_payloadSize = 0;
			_packetLen = PES_PACKET_LEN(_data);
			_pts = PES_NO_PTS;
			_dts = PES_NO_PTS;

			//	Check demux streamID
			_sID = PES_STREAM_ID(_data);
			if (!checkSID( _sID )) {
				//	Skip all packet
				LWARN( "PESDemuxer", "invalid stream ID: packetStreamID=%02x", _sID );
				SET_STATE( dmxSkip );
			}
			else if (!checkLen( PES_LEN(_data) )) {
				//	Check packet len
				LWARN( "PESDemuxer", "invalid pes len" );
				SET_STATE( dmxSkip );
			}
			else if (_sID != PES_SID_MAP &&
				_sID != PES_SID_PADDING &&
				_sID != PES_SID_PRIVATE_2 &&
				_sID != PES_SID_ECM &&
				_sID != PES_SID_EMM &&
				_sID != PES_SID_DIRECTORY &&
				_sID != PES_SID_DSMCC &&
				_sID != PES_SID_ITU_TYPE_E)
			{
				SET_STATE( dmxFields );
			}
			else if (_sID == PES_SID_PADDING) {
				SET_STATE( dmxSkip );
			}
			else {
				SET_STATE( dmxPayload );
			}
		}
		else if (_state == dmxFields) {
			//	Fill PES fields
			if (!fillData( PES_FIELDS_HEADER, _fields, _dataOff, ptr, len, off )) {
				return;
			}

			LDEBUG( "PESDemuxer", "Fields header: first=%02x, second=%02x, len=%02x", _fields[0], _fields[1], PES_FIELDS_LEN(_fields) );
			SET_STATE( dmxFieldsPayload );
		}
		else if (_state == dmxFieldsPayload) {
			//	Fill fields payload
			if (!fillData( PES_FIELDS_LEN(_fields), true, ptr, len, off )) {
				return;
			}

			//	GET PTS
			if (_fields[1] & PES_PTS_FLAG) {
				_pts=impl::getValue(_data);
			}

			//	Get DTS
			if (_fields[1] & PES_DTS_FLAG) {
				_dts=impl::getValue(_data+5);
			}

			LDEBUG( "PESDemuxer", "PTS=%llx, DTS=%llx", _pts, _dts );

			//	PES have private stream header?
			if (_sID == PES_SID_PRIVATE_1) {
				SET_STATE( dmxSyncHeader );
			}
			else {
				SET_STATE( dmxPayload );
			}
		}
		else if (_state == dmxSyncHeader) {
			//	Fill private stream header
			if (!fillData( 3, true, ptr, len, off )) {
				return;
			}

			//	Parse
			size_t syncOff=0;
			BYTE dataID=RB(_data,syncOff);
			BYTE privateID=RB(_data,syncOff);
			_syncHeaderLen = RB(_data,syncOff);
			_syncHeaderLen &= 0x0F;

			LDEBUG( "PESDemuxer", "Sync header: dataID=%02x, privStreamID=%02x, headerLen=%d",
				dataID, privateID, _syncHeaderLen );

			if (!checkSynHeader( dataID, privateID ))  {
				SET_STATE( dmxSkip );
			}
			else if (_syncHeaderLen) {
				SET_STATE( dmxSyncHeaderData );
			}
			else {
				SET_STATE( dmxPayload );
			}
		}
		else if (_state == dmxSyncHeaderData) {
			//	Fill private stream header
			if (!fillData( _syncHeaderLen, false, ptr, len, off )) {
				return;
			}

			SET_STATE( dmxPayload );
		}
		else if (_state == dmxPayload) {
			int state=parsePayload( ptr, len, off );
			if (state != _state) {
				if (state == dmxSync) {
					SET_STATE_SYN( state );
				} else {
					SET_STATE( state );
				}
			}
		}
		else if (_state == dmxSkip) {
			if (!_packetLen || fillData( _packetLen, NULL, _payloadSize, ptr, len, off )) {
				//	Skip all packet
				LDEBUG( "PESDemuxer", "Payload complete: packetLen=%d, bytes=%ld", _packetLen, len-off );
				SET_STATE_SYN( dmxSync );
			}
		}
		else {
			assert(false);
			return;
		}
	}
}

void PESDemuxer::parse( BYTE * /*tsPayload*/, size_t /*len*/, bool /*start*/ ) {
}

int PESDemuxer::parsePayload( BYTE * /*ptr*/, size_t /*len*/, size_t & /*off*/ ) {
	return dmxSkip;
}

//	Getters
bool PESDemuxer::checkSID( BYTE sID ) {
	if (_pesType == pes::audio) {
		return (sID & 0xF0) == 0xC0;
	}
	else if (_pesType == pes::video) {
		return (sID & 0xF0) == 0xE0;
	}
	else {
		return true;
	}
}

bool PESDemuxer::checkLen( WORD /*len*/ ) const {
	return true;
}

bool PESDemuxer::checkSynHeader( BYTE /*dataID*/, BYTE /*privateID*/ ) const {
	return true;
}

QWORD PESDemuxer::pts() const {
	return _pts;
}

QWORD PESDemuxer::dts() const {
	return _dts;
}

BYTE *PESDemuxer::data() const {
	return _data;
}

int PESDemuxer::dataOff() const {
	return _dataOff;
}

}
