/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-mpegparser implementation.

    DTV-mpegparser is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-mpegparser is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-mpegparser.

    DTV-mpegparser es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-mpegparser se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "frontend.h"
#include "provider.h"
#include <util/log.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <stropts.h>
#include <linux/dvb/dmx.h>
#include <linux/dvb/version.h>


#define MAKE_PROP(name, cmdName)	\
	struct dtv_property name; \
	name.cmd = cmdName;

#define MAKE_PROP1(name, cmdName, cmdValue)	  \
	struct dtv_property name; \
	name.cmd = cmdName; \
	name.u.data = cmdValue;	

namespace tuner {
namespace dvb {

Frontend::Frontend( const std::string &frontend )
	: _frontend(frontend)
{
	_frontendFD = -1;
	_spectral_inversion = INVERSION_AUTO;
}

Frontend::~Frontend()
{
}

bool Frontend::initialize() {
	struct dvb_frontend_info fe_info;

	//	Open frontend
	if ((_frontendFD = open( _frontend.c_str(), O_RDWR )) < 0) {
		LERROR( "Frontend", "Error, failed to open '%s': %d", _frontend.c_str(), errno );
		return false;
	}
	
	//	Determine FE type and caps
	if (ioctl(_frontendFD, FE_GET_INFO, &fe_info) == -1) {
		LERROR( "Frontend", "Error, FE_GET_INFO failed: %d", errno );
		return false;
	}
	_spectral_inversion = (fe_info.caps & FE_CAN_INVERSION_AUTO) ? INVERSION_AUTO : INVERSION_OFF;	

	LDEBUG( "Frontend", "Frontend info:" );
	LDEBUG( "Frontend", "\t name=%s, type=%x", fe_info.name, fe_info.type );
	LDEBUG( "Frontend", "\t frequency min=%d, max=%d, step=%d, tolerance=%d",
		fe_info.frequency_min, fe_info.frequency_max, fe_info.frequency_stepsize, fe_info.frequency_tolerance );
	LDEBUG( "Frontend", "\t symbol rate min=%d, max=%d, tolerance=%d",
		fe_info.symbol_rate_min, fe_info.symbol_rate_max, fe_info.symbol_rate_tolerance );
	LDEBUG( "Frontend", "\t notifier delay=%d", fe_info.notifier_delay );
	LDEBUG( "Frontend", "\t caps=%08X", fe_info.caps );

	return true;
}

void Frontend::finalize() {
	close( _frontendFD );
}

bool Frontend::setFrequency( int freq, int khBandwidth ) {
	LDEBUG( "Frontend", "Tune started begin: freq=%d, bandwidth=%d", freq, khBandwidth );

#if DVB_API_VERSION == 5 && DVB_API_VERSION_MINOR <= 5
	struct dvb_frontend_parameters param;	
	
	memset( &param, 0, sizeof(dvb_frontend_parameters) );
	param.frequency                    = freq * 1000;
	param.inversion                    = _spectral_inversion;
	param.u.ofdm.code_rate_HP          = FEC_3_4;
	param.u.ofdm.code_rate_LP          = FEC_AUTO;
	param.u.ofdm.constellation         = QAM_AUTO;
	param.u.ofdm.transmission_mode     = TRANSMISSION_MODE_AUTO;
	param.u.ofdm.guard_interval        = GUARD_INTERVAL_AUTO;
	param.u.ofdm.hierarchy_information = HIERARCHY_NONE;

	switch (khBandwidth) {
		case 6000:
			param.u.ofdm.bandwidth = BANDWIDTH_6_MHZ;
			break;
		case 7000:
			param.u.ofdm.bandwidth = BANDWIDTH_7_MHZ;
			break;
		case 8000:
			param.u.ofdm.bandwidth = BANDWIDTH_8_MHZ;
			break;			
		default:
			param.u.ofdm.bandwidth = BANDWIDTH_AUTO;
			break;
	}

	//	Tune!
	if (ioctl( _frontendFD, FE_SET_FRONTEND, &param ) == -1) {
		LERROR("Frontend", "Setting frontend parameters failed");
		return false;
	}

	return true;

#else
	struct dtv_properties props = { 0, NULL };

	MAKE_PROP(clearCmd,DTV_CLEAR);
	MAKE_PROP1(deliveryCmd,DTV_DELIVERY_SYSTEM,SYS_DVBT);
	MAKE_PROP1(freqCmd,DTV_FREQUENCY,freq * 1000);
	MAKE_PROP1(modCmd,DTV_MODULATION,QAM_AUTO);
	MAKE_PROP1(invCmd,DTV_INVERSION,INVERSION_AUTO);
	MAKE_PROP(tuneCmd,DTV_TUNE);
	
	struct dtv_property params[] = {
		clearCmd,
		deliveryCmd,
		freqCmd,
		modCmd,
		invCmd,
		tuneCmd
	};

	props.num = sizeof(params)/sizeof(struct dtv_property);
	props.props = params;
	if (ioctl( _frontendFD, FE_SET_PROPERTY, &props ) != 0 ) {
		LERROR( "Frontend", "Cannot set property" );
		return false;
	}
	
	return true;
#endif
}

//	Status/signal
bool Frontend::getStatus( state::type &newState, signal::type &newSignal ) const {
	fe_status_t st;
	uint32_t ber;
	int16_t value;
	util::WORD /*snr, */signal;

	//	Get status
	if (ioctl(_frontendFD, FE_READ_STATUS, &st) == -1) {
		LERROR( "Frontend", "FE_READ_STATUS failed" );
		return false;
	}

	//	Get Signal-to-Noise ratio
	// if (ioctl(_frontendFD, FE_READ_SNR, &value) == -1) {
	// 	LERROR( ""Frontend", "FE_READ_SNR failed" );
	// 	return false;
	// }
	// snr = (value & 0x0000FFFF);

	//	Get Signal Strength
	if (ioctl(_frontendFD, FE_READ_SIGNAL_STRENGTH, &value) == -1) {
		LERROR( "Frontend", "FE_READ_SIGNAL_STRENGTH failed" );
		return false;
	}
	signal = (value & 0x0000FFFF);

	//	Get Bit Error Rate
	if (ioctl(_frontendFD, FE_READ_BER, &ber) == -1) {
		LERROR( "Frontend", "FE_READ_BER failed" );
		return false;
	}

	newState = (st & FE_HAS_LOCK && st & FE_HAS_CARRIER) ? state::tuned : state::tunning;
	if (newState == state::tuned) {
		// LDEBUG( "Frontend", "st=%04x, snr=%04x, signal=%04x, ber=%08x",
		// 	st, snr, signal, ber );
		if (signal <= 0x7700) {
			newSignal = signal::poor;
		}
		else if (signal >= 0x9900) {
			newSignal = signal::excelent;
		}
		else {
			newSignal = signal::good;
		}
	}

	return true;
}

}
}
