/*******************************************************************************

  Copyright (C) 2010, 2011 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-connector implementation.

    DTV-connector is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-connector is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-connector.

    DTV-connector es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-connector se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "provider.h"
#include "frontend.h"
#include "timer.h"
#include "ts/provider.h"
#include "generated/config.h"
#if DVB_USE_DVB
#include	"dvb/provider.h"
#endif
#include "filter.h"
#include "../demuxer/psi/psidemuxer.h"
#include "../demuxer/psi/psi.h"
#include "../demuxer/psi/cache.h"
#include "../demuxer/pes/pesdemuxer.h"
#include <util/task/dispatcher.h>
#include <util/cfg/configregistrator.h>
#include <util/pool/circularpool.h>
#include <util/buffer.h>
#include <util/log.h>
#include <util/functions.h>
#include <util/mcr.h>
#include <boost/bind.hpp>
#include <boost/foreach.hpp>

namespace tuner {

REGISTER_INIT_CONFIG( tuner ) {
	root().addNode( "tuner" ).addNode( "provider" )
		.addValue( "cache", "Use PSI cache", false )
#if DVB_USE_DVB
		.addValue( "use", "Provider to initialize", std::string("dvb") );
#else
		.addValue( "use", "Provider to initialize", std::string("tsdata") );
#endif
}

template<typename T>
struct FilterFinder {
	ID _pid;
	FilterFinder( ID pid ) : _pid( pid ) {}
	bool operator()( const T filter ) const {
		return filter->pid() == _pid;
	}
};

template<typename T>
	inline typename std::list<T>::iterator findByPid( std::list<T> &list, ID pid ) {
	return std::find_if( list.begin(), list.end(), FilterFinder<T>(pid) );
}

template<typename T>
struct RemoveSections {
	RemoveSections( ID pid ) : _pid(pid) {}
	bool operator()( const T &item ) const {
		return _pid == TS_PID_NULL || item.pid == _pid;
	}

	ID _pid;
};

Provider* Provider::create() {
	Provider *provider=NULL;

	const std::string &use = util::cfg::getValue<std::string>("tuner.provider.use");
	if (use == "tsdata") {
		provider = new ts::Provider();
	}
#if DVB_USE_DVB
	else if (use == "dvb") {
		provider = new dvb::Provider();
	}
#endif
	else {
		return NULL;
	}

	return provider;
}

Provider::Provider()
{
	_frontend   = NULL;
	_current    = size_t(-1);
	_state      = state::off;
	_signal     = signal::invalid;
	_statusPoll = 500;
	_pool       = NULL;
	_cache      = NULL;
	_disp       = NULL;
}

Provider::~Provider()
{
	assert(_state < state::idle);
}

bool Provider::start( DWORD networkBlocks/*=1000*/ ) {
	LDEBUG( "Provider", "Start" );

	state( state::fail );

	//	Register into dispatcher
	if (!_disp) {
		LERROR( "Provider", "Dispatcher empty" );
		return false;
	}

	//	Create network resource pool
	_pool = new SectionsPool( "provider", networkBlocks, TSS_PRI_MAX_BYTES );

	//	Create sections cache
	if (util::cfg::getValue<bool>("tuner.provider.cache")) {
		_cache = new psi::Cache( _pool );
		if (!_cache->load()) {
			LERROR( "Provider", "Cannot start engine filter" );
			return false;
		}
	}

	//	Create and initialize frontend
	_frontend = createFrontend();
	if (!_frontend || !_frontend->initialize()) {
		LERROR( "Provider", "Cannot start tuner" );
		return false;
	}

	//	Start filter engine
	if (!startEngineFilter()) {
		LERROR( "Provider", "Cannot start engine filter" );
		return false;
	}
	
	//	Start demuxer thread
	_thread = boost::thread( boost::bind( &Provider::providerThread, this ) );

	state( state::idle );		
	
	return true;
}

void Provider::stop() {
	if (state() >= state::idle) {
		LDEBUG( "Provider", "Stop" );
		
		//	Unregister in dispatcher
		_disp->unregisterTarget( this );

		//	Stop current network
		stopCurrent();
	
		//	Wait section demuxer thread
		_pool->notify_all( util::pool::mask::exit );
		_thread.join();
		LDEBUG( "Provider", "Join to thread successfully" );

		//	Stop lost filters
		_mFilters.lock();
		BOOST_FOREACH( Filter *filter, _filters ) {
			LWARN( "Provider", "Filter lost: pid=%04x", filter->pid() );
			stopFilter( filter, true );
		}
		_filters.clear();
		_mFilters.unlock();

		//	Clear timers
		_mTimers.lock();
		CLEAN_ALL( impl::Timer *, _timers );
		_mTimers.unlock();
	
		//	Stop filter engine
		stopEngineFilter();

		//	Stop tuner
		_frontend->finalize();
		DEL(_frontend);

		//	Store cache
		if (_cache) {
			_cache->unload();
			DEL(_cache)
		}

		DEL(_pool);
	}

	//	Set state always to off
	state( state::off );	
}

//	State
void Provider::statusPollTimeout( int ms ) {
	_statusPoll = ms;
}

int Provider::statusPollTimeout() const {
	return _statusPoll;
}

state::type Provider::state() {
	state::type res;

	_mutex.lock();
	res = _state;
	_mutex.unlock();
	
	return res;
}

void Provider::state( state::type newState, signal::type newSignal/*=signal::invalid*/ ) {
	bool changed=false;
	signal::type oldSignal;
	assert( newState >= 0 && newState <= state::tuned );
	
	_mutex.lock();

	//	Save state
	state::type old=_state;
	_state = newState;

	//	Save signal
	oldSignal=_signal;

	//	Only change signal if state is tuned
	if (newState == state::tuned) {
		_signal = newSignal;
	}
	else {
		newSignal = _signal = signal::invalid;
	}
	_mutex.unlock();
	
	//	Process change
	if (old != newState) {
		LDEBUG( "Provider", "Changing state: old=%02x, new=%02x", old, newState );

		//	idle -> [tunning|tuned] : poll status
		if (old <= state::idle && newState > state::scanning) {
			addTimer( TS_PID_NULL, statusPollTimeout() );
		}
		else if (old > state::scanning && newState <= state::idle) {
			rmTimer( TS_PID_NULL );
		}
		
		changed = true;
	}

	if (oldSignal != newSignal) {
		changed=true;
	}

	if (changed && !_onStatusChanged.empty()) {
		_onStatusChanged( newState, newSignal );
	}
}

void Provider::onStatusChanged( const OnStatusChanged &callback ) {
	_onStatusChanged = callback;
}

void Provider::checkStatus() {
	//LDEBUG( "Provider", "Check status begin" );
	
	//      Only called when state > idle
	state::type st;
	signal::type signal;
	if (_frontend->getStatus( st, signal )) {
		state( st, signal );
	}

	//LDEBUG( "Provider", "Check status end" );	
}

//	Filter methods
bool Provider::startFilter( PSIDemuxer *demux ) {
 	assert(demux);

	LDEBUG( "Provider", "Start section filter: demux=%p, pid=%04x, timeout=%d", demux, demux->pid(), demux->timeout() );

	//	Check if can alloc a new filter
	bool result = canAllocFilter( true, demux->pid() );
	if (result) {
		//	Create filter
		Filter *filter = createFilter( demux );
		if (filter) {
			result=startFilter( filter );
			if (result) {
				//	Setup demuxer
				demux->setDispatcher( boost::bind(&Provider::onNotification,this,_1,_2) );
				if (_cache) {
					demux->setCache( _cache );
				}

				_mDemuxers.lock();

				//	Add demux to list of demuxers
				_demuxers.push_back( demux );

				//	Try push PSI section from cache
				if (_cache && demux->supportCache()) {
					util::Buffer *buf= _cache->get( demux->pid() );
					if (buf) {
						enqueue( demux->pid(), buf );
					}
				}
				
				_mDemuxers.unlock();					
			}
		}
		else {
			LWARN( "Provider", "Cannot create filter: pid=%04x", demux->pid() );
			result=false;
		}
	}

	if (!result) {
		delete demux;
	}
	
	return result;
}

bool Provider::startFilter( ID pid, pes::type pesType, stream::Pipe *pipe ) {
	LDEBUG( "Provider", "Start stream filter: pid=%04x, pesType=%d, pipe=%p",
		pid, pesType, pipe );

	//	Check if can alloc a new filter
	bool result = canAllocFilter( false, pid );
	if (result) {
		//	Create filter
		Filter *filter = createFilter( pid, pesType, pipe );
		if (filter) {
			result=startFilter( filter );
		}
		else {
			LWARN( "Provider", "Cannot create filter: pid=%04x", pid );
			result=false;
		}
	}
	
	return result;
}

void Provider::stopFilter( ID pid, bool destroy/*=true*/ ) {
	if (pid != TS_PID_NULL) {
		//	Try remove from list
		Filter *filter=removeFilter( pid );
		if (filter) {
			LDEBUG( "Provider", "Stop filter: pid=%04x", pid );
			stopFilter( filter, destroy );
		}
	}
}

int Provider::maxFilters() const {
	return 100;
}

bool Provider::haveTimers() const {
	return true;
}

//	Aux filters engine
bool Provider::startEngineFilter() {
	LDEBUG( "Provider", "Start engine filter" );
	return true;
}

void Provider::stopEngineFilter() {
	LDEBUG( "Provider", "Stop engine filter" );
}

bool Provider::checkCRC() const {
	return true;
}

//	Network methods
size_t Provider::getNetworksCount() const {
	return _frontend->getCount();
}

size_t Provider::currentNetwork() const {
	return _current;
}

std::string Provider::getNetworkName( size_t index ) const {
	return _frontend->getName( index );
}

size_t Provider::findNetwork( const std::string &name ) const {
	return _frontend->find( name );
}

bool Provider::setNetwork( size_t index ) {
	bool result=false;
	
	if (state() >= state::idle) {
		state::type st = tuner::state::idle;		
		LINFO( "Provider", "Set current: %s", getNetworkName(index).c_str() );
	
		//	Stop current network
		stopCurrent();

		//	Tune network
		result=_frontend->start( index );
		if (result) {
			//	Initialize cache
			if (_cache) {
				_cache->setNetwork( index );
			}
			
			//	Check state
			st = tuner::state::tunning;

			_current = index;
   		}
		//	Set state
		state( st );
   	}
	
	return result;
}

//	Scan networks
bool Provider::scan( size_t index ) {
	//	Stop current network
	stopCurrent();

	//	Clear cache
	if (_cache) {
		_cache->clear( index );
	}

	//	Tune the first scanned network
	bool result=_frontend->start( index );
	if (result) {
		//	Set state
		state( tuner::state::scanning );

		_current = index;
	}
	return result;
}

void Provider::cancelScan() {
	stopCurrent();
}

//	Stop current network
void Provider::stopNetwork() {
}

void Provider::stopCurrent() {
	if (state() > tuner::state::idle) {
		LDEBUG( "Provider", "Stop current network" );

		//	Set state to idle (no accept more sections or tasks)
		state( tuner::state::idle );

		//	Stop frontend
		_frontend->stop();

		//	Stop network
		stopNetwork();

		//	Clean all data of sections
		clear( TS_PID_NULL );

		_current = size_t(-1);
	}
}

//	Aux filters
bool Provider::startFilter( Filter *filter ) {
	bool result=true;
	
	//	Initialize filter
	if (!filter->initialized()) {
		result=filter->initialize();
	}

	if (result) {
		//	Start filter
		result=filter->start();
		if (result) {
			//	Add to list of filters
			_mFilters.lock();
			_filters.push_back( filter );
			_mFilters.unlock();

			//	Add timer if filter not support it!
			if (!haveTimers() && filter->timeout()) {
				addTimer( filter->pid(), filter->timeout() );
			}
		}
		else {
			filter->deinitialize();
			LWARN( "Provider", "Filter cannot be started: pid=%04x", filter->pid() );
		}
	}
	else {
		LWARN( "Provider", "Filter initialization fail: filter=%04x", filter->pid() );
	}

	return result;
}

void Provider::stopFilter( Filter *filter, bool destroy ) {
	assert(filter);

	ID pid = filter->pid();
	
	//	Stop filter
	filter->stop();
	
	//	Remove demuxer
	_mDemuxers.lock();
	DemuxerList::iterator it = findByPid(_demuxers,pid);
	if (it != _demuxers.end()) {
		//	Remove from list
		delete (*it);
		_demuxers.erase( it );
	}
	_mDemuxers.unlock();

	//	Cleanup data from filter && demuxer
	clear( pid );

	//	Remove timer if necesary
	if (!haveTimers()) {
		rmTimer( pid );
	}

	//	Destroy filter
	if (destroy) {
		filter->deinitialize();
		delete filter;
	}
}

void Provider::restartFilter( ID pid ) {
	LDEBUG( "Provider", "Restart filter: pid=%04x", pid );
	
	_mFilters.lock();
	FilterList::iterator it = findByPid(_filters,pid);
	if (it != _filters.end()) {
		//	Get filter
		(*it)->restart();
	}
	_mFilters.unlock();
}

Filter *Provider::removeFilter( ID pid ) {
	Filter *filter=NULL;
	//LDEBUG( "Provider", "Remove filter: pid=%04x", pid );	
	
	_mFilters.lock();
	FilterList::iterator it = findByPid(_filters,pid);
	if (it != _filters.end()) {
		//	Get filter
		filter = (*it);

		//	Remove from list
		_filters.erase( it );
	}
	_mFilters.unlock();

	return filter;
}

bool Provider::canAllocFilter( bool isSection, ID pid ) {
	bool result;
	bool maxAllocated=true;
	
	_mFilters.lock();
	result = isSection ? (_filters.size() < static_cast<size_t>(maxFilters())) : true;
	if (result) {
		maxAllocated=false;
		FilterList::const_iterator it = findByPid( _filters, pid );
		result = (it == _filters.end());
	}
	_mFilters.unlock();

	if (!result) {
		std::string err="Cannot alloc a new filter. ";
		err += maxAllocated ? "Maximum already allocated" : "PID already filtrated";
		err += ": pid=%04x";
		LWARN( "Provider", err.c_str(), pid );
	}

	return result;
}

void Provider::timeout( ID pid ) {
	enqueue( pid, NULL );
}

void Provider::enqueue( ID pid, util::BYTE *data, DWORD size ) {
	assert(data);
	util::Buffer *buf = getNetworkBuffer();
	buf->copy( (char *)data, size );
	enqueue( pid, buf );	
}

void Provider::enqueue( ID pid, util::Buffer *buf ) {
	_mutex.lock();

	//	Only enqueue if network is tunned or scanning
	if (_state > tuner::state::idle) {
		//	Enqueue filter data into provider thread
		impl::SectionBuffer data;
		data.pid = pid;
		data.buf = buf;
		_pool->put( data );
		if (!haveTimers()) {
			resetTimer(pid);
		}
		//LDEBUG( "Provider", "Enqueue section: pid=%04x", pid );
	}
	else if (buf) {
		LWARN( "Provider", "Ignoring section: Provider idle: pid=%04x", pid );
		
		//	Returns network buffer
		freeNetworkBuffer( buf );
	}
	
	_mutex.unlock();
}
	
void Provider::clear( ID pid ) {
	//	Clean sections from pid
	_pool->remove_if( RemoveSections<impl::SectionBuffer>(pid) );

	//	Clean notifications from pid
	_mNotifications.lock();
	util::erase_if( _notifications, RemoveSections<impl::SectionNotification>(pid) );
	_mNotifications.unlock();	
}

void Provider::processSectionBuffer( PSIDemuxer *demux, util::Buffer *buf ) {
	demux->process( buf );
}

void Provider::processSection( impl::SectionBuffer &data ) {
	//	Process section buffer if demuxer exists
	_mDemuxers.lock();	
	DemuxerList::const_iterator it = findByPid( _demuxers, data.pid );
	if (it != _demuxers.end()) {
		processSectionBuffer( (*it), data.buf );
	}
	_mDemuxers.unlock();

	//	Free buffer
	if (data.buf) {
		_pool->free( data.buf );
	}
}

void Provider::onNotification( ID pid, const boost::function<void (void)> &callback ) {
	_mNotifications.lock();
	impl::SectionNotification data;
	data.callback = callback;
	data.pid = pid;
	_notifications.push_back( data );
	_mNotifications.unlock();
	_disp->post( this, boost::bind(&Provider::postNotification,this) );
}

void Provider::postNotification() {
	impl::SectionNotification evt;
	bool exit=false;

	while (!exit) { 
		//	Get element from list
		_mNotifications.lock();
		if (_notifications.empty()) {
			exit=true;
		}
		else {
			evt = _notifications.front();
			_notifications.pop_front();
		}
		_mNotifications.unlock();

		if (!exit) {
			evt.callback();
		}
	}
}

void Provider::setDispatcher( util::task::Dispatcher *disp ) {
	_disp = disp;
	_disp->registerTarget( this );	
}

util::task::Dispatcher *Provider::dispatcher() const {
	return _disp;
}

//	Aux timers
void Provider::addTimer( ID pid, int ms ) {
	//LDEBUG( "Provider", "Add timer: pid=%04x", pid );
	_mTimers.lock();
	_timers.push_back( new impl::Timer( pid, ms ) );
	_mTimers.unlock();

	//	Notify	
	_pool->notify_all( util::pool::mask::user );
}

void Provider::rmTimer( ID pid ) {
	//LDEBUG( "Provider", "Remove timer: pid=%04x", pid );	
	_mTimers.lock();	
	TimerList::iterator it=std::find_if(
		_timers.begin(),
		_timers.end(),
		boost::bind(&impl::Timer::pid,_1) == pid );
	if (it != _timers.end()) {
		delete (*it);
		_timers.erase( it );
	}
	_mTimers.unlock();

	//	Notify
	_pool->notify_all( util::pool::mask::user );
}

void Provider::resetTimer( ID pid ) {
	_mTimers.lock_shared();
	TimerList::const_iterator it=std::find_if(
		_timers.begin(),
		_timers.end(),
		boost::bind(&impl::Timer::pid,_1) == pid );
	if (it != _timers.end()) {
		//LDEBUG( "Provider", "Reset timer: pid=%04x", pid );		
		(*it)->reset();
	}
	_mTimers.unlock_shared();
}

int Provider::checkTimers() {
	int min = std::numeric_limits<int>::max();
	
	if (!_timers.empty()) {
		std::vector<ID> pids;
		impl::pt::ptime now = impl::pt::microsec_clock::local_time();

		//	Consume elapsed time, if any expired add to post processing
		_mTimers.lock_shared();
		BOOST_FOREACH( impl::Timer *timer, _timers ) {
			if (timer->consume( now )) {
				pids.push_back( timer->pid() );
			}
			if (timer->timer() < min) {
				min = timer->timer();
			}
		}
		_mTimers.unlock_shared();

		//	Process timers
		BOOST_FOREACH( ID pid, pids ) {
			if (pid != TS_PID_NULL) {
				timeout( pid );
			}
			else {
				checkStatus();
			}
		}
	}

	if (min == std::numeric_limits<int>::max()) {
		min = -1;
	}

	//LDEBUG( "Provider", "Check timers end: ms=%d", min );
	
	return min;
}

//	Aux network buffer
util::Buffer *Provider::getNetworkBuffer() {
	return _pool->alloc();
}

void Provider::freeNetworkBuffer( util::Buffer *buf ) {
	_pool->free( buf );
}

//	Aux thread
void Provider::providerThread() {
	impl::SectionBuffer item;
	util::DWORD mask;
	bool exit=false;
	int ms;

	LINFO( "Provider", "Begin provider thread" );
	
	while (!exit) {
		//	Check timers and get minimum to wait
		ms = checkTimers();
			
		//	Get section
		if (_pool->get( item, mask, ms )) {
			processSection( item );
		}

		if (mask & util::pool::mask::exit) {
			exit = true;
		}
	}

	LINFO( "Provider", "End provider thread" );
}

}

