/*******************************************************************************

  Copyright (C) 2010, 2011 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-connector implementation.

    DTV-connector is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-connector is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-connector.

    DTV-connector es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-connector se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "spawner.h"
#include "taskitem.h"
#include <util/mcr.h>
#include <boost/foreach.hpp>
#include <algorithm>
#include <util/log.h>

namespace util {
namespace spawner {

struct Spawner::MaxTaskFinder {
	MaxTaskFinder() : _item( NULL ) { _pri = 0; }
	void operator()( TaskItem *item ) {
		if (!item->isRunning() && item->priority() > _pri) {
			_item = item;
			_pri  = item->priority();
		}
	}
	TaskItem *_item;
	int _pri;
};

struct Spawner::MinTaskFinder {
	MinTaskFinder( int priority ) : _item( NULL ) { _pri = priority; }
	void operator()( TaskItem *item ) {
		if (item->isRunning() && item->priority() < _pri) {
			_item = item;
			_pri  = item->priority();
		}
	}
	TaskItem *_item;
	int _pri;
};

Spawner::Spawner( int maxParallel )
{
	_enable = true;
	_maxParallel = maxParallel;
}

Spawner::~Spawner( void )
{
	CLEAN_ALL( TaskItem*, _tasks );
}

//	Spawner methods
status::type Spawner::run( TaskItem *item ) {
	assert( item );

	//	Try start
	status::type st=exec( item );
	if (st > status::failed) {
		//	Add item into tasks
		_tasks.push_back( item );
	}
	else {
		delete item;
	}
	
	return st;
}

void Spawner::stop( TaskItem *item, int msTimeout/*=0*/ ) {
	Tasks::iterator it=std::find( _tasks.begin(), _tasks.end(), item );
	if (it != _tasks.end()) {
		if (item->isRunning()) {
			item->stop( msTimeout);
		}

		//	Remove task from list
		delete item;
		_tasks.erase( it );

		//	Try run queued
		runQueued();
	}
}

void Spawner::enable( bool mustEnable ) {
	if (_enable != mustEnable) {
		LINFO("util::Spawner", "Enable: old=%d, new=%d", _enable, mustEnable);
		
		_enable = mustEnable;
		
		if (_enable) {
			// Spawn pending process
			for (int i=0; i<static_cast<int>(_tasks.size()); i++) {
				if (!runQueued()) {
					break;
				}
			}
		}
		else {
			// Stop all process and enqueue
			BOOST_FOREACH( TaskItem *item, _tasks ) {
				if (item->isRunning()) {
					stopAndQueue(item);
				}
			}
		}
	}
}

void Spawner::stopAndQueue( TaskItem *item ) {
	LINFO("util::Spawner", "Stop and queue: pid=%d", item->process()->pid());

	//	Notify queued
	item->stop();
	item->queue();
}

int Spawner::maxParallel() const {
	return _maxParallel;
}

int Spawner::running() const {
	int res=0;
	BOOST_FOREACH( TaskItem *item, _tasks ) {
		if (item->isRunning()) {
			res++;
		}
	}
	return res;
}

int Spawner::tasks() const {
	return _tasks.size();
}

bool Spawner::canRun( int priority ) {
	bool result=false;

	if (_enable) {
		if (running() < maxParallel()) {
			result=true;
		}
		else {
			//	Check if some task running has low priority
			MinTaskFinder minTask(priority);
			if (find( minTask )) {
				stopAndQueue( minTask._item );
				result=true;
			}
		}
	}

	LDEBUG( "util::Spawner", "Can run: priority=%d, enable=%d, result=%d",
		priority, _enable, result );
	
	return result;
}

status::type Spawner::exec( TaskItem *item ) {
	status::type st=status::failed;
	if (canRun( item->priority() )) {
		//	Try fork process
		if ( item->start() ) {
			st = status::started;
		}
	}
	else {
		//	Enqueue process
		item->queue();
		st=status::queued;
	}

	return st;
}

bool Spawner::runQueued( void ) {
	bool result=false;
	LDEBUG("util::Spawner", "Run queued process begin: size=%d", _tasks.size());
	
	//	Get task with maximum priority
	MaxTaskFinder max;
	if (find( max )) {
		//	Run item
		result = (exec( max._item ) == status::started);
	}
	LDEBUG("util::Spawner", "Run queued process end");
	return result;
}

}
}
