/*******************************************************************************

  Copyright (C) 2010, 2011 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-connector implementation.

    DTV-connector is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-connector is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-connector.

    DTV-connector es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-connector se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/
#include "nclapplication.h"
#include "nclprofile.h"
#include "generated/config.h"
#include "../../../controller/applicationcontroller.h"
#include "../../../controller/spawner.h"
#include "../../../controller/spawner/taskitem.h"
#include "../../../applicationextension.h"
#include "../../../../../servicemanager.h"
#include "../../../../../../resourcemanager.h"
#include "../../../../../../demuxer/psi/dsmcc/event.h"
#include <connector/connector.h>
#include <connector/handler/keepalivehandler.h>
#include <connector/handler/keyregisterhandler.h>
#include <connector/handler/keyeventhandler.h>
#include <connector/handler/videoresizehandler.h>
#include <connector/handler/exitcommandhandler.h>
#include <connector/handler/editingcommandhandler.h>
#include <connector/handler/startcommandhandler.h>
#include <util/task/dispatcher.h>
#include <util/cfg/configregistrator.h>
#include <util/string.h>
#include <boost/lexical_cast.hpp>
#include <boost/foreach.hpp>
#include <set>
#include <assert.h>
#include <util/log.h>

namespace tuner {
namespace app {

namespace SpawnerStatus = util::spawner::status;

NCLApplication::NCLApplication( ApplicationExtension *ext, const ApplicationID &id, const std::string &script )
	: Application( ext, id ), _script( script )
{
	name( script );
	_connector = NULL;
	_proc = NULL;
}

NCLApplication::~NCLApplication()
{
	assert(!_connector);
	assert(!_proc);
}

void NCLApplication::show() const {
	Application::show();
	LINFO("app::NCLApplication", "\tscript=%s", _script.c_str());
}

const std::string &NCLApplication::script() const {
	return _script;
}

status::type NCLApplication::doStart() {
	//	Setup item to spawn
	_proc = new util::spawner::TaskItem( NCLProfile::ginga() );
	_proc->priority( priority() );
	_proc->onStart( boost::bind( &NCLApplication::setupProcess, this ) );
	_proc->onQueue( boost::bind( &Application::status, this, status::waiting ) );

	//	Run ginga
	SpawnerStatus::type st=ctrl()->spawner()->run( _proc );
	if (st == SpawnerStatus::failed) {
		_proc = NULL;

		//	Cleanup connector
		delete _connector;
		_connector = NULL;
	}
	else {
		//	Mark status to starting
		status( status::starting );
	}

	//	Register into dispatcher to get called
	extension()->srvMgr()->dispatcher()->registerTarget( this );

	return status();
}

void NCLApplication::doStop( bool kill/*=false*/ ) {
	assert( _proc );

	//	UnRegister into dispatcher to get called
	extension()->srvMgr()->dispatcher()->unregisterTarget( this );	
	
	//	Stop ginga
	if (_connector && status() > status::starting) {
		connector::ExitCommandHandler exit;
		exit.send( _connector );
	}

	//	Kill process if ginga exit fail
	ctrl()->spawner()->stop( _proc, kill ? 0 : util::cfg::getValue<int>("tuner.service.ginga-ncl.timeout") );

	//	Cleanup connector
	delete _connector;
	_connector = NULL;

	_proc = NULL;
}

//	NCL Editing command
void NCLApplication::onEditingCommand( const Buffer &buf ) {
	LDEBUG("app::NCLApplication", "Editing command received: size=%d", buf.length());
	if (_connector) {
		connector::EditingCommandHandler cmd;
		cmd.send( _connector, buf );
	}
}

struct NCLEventFinder {
	NCLEventFinder( std::vector<std::string> &events ) : _events( events ) {}
	void operator()( dsmcc::Event *evt ) {
		util::WORD eventID;
		evt->show();
		if (evt->includes( "nclEditingCommand", eventID )) {
			_events.push_back( evt->url() );
		}
	}

	std::vector<std::string> &_events;

private:
	// Avoid VisualStudio warning C4512
	NCLEventFinder &operator=(NCLEventFinder &/*ef*/) { return *this; }
};

void NCLApplication::registerEditingCommands() {
	LDEBUG("app::NCLApplication", "Register editing commands");
	
	std::vector<std::string> names;
	NCLEventFinder iter(names);
	loopEvents( iter );
	BOOST_FOREACH( const std::string &name, names ) {
		LDEBUG("app::NCLApplication", "Editing command found: event=%s", name.c_str());
		registerEvent( name, "nclEditingCommand", boost::bind(&NCLApplication::onEditingCommand,this,_1) );
	}
}

//	Connector aux
void NCLApplication::setupProcess() {
	//	Connector can be active, if so, stop process and continue
	assert(!_connector);
	assert(_proc);

	//	Make pipe
	std::string pipe = extension()->resMgr()->mkTempFileName( "connector_%%%%%%" );
	LDEBUG("app::NCLApplication", "Setup process: pipe=%s", pipe.c_str());

	_proc->process()->clearParams();
	_proc->process()->clearEnvironment();

	std::string setting="--set=";
	
	//	Setup remote canvas
	setting += "gui.use=custom:gui.window.use=remote:";

	//	Setup NIT
	ID nitID = extension()->srvMgr()->currentNit();
	if (nitID != NIT_ID_RESERVED) {
		_proc->process()->addParam( "--baseId" );
		_proc->process()->addParam( boost::lexical_cast<std::string>( nitID ) );
	}

	{	//	Setup resolution
		int w,h;
		setupLayerResolution( w, h );
		
		setting += "gui.window.size.width=" + boost::lexical_cast<std::string>(w);
		setting += ":gui.window.size.height=" + boost::lexical_cast<std::string>(h);
	}

	_proc->process()->addParam( setting );

	//	Setup pipe
	_proc->process()->addParam( "-c" );
	_proc->process()->addParam( pipe );

	//	Setup NCL name
	std::string ncl = path() + "/" + _script;
	_proc->process()->addParam( "--ncl" );
	_proc->process()->addParam( ncl );

	//	Setup envirioment
	_proc->process()->environment( "GINGA_PROCESS_ENV" );

	//	Create connector
	_connector = new connector::Connector( pipe, true );
	_connector->setDispatcher( boost::bind(&util::task::Dispatcher::post,extension()->srvMgr()->dispatcher(),this,_1) );

	//	Handle connect
	_connector->onConnect( boost::bind( &NCLApplication::onGingaConnected, this ) );

	//	Handle error cases
	boost::function<void (void)> fnc = boost::bind(&Application::stop, this, false);
	_connector->onConnectTimeout( fnc );
	_connector->onDisconnect( fnc );
	_connector->onTimeout( fnc );
	
	{	//	Reserve/Release keys messages
		connector::KeyRegisterHandler *keyHandler = new connector::KeyRegisterHandler();
		keyHandler->onKeyEvent( boost::bind(&NCLApplication::onRegisterKeyEvent,this,_1) );
		_connector->addHandler( keyHandler );
	}

	{	//	Video resize
		connector::VideoResizeHandler<int> *videoResizeHandler = new connector::VideoResizeHandler<int>();
		videoResizeHandler->onVideoResizeEvent( boost::bind(&NCLApplication::onResizeVideoEvent,this,_1) );
		_connector->addHandler( videoResizeHandler );
	}

	{	//	Setup start presentation
		connector::StartCommandHandler *handler = new connector::StartCommandHandler();
		handler->onStartPresentationEvent( boost::bind( &NCLApplication::onStartPresentation, this, _1 ) );
		_connector->addHandler( handler );
	}		
}

//	Status changes
void NCLApplication::onStatusChanged( status::type /*old*/, status::type actual ) {
	switch (actual) {
		case status::waiting: {
			//	Kill connnector
			delete _connector;
			_connector = NULL;
			break;
		}
		default:
			break;
	};
}

void NCLApplication::onGingaConnected() {
	//	If application is starting
	if (status() == status::starting) {
		//	Change status
		status( status::started );
	}
	else {
		LWARN("app::NCLApplication", "ginga started but not processed becouse of a status change");
	}
}

void NCLApplication::onStartPresentation( bool start ) {
	LINFO("app::NCLApplication", "Ginga presentation started: %d", start);
	if (start) {
		//	Register to all nclEditingCommand events
		registerEditingCommands();
	}

	//	Change status
	status( start ? status::ready : status::started );
}

void NCLApplication::onResizeVideoEvent( const connector::NCLVideoDescriptorPtr &av ) {
	resizeVideo( av->getX(), av->getY(), av->getWidth(), av->getHeight() );
}

void NCLApplication::dispatchKeyImpl( util::key::type key, bool isUp ) {
	LDEBUG("app::NCLApplication", "Dispatch key: key=%d, isUp=%d", key, isUp);
	if (_connector) {
		connector::KeyEventHandler handler;
		handler.send( _connector, key, isUp );
	}
}

void NCLApplication::onRegisterKeyEvent( const connector::KeyRegisterDataPtr &data ) {
	std::vector<util::key::type> keys;
	BOOST_FOREACH( util::key::type key, *data.get() ) {
		keys.push_back( key );
	}
	reserveKeys( keys );
}

}
}
