/*******************************************************************************

  Copyright (C) 2010, 2011 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-connector implementation.

    DTV-connector is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-connector is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-connector.

    DTV-connector es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-connector se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/
#include "service.h"
#include "../demuxer/ts.h"
#include "../demuxer/descriptors.h"
#include "../demuxer/descriptors/generic.h"
#include "../demuxer/psi/psi.h"
#include "../demuxer/psi/nit.h"
#include "../demuxer/psi/sdt.h"
#include "../demuxer/psi/pmt.h"
#include "generated/config.h"
#include <util/log.h>
#include <util/string.h>
#include <util/cfg/configregistrator.h>
#include <boost/foreach.hpp>
#include <boost/filesystem.hpp>

REGISTER_INIT_CONFIG( tuner_service ) {
	root().addNode("service").addNode("ginga-ncl")
		.addValue("process", "Ginga process", std::string(GINGA_DEFAULT_PROCESS))
		.addValue("version", "Ginga version file", std::string(GINGA_VERSION_FILE))
		.addValue("timeout", "Ginga timeout", 5000 );
}

namespace tuner {

Service::Service()
{
	_networkID   = -1;
	_nitID       = NIT_ID_RESERVED;
	_tsID        = 0;
	_id          = 0;
	_pid         = TS_PID_NULL;
	_pcrPID      = TS_PID_NULL;
	_type        = service::type::digital_television;
	_eitSchedule = false;
	_eitPresent  = false;
	_status      = service::status::undefined;
	_state       = service::state::present;
	_caMode      = false;
	_remote      = 0;
}

Service::Service( size_t networkID, ID nitID, ID tsID, ID programID, ID pid )
{
	_networkID   = networkID;
	_nitID       = nitID;
	_tsID        = tsID;
	_id          = programID;
	_pid         = pid;
	_pcrPID      = TS_PID_NULL;
	_type        = service::type::digital_television;
	_eitSchedule = false;
	_eitPresent  = false;	
	_status      = service::status::undefined;
	_state       = service::state::present;	
	_caMode      = false;
	_remote      = 0;
}

Service::Service( const Service &other )
	: _elements( other._elements ), _descriptors( other._descriptors )
{
	_networkID   = other._networkID;
	_nitID       = other._nitID;
	_tsID        = other._tsID;
	_id          = other._id;
	_pid         = other._pid;
	_pcrPID      = other._pcrPID;
	_type        = other._type;
	_eitSchedule = other._eitSchedule;
	_eitPresent  = other._eitPresent;
	_status      = other._status;
	_state       = other._state;
	_caMode      = other._caMode;
	_remote      = other._remote;
}

Service::~Service()
{
}

size_t Service::networkID() const {
	return _networkID;
}

ID Service::nitID() const {
	return _nitID;
}

ID Service::tsID() const {
	return _tsID;
}

ID Service::id() const {
	return _id;
}

ID Service::pid() const {
	return _pid;
}

service::Type Service::type() const {
	return _type;
}

const std::string &Service::provider() const {
	return _provider;
}

const std::string &Service::name() const {
	return _name;
}

const std::string &Service::networkName() const {
	return _nitName;
}

const std::vector<ElementaryInfo> &Service::elements() const {
	return _elements;
}

int Service::remote() const {
	return _remote;
}

const std::string &Service::tsName() const {
	return _tsName;
}

std::string Service::logicalChannel() const {
	//	ABNT NBR 15603-2 Annex G and H
	return util::format( "%02d.%01d%01d",
		_remote, (_id & 0x0018) >> 3, (_id & 0x0007)+1 );
}

ID Service::pcrPID() const {
	return _pcrPID;
}

service::status::type Service::status() const {
	return _status;
}

service::state::type Service::state() const {
	return _state;
}

const desc::MapOfDescriptors &Service::descriptors() const {
	return _descriptors;
}

void Service::state( service::state::type st ) {
	_state = st;
}

void Service::show() const {
	LINFO("Service", "Service: id=(%04x,%04x,%04x), type=%02x, name=%s, provider=%s", _id, _tsID, _nitID, _type, _name.c_str(), _provider.c_str());
	LINFO("Service", "\t pid=%x, pcrPID=%x, elements=%d", _pid, _pcrPID, _elements.size());
	BOOST_FOREACH( const ElementaryInfo &info, _elements ) {
		LINFO( "Service", "\t\ttype=%02x, pid=%04x, Descriptors=%s",
			info.streamType, info.pid, desc::show( info.descriptors ).c_str() );
	}		
	LINFO("Service", "\t TS: name=%s, remote=%d, logicalChannel=%s", _tsName.c_str(), _remote, logicalChannel().c_str());
	LINFO("Service", "\t Network: name=%s", _nitName.c_str());
	LINFO("Service", "\t status=%x, EIT flags(schedule=%d,present=%d), caMode=%d", _status, _eitSchedule, _eitPresent, _caMode);
}

//	Process tables
bool Service::process( const boost::shared_ptr<Pmt> &pmt ) {
	bool changed=false;
	
	if (pmt->programID() == _id &&
		pmt->pid() == _pid)
	{
		_pcrPID   = pmt->pcrPID();
		_elements = pmt->elements();
		_state    = service::state::complete;
		process( pmt->descriptors() );
		changed   = true;
	}
	
	return changed;
}

bool Service::process( const boost::shared_ptr<Nit> &nit ) {
	bool changed = false;
	
	if (nit->actual()) {
		//	Save nitID
		_nitID = nit->networkID();
		
		//	Process NIT network descriptors
		changed |= process( nit->descriptors() );
	
		//	Process NIT networks
		BOOST_FOREACH( Nit::ElementaryInfo info, nit->elements() ) {
			//	Process NIT service descriptor
			changed |= process( info.descriptors );
		}
	}

	return changed;
}

bool Service::process( const boost::shared_ptr<Sdt> &sdt ) {
	bool changed = false;
	
	//	SDT is for actual network?
	if (sdt->actual()) {
		//	Process SDT services
		const std::vector<Sdt::Service> &services = sdt->services();
		BOOST_FOREACH( Sdt::Service srv, services ) {
			if (srv.serviceID == _id) {
				//	Update information
				_eitSchedule = srv.eitSchedule;
				_eitPresent  = srv.eitPresentFollowing;				
				_status      = srv.status;
				_caMode      = srv.caMode;

				//	Process descriptors!
				changed |= process( srv.descriptors );
			}
		}
	}

	return changed;
}

bool Service::process( const desc::MapOfDescriptors &descriptors ) {
	bool changed = false;
	
	BOOST_FOREACH( desc::DescriptorItem itDesc, descriptors ) {
		util::BYTE tag = itDesc.first;
		bool parsed=false;

		switch (tag) {
			case desc::ident::network_name: {
				//	Network Name descriptor (0x40)
				_nitName = itDesc.second.get<std::string>();
				parsed = true;
				break;
			}
			case desc::ident::service_list: {
				//	Service List Descriptor (0x41)
				const desc::ServiceListDescriptor &serviceList = itDesc.second.get<desc::ServiceListDescriptor>();
				BOOST_FOREACH( desc::Service srv, serviceList ) {
					if (srv.first == _id) {
 						_type   = srv.second;
						parsed = true;
					}
				}
				break;
			}
			case desc::ident::service: {
				//	Service Descriptor (0x48)
				const desc::ServiceDescriptor &srvDesc = itDesc.second.get<desc::ServiceDescriptor>();
				_type     = srvDesc.type;
				_name     = srvDesc.name;
				_provider = srvDesc.provider;
				parsed   = true;
				break;
			}
			case desc::ident::ts_information: {
				//	TS Information Descriptor
				const desc::TransportStreamInformationDescriptor &desc = itDesc.second.get<desc::TransportStreamInformationDescriptor>();
				_remote = desc.remoteControlKeyID;
				_tsName = desc.name;
				parsed = true;
				break;
			}
			default:
				LDEBUG("Service", "Descriptor not processsed: %x", tag);
		};

		if (!parsed) {
			//	Save descriptor
			_descriptors[tag] = itDesc.second;
		}
		changed |= parsed;
	}

	return changed;
}

}
