/*******************************************************************************

  Copyright (C) 2010, 2011 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-connector implementation.

    DTV-connector is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-connector is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-connector.

    DTV-connector es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-connector se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "servicemanager.h"
#include "service.h"
#include "extension/extension.h"
#include "../resourcemanager.h"
#include "../provider/provider.h"
#include <util/log.h>
#include <util/mcr.h>
#include <assert.h>

#define NOTIFY(f)	\
	BOOST_FOREACH(Extension *ext, _extensions) { if (ext->mustNotify(_inScanMode)) { ext->f; } }

namespace tuner {

ServiceManager::ServiceManager( ResourceManager *resMgr )
{
	assert(resMgr);
	_resMgr = resMgr;
	_initialized = false;
	_running = false;
	_inScan = false;
	_inScanMode = false;	
	_isReady = false;
}

ServiceManager::~ServiceManager()
{
	//	Destroy extensions
	CLEAN_ALL( Extension *, _extensions );
	delete _resMgr;

	//	Checks
	assert(!_initialized);
	assert(!_running);
	assert(!_inScan);
	assert(!_inScanMode);
	assert(!_isReady);	
	assert(_services.empty());
	assert(_servicesQueued.empty());
}

//	Initialization/finalization
void ServiceManager::initialize() {
	LINFO( "ServiceManager", "Initialize" );
	assert(!_initialized);
	assert(!_running);	
	assert(!_inScan);
	assert(!_inScanMode);
	assert(!_isReady);		
	assert(_services.empty());
	assert(_servicesQueued.empty());	
	BOOST_FOREACH( Extension *ext, _extensions ) {
		ext->initialize( this );
	}
	_initialized = true;	
}

void ServiceManager::finalize() {
	LINFO( "ServiceManager", "Finalize" );	

	//	Checks
	assert(!_running);
	assert(!_isReady);
	assert(_services.empty());
	assert(_servicesQueued.empty());

	if (_initialized) {
		BOOST_FOREACH( Extension *ext, _extensions ) {
			ext->finalize();
		}
		_initialized = false;
	}
}

//	Start/Stop
void ServiceManager::start( bool scan/*=false*/ ) {
	LDEBUG( "ServiceManager", "Start: scan=%d", scan );

	//	Checks
	assert(_initialized);
	assert(!_running);
	assert(!_inScan);
	assert(!_inScanMode);
	assert(!_isReady); 	
	assert(_services.empty());
	assert(_servicesQueued.empty());

	_inScanMode = scan;
	NOTIFY( onStart() );
	_running = true;
}

void ServiceManager::stop() {
	LDEBUG( "ServiceManager", "Stop" );
	if (_running) {
		//	Notify all extensions
		NOTIFY( onStop() );

		//	Clean state
		_servicesQueued.clear();
		_running = false;
		_inScan = false;
		_inScanMode = false;
		_isReady = false;

		//	Clean temporary files
		resMgr()->clean();
	}
}

bool ServiceManager::isRunning() const {
	return _running;
}

//	Start/Stop services
void ServiceManager::startService( ID service ) {
	LDEBUG("ServiceManager", "Start service by ID=%x", service);
	
	//	Check if service ready ...
	Service *srv = findService( service );
	if (srv && isReady() && srv->state() == service::state::ready) {
		//	Start service ...
		startService( srv );
	}
	else {
		//	Check if service already enqueued ...
		std::vector<ID>::iterator it=std::find( _servicesQueued.begin(), _servicesQueued.end(), service );
		if (it == _servicesQueued.end()) {		
			//	enqueue and wait ...
			_servicesQueued.push_back( service );
		}
	}
}

void ServiceManager::stopService( ID service ) {
	LDEBUG("ServiceManager", "Stop service by ID begin: %04x", service);

	//	Find service ...
	Service *srv = findService( service );
	if (srv) {
		//	Stop service ...
		stopService( srv );
	}
	else {
		//	Find service in services queued to run
		std::vector<ID>::iterator it=std::find( _servicesQueued.begin(), _servicesQueued.end(), service );
		if (it != _servicesQueued.end()) {
			//	Only remove from list ...
			_servicesQueued.erase( it );
		}
	}

	LDEBUG("ServiceManager", "Stop service by ID end: %04x", service);
}

void ServiceManager::stopServices() {
	LDEBUG("ServiceManager", "Stop services");

	//	Stop all services runnning
	BOOST_FOREACH( Service *srv, _services ) {
		stopService( srv );
	}

	//	Cleanup all services queued
	_servicesQueued.clear();
}

//	Services
ServiceManager::Services &ServiceManager::services() {
	return _services;
}

void ServiceManager::addService( Service *srv ) {
	_services.push_back( srv );
}

Service *ServiceManager::findService( ID service ) {
	Services::const_iterator it=std::find_if(
		_services.begin(),
		_services.end(),
		boost::bind(&Service::id,_1) == service
	);
	return (it != _services.end()) ? (*it) : NULL;
}

Service *ServiceManager::findServiceByPID( ID pid ) {
	Services::const_iterator it=std::find_if(
		_services.begin(),
		_services.end(),
		boost::bind(&Service::pid,_1) == pid
	);
	return (it != _services.end()) ? (*it) : NULL;
}

void ServiceManager::startService( Service *srv ) {
	LDEBUG("ServiceManager", "On start service begin: id=%x", srv->id());

	assert( srv->state() == service::state::ready );
	
	//	Notify to extensions of service was started
	NOTIFY( onServiceStarted( srv, true ) );
	
	//	Mark service as running
	srv->state( service::state::running );

	LDEBUG("ServiceManager", "On start service end: id=%x", srv->id());
}

bool ServiceManager::stopService( Service *srv ) {
	bool result=false;
	
	if (srv->state() == service::state::running) {
		LDEBUG("ServiceManager", "On stop service: %04x", srv->id());

		//	Mark service as ready
		srv->state( service::state::ready );

		//	Notify to extensions
		NOTIFY( onServiceStarted( srv, false ) );
		
		result=true;
	}
	
	return result;
}

void ServiceManager::setServiceReady( Service *srv ) {
	LDEBUG("ServiceManager", "Service ready: id=(%04x,%04x,%04x)",
		srv->nitID(), srv->tsID(), srv->id() );

	//	Set state
	assert( srv->state() == service::state::complete );
	srv->state( service::state::ready );

	//	Notify to extensions
	NOTIFY( onServiceReady( srv, true ) );
				
	//	If serviceID was queued to run
	std::vector<ID>::iterator it=std::find( _servicesQueued.begin(), _servicesQueued.end(), srv->id() );
	if (it != _servicesQueued.end()) {
		//	Start service
		startService( srv );
		_servicesQueued.erase( it );
	}
}

void ServiceManager::expireService( Service *srv, bool needExpire/*=true*/ ) {
	if (srv->state() >= service::state::ready) {
		LDEBUG("ServiceManager", "On expire service: serviceID=%04x, needExpire=%d", srv->id(), needExpire);
		
		//	Stop service if running
		bool stopped=stopService( srv );

		//	Set new status
		srv->state( needExpire ? service::state::expired : service::state::complete );

		//	Notify to extensions
		NOTIFY( onServiceReady( srv, false ) );

		//	Queue running state
		if (stopped) {
			startService( srv->id() );
		}
	}
}

void ServiceManager::serviceChanged( Service *srv ) {
	assert(srv);
	if (isReady() && srv->state() == service::state::complete) {
		setServiceReady( srv );
	}	
}

//	State
bool ServiceManager::isReady() const {
	return _isReady;
}

void ServiceManager::setReady() {
	if (!_isReady) {
		LDEBUG("ServiceManager", "Service provider ready");
		
		_isReady = true;
	
		//	Notify to extensions
		NOTIFY( onReady(true) );

		//	Notify all services ready
		BOOST_FOREACH( Service *srv, _services ) {
			if (srv->state() == service::state::complete) {
				setServiceReady( srv );
			}
		}

		//	Notify end scan
		notifyEndScan();
	}
}

void ServiceManager::expire() {
	if (_isReady) {
		LDEBUG("ServiceManager", "Service provider expired");

		_isReady = false;

		//	Expire all services
		BOOST_FOREACH( Service *srv, _services ) {
			expireService( srv, false );
		}

		//	Notify to extensions
		NOTIFY( onReady( false ) );
	}
}

//	Scan
void ServiceManager::onEndScan( const OnEndScan &callback ) {
	_onEndScan = callback;
}

void ServiceManager::notifyEndScan( bool allScanned/*=false*/ ) {
	if (!_inScan) {
		//	Check if last scanned ...
		if (!allScanned) {
			allScanned=true;
			BOOST_FOREACH( Service *srv, _services ) {
				allScanned &= (srv->state() >= service::state::timeout);
			}
		}

		if (allScanned) {
			LDEBUG("ServiceManager", "Notify end scan");
			_inScan = true;
			if (!_onEndScan.empty()) {
				_onEndScan();
			}
		}
	}
}

//	Extensions
bool ServiceManager::isAttached( Extension *ext ) const {
	return std::find( _extensions.begin(), _extensions.end(), ext ) != _extensions.end();
}

void ServiceManager::attach( Extension *extension ) {
	if (_initialized) {
		extension->initialize( this );
	
		if (_running) {
			extension->onStart();
	
			if (isReady()) {
				extension->onReady( true );

				BOOST_FOREACH( Service *srv, _services ) {
					if (srv->state() >= service::state::ready) {
						extension->onServiceReady( srv, true );

						if (srv->state() >= service::state::running) {
							extension->onServiceStarted( srv, true );
						}
					}
				}
			}
		}
	}

	_extensions.push_back( extension );
}

void ServiceManager::detach( Extension *extension ) {
	std::vector<Extension *>::iterator it=std::find(
		_extensions.begin(), _extensions.end(), extension );
	if (it != _extensions.end()) {
		if (_initialized) {
			if (_running) {
				if (isReady()) {
					BOOST_FOREACH( Service *srv, _services ) {
						if (srv->state() >= service::state::ready) {
							if (srv->state() >= service::state::running) {
								extension->onServiceStarted( srv, false );
							}

							extension->onServiceReady( srv, false );
						}
					}
					
					extension->onReady( false );
				}

				extension->onStop();
			}
			
			extension->finalize();
		}
			
		_extensions.erase( it );
	}
}

//	Getters
ResourceManager *ServiceManager::resMgr() {
	return _resMgr;
}

}
