/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-util implementation.

    DTV-util is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-util is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-util.

    DTV-util es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-util se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "storage.h"
#include "../../string.h"
#include "../../mcr.h"
#include "../../log.h"
#include <boost/filesystem.hpp>
#include <boost/algorithm/string/predicate.hpp>
#include <db.h>
#include <cstdio>

namespace fs = boost::filesystem;

namespace util {
namespace storage {
namespace bdb {

Storage::Storage( const std::string &filename )
{
	_dbName = filename;
	if (!boost::algorithm::iends_with(_dbName, ".db")) {
		_dbName += ".db";
	}
	_buf = NULL;
	_pos = 0;
	_nestedCalls = 0;
}

Storage::~Storage( void ) {}

bool Storage::initialize( void ) {
	u_int32_t flags;
	int ret;
		
	//	Initialize db structure
	ret = db_create(&_dbp, NULL, 0);
	if (ret) {
		throw StorageError( format( "Unable to create db structure: %s\n", db_strerror(ret)) );
	}

	// //	Permit duplicate keys
	// ret = _dbp->set_flags( _dbp, DB_DUPSORT );
	// if (ret) {
	// 	_dbp->close( _dbp, 0 );
	// 	throw StorageError( format( "Attempt to set DUPSORT flag failed!: %s\n", db_strerror(ret)) );
	// }

	//	Database open flags
	flags = DB_CREATE;	//	If the database does not exist, create it.

	//	Try open database ...
	ret = _dbp->open( _dbp, NULL, _dbName.c_str(), NULL, DB_BTREE, flags, 0 );
	if (ret) {
		//	Remove database
		_dbp->remove( _dbp, _dbName.c_str(), NULL, 0 );
		
		//	Invalid? truncate and try open again!
		flags = DB_TRUNCATE;
		ret = _dbp->open(_dbp, NULL, _dbName.c_str(), NULL, DB_BTREE, (DBTYPE)flags, 0 );
		if (ret) {
			throw StorageError( util::format( "Unable to open %s db: %s\n", _dbName.c_str(), db_strerror(ret)) );
		}
	}
	LDEBUG("storage::bdb", "Storage initialized");
	return true;
}

void Storage::finalize( void ) {
	if (_dbp) {
		_dbp->close(_dbp, 0);
	}
	LDEBUG("storage::bdb", "Storage finalized");
}

void Storage::destroyDB() {
	if (std::remove(_dbName.c_str())) {
		LWARN("storage::bdb", "Could not erase database file");
	} else {
		LDEBUG("storage::bdb", "Database file erased successfully");
	}
}

void Storage::endCall( const std::string &key, bool load /*=true*/ ) {
	_nestedCalls--;
	if (_nestedCalls==0) {
		if (!load) {
			saveKey( key );
		}
		DEL(_buf);
	}
}

void Storage::saveKey( const std::string &key ) {
	DBT dKey, dData;
	//	Zero out the DBTs before using them.
	memset(&dKey, 0, sizeof(DBT));
	memset(&dData, 0, sizeof(DBT));

	//	Assign key
	dKey.data = (void *)key.c_str();
	dKey.size = key.length()+1;

	dData.data = (void *)_buf->buffer();
	dData.size = _buf->length();

	//	Store in db
	int ret = _dbp->put( _dbp, NULL, &dKey, &dData, 0 );
	if (ret) {
		throw StorageError( util::format( "Unable to save key %s into db: %s", key.c_str(), db_strerror(ret) ) );
	}
	_dbp->sync( _dbp, 0 );
}

bool Storage::getKey( const std::string &key ) {
	DBT dKey, dData;

	//	Zero out the DBTs before using them.
	memset(&dKey, 0, sizeof(DBT));
	memset(&dData, 0, sizeof(DBT));

	//	Assign key
	dKey.data = (void *)key.c_str();
	dKey.size = key.length()+1;

	//	Retrieve data from db
	int ret = _dbp->get( _dbp, NULL, &dKey, &dData, 0 );
	if (ret == DB_NOTFOUND) {
		return false;
	} else if (ret) {
		throw StorageError( util::format( "Unable to load key %s from db: %s", key.c_str(), db_strerror(ret)) );
	}

	//	Unmarshall data
	_buf->assign( (char *)dData.data, dData.size );
	
	return true;
}

}
}
}
