/*******************************************************************************

  Copyright (C) 2010, 2011 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-util implementation.

    DTV-util is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-util is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementacin de DTV-util.

    DTV-util es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los trminos de la Licencia Pblica General Reducida GNU como es publicada por la
  Free Software Foundation, segn la versin 2 de la licencia.

    DTV-util se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTA; ni siquiera la garanta implcita de COMERCIALIZACIN o ADECUACIN
  PARA ALGN PROPSITO PARTICULAR. Para ms detalles, revise la Licencia Pblica
  General Reducida GNU.

    Ud. debera haber recibido una copia de la Licencia Pblica General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "dispatcher.h"
#include "../../task/dispatcherimpl.h"
#include "../../log.h"
#include "../../mcr.h"
#include <boost/foreach.hpp>
#include <boost/bind.hpp>
#include <ev.h>

#define INIT(data,callback)	\
	assert( !callback.empty() ); \
	data.callback = callback; \
	data.id = allocID();	\
	LDEBUG("io::ev::Dispatcher", "Init event: id=%ld", data.id->getID());

#define STOP_EVENT(t, l, id, method)	\
	std::list<t *>::iterator it=std::find_if( l.begin(), l.end(), impl::EventFinder<t>(id) ); \
	if (it != l.end()) { \
		t *ev=(*it); \
		l.erase(it); \
		ev_##method##_stop( _loop, &ev->ev ); \
		ev->data.id.reset(); \
		free( ev ); \
	} \
	else { LWARN("io::ev::Dispatcher", "invalid " #method ": %ld", id); }

namespace util {
namespace io {
namespace ev {

namespace impl {

struct my_data {
	id::Ident id;
	io::Dispatcher::Callback callback;
};
	
struct my_timer_ev {
	struct ev_timer ev;
	struct my_data data;
};

struct my_io_ev {
	struct ev_io ev;
	struct my_data data;	
};

static void wakeup_cb( struct ev_loop *loop, struct ev_async */*w*/, int /*revents*/ ) {
	Dispatcher *sys = (Dispatcher *)ev_userdata(loop);
	assert(sys);	
	sys->wakeup();
}

static void timer_cb( struct ev_loop *loop, struct ev_timer *timer, int /*revents*/ ) {
	Dispatcher *sys = (Dispatcher *)ev_userdata(loop);
	assert(sys);
	struct my_timer_ev *ev = (struct my_timer_ev *)timer;
	id::Ident id(ev->data.id);
	ev->data.callback( id );	
	sys->stopTimer( id );
}

static void io_cb( struct ev_loop */*loop*/, struct ev_io *w, int /*revents*/ ) {
	struct my_io_ev *ev = (struct my_io_ev *)w;
	ev->data.callback( ev->data.id );	
}

template<class T>
class EventFinder {
public:
	EventFinder( id::ID_TYPE id ) : _id(id) {}
	
	bool operator()( const T *io ) const {
		return io->data.id->getID() == _id;
	}

private:
	id::ID_TYPE _id;
};

}

Dispatcher::Dispatcher()
{
	_loop = NULL;
	_wakeup = NULL;
	_threadID = boost::this_thread::get_id();
	_tasks = NULL;
}

Dispatcher::~Dispatcher()
{
}

//	Initialization
bool Dispatcher::initialize() {
	//	Create dispatcher
	_tasks = new task::DispatcherImpl();
	_tasks->onPost( boost::bind(&Dispatcher::onPost,this) );
	
	//	Create loop
	_loop = ev_loop_new(EVFLAG_AUTO);
	if (!_loop) {
		LERROR("io::ev::Dispatcher", "cannot create loop");
		return false;
	}
	ev_set_userdata( _loop, this );

	//	Create and initialize wakeup watcher
	_wakeup = (ev_async *)malloc( sizeof(struct ev_async) );
	if (!_wakeup) {
		free(_loop);
		LWARN("io::ev::Dispatcher", "cannot create async event");
		return false;
	}
	ev_async_init (_wakeup, impl::wakeup_cb);
	ev_async_start (_loop, _wakeup);

	//	Register into dispatcher
	_tasks->registerTarget( this );

	return true;
}

void Dispatcher::finalize() {
	//	Run all tasks in queue
	_tasks->runAll();

	//	Remove io from dispatcher
	_tasks->unregisterTarget( this );

	//	Cleanup all events
	while (!_io.empty()) {
		ListOfIO::iterator it = _io.begin();
		LWARN("io::ev::Dispatcher", "stopping lost io event: fd=%d", (*it)->ev.fd);
		stopIOImpl( (*it)->data.id->getID() );
	}

	while (!_timers.empty()) {
		ListOfTimer::iterator it = _timers.begin();
		LWARN("io::ev::Dispatcher", "stopping lost timer event");
		stopTimerImpl( (*it)->data.id->getID() );
	}

	//	Destroy loop
	ev_loop_destroy( _loop );
	_loop = NULL;
	free(_wakeup);
	_wakeup = NULL;

	DEL(_tasks);
}

void Dispatcher::run() {
	LINFO( "io::ev::Dispatcher", "Run begin" );

	//	Setup threadID
	_threadID = boost::this_thread::get_id();	

	//	Wait for events to arrive
	ev_run (_loop, 0);

	LINFO( "io::ev::Dispatcher", "Run end" );	
}

void Dispatcher::exit() {
	LINFO( "io::ev::Dispatcher", "Exit" );		

	//	Stop loop
	_tasks->post( this, boost::bind(&Dispatcher::exitLoop, this) );
}

//	Dispatcher
task::Dispatcher *Dispatcher::dispatcher() const {
	return _tasks;
}

void Dispatcher::onPost() {
	//	wakeup loop
	ev_async_send( _loop, _wakeup );
}

//	Timer events
id::Ident Dispatcher::addTimer( int ms, const Callback &callback ) {
	LDEBUG("io::ev::Dispatcher", "Add timer: ms=%d", ms);

	//	Create timer
	struct impl::my_timer_ev *timer = (struct impl::my_timer_ev *)calloc( 1, sizeof(struct impl::my_timer_ev) );
	if (!timer) {
		LERROR("io::ev::Dispatcher", "cannot create timer event");
		return id::Ident();
	}
	else {
		//	Setup
		INIT( timer->data, callback );

		{
			struct ev_timer *ev = &timer->ev;
			ev_timer_init( ev, impl::timer_cb, ms/1000., 0. );
		}

		//	Add to loop
		_tasks->post( this, boost::bind(&Dispatcher::startTimer, this, timer ) );

		return timer->data.id;
	}
}

void Dispatcher::startTimer( struct impl::my_timer_ev *timer ) {
	ev_timer_start (_loop, &timer->ev);
	_timers.push_back( timer );
}

void Dispatcher::stopTimer( id::Ident &id ) {
	if (id::isValid(id)) {
		_tasks->post( this, boost::bind(&Dispatcher::stopTimerImpl, this, id->getID() ) );
		id.reset();
	}
}

void Dispatcher::stopTimerImpl( id::ID_TYPE id ) {
	STOP_EVENT( struct impl::my_timer_ev, _timers, id, timer );	
}

//	IO events
id::Ident Dispatcher::addIO( DESCRIPTOR fd, const Callback &callback ) {
	struct impl::my_io_ev *io = (struct impl::my_io_ev *)calloc( 1, sizeof(struct impl::my_io_ev) );
	if (!io) {
		LERROR("io::ev::Dispatcher", "cannot create io event");
		return id::Ident();
	}
	else {
		//	Setup
		INIT( io->data, callback );

		{
			struct ev_io *ev = &io->ev;
			ev_io_init( ev, impl::io_cb, fd, EV_READ );
		}

		//	Add to loop
		_tasks->post( this, boost::bind(&Dispatcher::startIO, this, io ) );

		return io->data.id;
	}
}
	
void Dispatcher::stopIO( id::Ident &id ) {
	if (id::isValid(id)) {
		_tasks->post( this, boost::bind(&Dispatcher::stopIOImpl, this, id->getID() ) );
		id.reset();
	}
}

void Dispatcher::startIO( struct impl::my_io_ev *io ) {
	ev_io_start (_loop, &io->ev);
	_io.push_back( io );
}

void Dispatcher::stopIOImpl( id::ID_TYPE id ) {
	STOP_EVENT( struct impl::my_io_ev, _io, id, io );
}

//	Socket events
id::Ident Dispatcher::addSocket( SOCKET sock, const Callback &callback ) {
	return addIO( sock, callback );
}

void Dispatcher::stopSocket( id::Ident &id ) {
	stopIO( id );
}

//	Aux
void Dispatcher::wakeup() {
	_tasks->runAll();
}

void Dispatcher::exitLoop() {
	ev_break( _loop );
}

}
}
}

