/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-util implementation.

    DTV-util is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-util is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-util.

    DTV-util es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-util se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "storage.h"
#include "../../../src/db/storage.h"
#include "generated/config.h"
#if STORAGE_USE_BDB
#	include "../../../src/db/bdb/storage.h"
#	define STORAGE util::storage::bdb::Storage
#endif
#include "../../../src/buffer.h"
#include <string>
#include <vector>

namespace impl {

static STORAGE *_storage = NULL;

static STORAGE* get() {
	if (!_storage) {
		_storage = new STORAGE("test");
		_storage->initialize();
	}
	return _storage;
}

static void destroyStorage() {
	assert(_storage);
	_storage->finalize();
	_storage->destroyDB();
	DEL(_storage);
}

static void resetStorage() {
	assert(_storage);
	_storage->finalize();
	DEL(_storage);
}

class Integer {
public:
	Integer() : _i(0) {};
	explicit Integer( int i ) : _i(i) {}

	int i() const { return _i; }
	void i( int i ) { _i = i; }

	template<class Storage>
	void serialize( Storage &storage ) {
		storage.serialize( "i", _i );
	}

private:
	int _i;
};

class Object {
public:
	Object( int valI, double valD, char valC, const std::string &valS, Integer integer )
		: _valI(valI), _valD(valD), _valC(valC), _valS(valS), _Integer(integer) {}


	int &valI() { return _valI; }
	double &valD() { return _valD; }
	char &valC() { return _valC; }
	std::string &valS() { return _valS; }
	Integer &integer() { return _Integer; }

private:
	int _valI;
	double _valD;
	char _valC;
	std::string _valS;
	Integer _Integer;
};

enum type {
	uno,
	dos
};

}

namespace util {
namespace storage {

template<>
void serialize( STORAGE &storage, impl::Object &obj ) {
	storage.serialize( "valI", obj.valI() );
	storage.serialize( "valD", obj.valD() );
	storage.serialize( "valC", obj.valC() );
	storage.serialize( "valS", obj.valS() );
	storage.serialize( "Integer", obj.integer() );
}

}
}

void StorageTest::TearDown( void ) {
	impl::destroyStorage();
}

TEST_F( StorageTest, save_and_load_string ) {
	std::string val = "value";
	impl::get()->save("keyS", val);
	val = "";
	impl::get()->load("keyS", val);

	ASSERT_TRUE( val == "value" );
}

TEST_F( StorageTest, save_and_load_int ) {
	int val = 10;
	impl::get()->save( "int", val );
	val = 0;
	impl::get()->load( "int", val );

	ASSERT_TRUE( val==10 );
}

TEST_F( StorageTest, save_and_load_true ) {
	bool val = true;
	impl::get()->save("keyB1", val);
	val = false;
	impl::get()->load("keyB1", val);

	ASSERT_TRUE( val );
}

TEST_F( StorageTest, save_and_load_false ) {
	bool val = false;
	impl::get()->save("keyB2", val);
	val = true;
	impl::get()->load("keyB2", val);

	ASSERT_FALSE( val );
}

TEST_F( StorageTest, load_no_stored_data ) {
	bool val = false;
	impl::get()->load("keyB3", val);

	ASSERT_FALSE( val );
}

TEST_F( StorageTest, load ) {
	std::string val = "none";
	impl::get()->load("keyStr", val);
	ASSERT_TRUE( val == "none" );

	val = "value";
	impl::get()->load("keyStr", val);
	ASSERT_TRUE( val == "none" );
}

TEST_F( StorageTest, save_two_values ) {
	std::string val = "none";
	
	impl::get()->save("key1", std::string("value"));
	impl::get()->load("key1", val);

	ASSERT_TRUE( val == "value" );

	impl::get()->save("key1", std::string("anotherValue"));
	impl::get()->load("key1", val);

	ASSERT_TRUE( val == "anotherValue" );
}

TEST_F( StorageTest, save_two_values2 ) {
	bool valB = false;
	std::string valS = "none";

	impl::get()->save("key2", true);
	impl::get()->load("key2", valB);

	ASSERT_TRUE( valB );

	impl::get()->save("key2", std::string("anotherValue"));
	impl::get()->load("key2", valS);

	ASSERT_TRUE( valS == "anotherValue" );
}

TEST_F( StorageTest, save_and_load_string_destroying_cache ) {
	std::string val = "none";

	impl::get()->save("string", std::string("value"));
	impl::resetStorage();
	impl::get()->load("string", val);

	ASSERT_TRUE( val == "value" );
}

TEST_F( StorageTest, save_and_load_int_destroying_cache ) {
	int val = 20;

	impl::get()->save("int", 10);
	impl::resetStorage();
	impl::get()->load("int", val);

	ASSERT_TRUE( val == 10 );
}

TEST_F( StorageTest, save_and_load_bool_destroying_cache ) {
	bool val = false;

	impl::get()->save("bool", true);
	impl::resetStorage();
	impl::get()->load("bool", val);

	ASSERT_TRUE( val );
}

TEST_F( StorageTest, save_and_load_vector) {
	int myints[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9};

	impl::get()->save("ints", std::vector<int>(myints, myints + sizeof(myints) / sizeof(int) ));
	std::vector<int> loadedVector;
	impl::get()->load("ints", loadedVector);
	for (int i=0; i<10; i++) {
		ASSERT_TRUE( loadedVector[i] == myints[i]);
	}
}

TEST_F( StorageTest, save_and_load_vector_destroying_cache ) {
	int myints[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9};

	impl::get()->save("ints", std::vector<int>(myints, myints + sizeof(myints) / sizeof(int) ));
	impl::resetStorage();
	std::vector<int> loadedVector;
	impl::get()->load("ints", loadedVector);
	ASSERT_TRUE(loadedVector.size()==10);
	for (int i=0; i<10; i++) {
		ASSERT_TRUE( loadedVector[i] == myints[i]);
	}
}

TEST_F( StorageTest, save_and_load_class) {
	impl::Integer integer(10);

	impl::get()->save("integer", integer);
	impl::get()->load("integer", integer);
	ASSERT_TRUE( integer.i() == 10 );
}

TEST_F( StorageTest, save_and_load_class_destroying_cache ) {
	impl::Integer integer(10);

	impl::get()->save("integer", integer);
	impl::resetStorage();
	impl::get()->load("integer", integer);
	ASSERT_TRUE( integer.i() == 10 );
}

TEST_F( StorageTest, save_and_load_vector_of_pointers) {
	std::vector<impl::Integer*> integers;
	for (int i=0; i<10; i++) {
		integers.push_back( new impl::Integer(i) );
	}

	impl::get()->save("integers", integers);
	std::vector<impl::Integer*> loadedVector;
	impl::get()->load("integers", loadedVector);
	ASSERT_TRUE(loadedVector.size()==10);
	for (int i=0; i<10; i++) {
		ASSERT_TRUE( loadedVector[i]->i() == i );
	}
}

TEST_F( StorageTest, save_and_load_vector_of_pointers_destroying_cache ) {
	std::vector<impl::Integer*> integers;
	for (int i=0; i<10; i++) {
		integers.push_back( new impl::Integer(i) );
	}

	impl::get()->save("integers", integers);
	impl::resetStorage();
	std::vector<impl::Integer*> loadedVector;
	impl::get()->load("integers", loadedVector);
	ASSERT_TRUE(loadedVector.size()==10);
	for (int i=0; i<10; i++) {
		ASSERT_TRUE( loadedVector[i]->i() == i );
	}
}

TEST_F( StorageTest, save_and_load_object ) {

	impl::Object obj(10, 20.5, 'a', "un string", impl::Integer(15));
	impl::get()->save( "obj", obj );

	impl::Object obj2(0, 0.0, '\0', "", impl::Integer(0));
	impl::get()->load( "obj", obj2);

	ASSERT_TRUE( obj2.valI() == 10 );
	ASSERT_TRUE( obj2.valD() == 20.5 );
	ASSERT_TRUE( obj2.valC() == 'a' );
	ASSERT_TRUE( obj2.valS() == "un string" );
	ASSERT_TRUE( obj2.integer().i() == 15 );
}

TEST_F( StorageTest, save_and_load_object_destroying_cache ) {

	impl::Object obj(10, 20.5, 'a', "un string", impl::Integer(15));
	impl::get()->save( "obj", obj );

	impl::resetStorage();

	impl::Object obj2(0, 0.0, '\0', "", impl::Integer(0));
	impl::get()->load( "obj", obj2);

	ASSERT_TRUE( obj2.valI() == 10 );
	ASSERT_TRUE( obj2.valD() == 20.5 );
	ASSERT_TRUE( obj2.valC() == 'a' );
	ASSERT_TRUE( obj2.valS() == "un string" );
	ASSERT_TRUE( obj2.integer().i() == 15 );
}

typedef int AnInt;

TEST_F( StorageTest, save_and_load_typedef ) {

	AnInt i = 10;
	impl::get()->save( "anInt", i );
	i=0;
	impl::get()->load( "anInt", i );

	ASSERT_TRUE( i == 10 );
}

TEST_F( StorageTest, save_and_load_typedef_destroying_cache ) {

	AnInt i = 10;
	impl::get()->save( "anInt", i );

	impl::resetStorage();

	i=0;
	impl::get()->load( "anInt", i );

	ASSERT_TRUE( i == 10 );
}

TEST_F( StorageTest, save_and_load_enum ) {

	impl::type t = impl::uno;
	impl::get()->save( "enumType", t );
	t=impl::dos;
	impl::get()->load( "enumType", t );

	ASSERT_TRUE( t == impl::uno );
}

TEST_F( StorageTest, save_and_load_enum_destroying_cache ) {

	impl::type t = impl::uno;
	impl::get()->save( "enumType", t );

	impl::resetStorage();

	t=impl::dos;
	impl::get()->load( "enumType", t );

	ASSERT_TRUE( t == impl::uno );
}
