/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "service.h"
#include "mixer.h"
#include "playermixer.h"
#include "../plugin/pluginmanager.h"
#include <util/log.h>

namespace zapper {
namespace audio {

Service::Service()
	: plugin::ServiceImpl<Service>( "Mixer" )
{
	_mixer = NULL;
	_volume = -1;
	_oldVolume = 0;
	_isMuted = false;
	_channels = audio::channel::LAST_CHANNEL;
}
	
Service::Service( Mixer *mixer )
	: plugin::ServiceImpl<Service>( "Mixer" )
{
	assert(mixer);
	_mixer = mixer;
	_volume = -1;
	_oldVolume = 0;
	_isMuted = false;
	_channels = audio::channel::LAST_CHANNEL;
}

Service::~Service()
{
	delete _mixer;
}

//	Volume methods
Volume Service::volume( Volume vol ) {
	return setVolume( vol );
}

Volume Service::setVolume( Volume vol ) {
	assert(_mixer);

	LDEBUG( "mixer::Service", "Set volume: vol=%d", vol );
	
	if (isMuted()) {
		toggleMute();
	}

	if (vol != _volume) {
		if (!_mixer->setVolume( vol )) {
			LWARN( "mixer::Service", "cannot set volume." );
		}
		_volume = vol;
		mgr()->save( "audio.volume", _volume );
	}
	return _volume;
}

Volume Service::volume() const {
	return isMuted() ? MIN_VOLUME : _volume;
}

bool Service::isMuted() const {
	return _isMuted;
}

Volume Service::volumeUp() {
	if (isMuted()) {
		toggleMute();
	}

	Volume old = volume();
	if (old < MAX_VOLUME) {
		old = (old + STEP_VOLUME > MAX_VOLUME) ? MAX_VOLUME : (old + STEP_VOLUME);
	}
	return volume( old );
}

Volume Service::volumeDown() {
	if (isMuted()) {
		toggleMute();
	}

	Volume old = volume();
	if (old > MIN_VOLUME) {
		old = (old - STEP_VOLUME < MIN_VOLUME) ? MIN_VOLUME : (old - STEP_VOLUME);
	}
	return volume( old );
}

bool Service::toggleMute() {
	return setMute( !_isMuted, true );
}

bool Service::mute( bool needMute ) {
	return setMute( needMute, true );
}

bool Service::setMute( bool needMute, bool save ) {
	assert(_mixer);
	
	LDEBUG( "mixer::Service", "Set mute: mute=%d", needMute );
	
	Volume old = volume();

	if (_mixer->hasMute()) {
		if (!_mixer->mute( needMute )) {
			LWARN( "mixer::Service", "Toggle mute fail." );
		}
	}
	else if (needMute) {
		setVolume( MIN_VOLUME );
		_oldVolume = old;
	}
	else {
		setVolume( _oldVolume );
	}

	_isMuted = needMute;
	if (save) {
		mgr()->save( "audio.muted", _isMuted );
	}
	
	return _isMuted;
}

//	Audio channels
std::vector<audio::channel::type> Service::supportedAudioChannels() const {
	assert(_mixer);
	return _mixer->supportedAudioChannels();
}

void Service::setAudioChannel( audio::channel::type chs ) {
	assert(_mixer);	
	if (_channels != chs) {
		//	Is audio channels valid?
		std::vector<audio::channel::type> channels = _mixer->supportedAudioChannels();
		std::vector<audio::channel::type>::const_iterator it=std::find( channels.begin(), channels.end(), chs );
		if (it != channels.end()) {
			//	Get audio channels info
			audio::Channel &chInfo = audio::getChannelInfo( chs );
			LDEBUG( "mixer::Service", "Set audio channel: channel=%d, name=%s, count=%d, lfe=%d",
				chs, chInfo.name, chInfo.count, chInfo.lfe );

			if (_mixer->setAudioChannel(chs)) {
				_channels = chs;
				mgr()->save( "audio.audioChannel", _channels );
			}
			else {
				LWARN( "mixer::Service", "Cannot set audio channel: %d", chs );
			}
		}
		else {
			LWARN( "mixer::Service", "Audio channel not supported: %d", chs );
		}
	}
}

audio::channel::type Service::getCurrentAudioChannel() const {
	return _channels;
}

//	Service virtual methdos
void Service::standby() {
	setMute( true, false );
}

void Service::wakeup() {
	loadMute();
}

void Service::onAttach() {
	//	If no mixer, add channelservice as dependency
	if (!_mixer) {
		addSimpleDependency( "ChannelService" );
	}
}

bool Service::onOnline() {
	if (!_mixer) {
		LINFO( "mixer::Service", "Using mixer from player" );
		_mixer = new PlayerMixer( mgr() );
	}

	return _mixer->initialize();
}

void Service::onOffline() {
	assert(_mixer);	
	return _mixer->finalize();	
}

void Service::onLoadConfig() {
	assert(_mixer);

	{
		Volume vol = DEFAULT_VOLUME;
		mgr()->load( "audio.volume", vol );
		setVolume(vol);
	}

	loadMute();

	{
		audio::channel::type chs = _mixer->defaultAudioChannel();
		mgr()->load( "audio.audioChannel", chs );
		setAudioChannel(chs);
	}
}

void Service::onResetConfig() {
	assert(_mixer);
	
	volume( DEFAULT_VOLUME );
	if (isMuted()) {
		toggleMute();
	}
	setAudioChannel( _mixer->defaultAudioChannel() );
}

void Service::loadMute() {
	bool muted = false;
	mgr()->load( "audio.muted", muted );
	setMute(muted,false);
}

}
}

