/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "channel.h"
#include "show.h"
#include "channelmanager.h"
#include <mpegparser/demuxer/psi/eit.h>
#include <mpegparser/service/service.h>
#include <mpegparser/time.h>
#include <util/string.h>
#include <util/url.h>
#include <util/cfg/cfg.h>
#include <util/log.h>

namespace zapper {
namespace channel {

using namespace boost::gregorian;

namespace impl {
	static Show defaultShow;
}

Channel::Channel()
{
	_network = -1;
	_channelID = -1;
	_type  = 0;
	_parentalAge = parental::age::none;
	_parentalContent = parental::content::none;
	_isFavorite = false;
	_isBlocked = false;
	_mgr = NULL;
	_present = NULL;
	_following = NULL;
}

Channel::Channel( tuner::Service *srv )
	: _network(srv->networkID()), _id( srv->nitID(), srv->tsID(), srv->id() ),
	  _name( srv->name() ),
	  _type( srv->type() ),
	  _channel( srv->logicalChannel() )
{
	_channelID = -1;
	_isFavorite = false;
	_isBlocked = false;
	_mgr = NULL;
	_present = NULL;
	_following = NULL;	

	{	//	Get parental rating
		const std::string &code = util::cfg::getValue<std::string>("zapper.countryCode");
		parental::parse( srv->descriptors(), code, _parentalAge, _parentalContent );
	}
}

Channel::~Channel()
{
	delete _present;
	delete _following;
	for (int i=0; i<MAX_EIT_TABLES; i++) {
		CLEAN_ALL( Show *, _shows[i] );
	}
}

//	Attach to channel manager
void Channel::attach( ChannelManager *mgr ) {
	assert(mgr);
	_mgr = mgr;
}

//	Channel ID
int Channel::channelID() const {
	return _channelID;
}

void Channel::channelID( int channelID ) {
	_channelID = channelID;
}

void Channel::toggleFavorite() {
	assert(_mgr);
	_isFavorite = !_isFavorite;
	_mgr->onUpdateChannel( this, true );
}

bool Channel::isFavorite() const {
	return _isFavorite;
}

void Channel::toggleBlocked() {
	assert(_mgr);
	_isBlocked = !_isBlocked;
	_mgr->onUpdateChannel( this, true );
}

void Channel::unblock() {
	_isBlocked = false;
}

bool Channel::isBlocked() const {
	return _isBlocked;
}

const ServiceID &Channel::id() const {
	return _id;
}

//	Player info
const std::string Channel::getParameters() const {
	util::Url url;
	url.type( "service" );
	url.addParam( "network", _network );
	url.addParam( "nitID", _id.nitID() );
	url.addParam( "tsID", _id.tsID() );
	url.addParam( "srvID", _id.programID() );	
	return url.id();
}

//	Shows
void Channel::process( tuner::Eit *eit ) {
	eit->show();

	const tuner::eit::Events &events = eit->events();
	if (eit->isPresentFollowing()) {
		//	Update p/f information (current and next show)
		DEL(_present);
		if (events.size() > 0 && events[0]) {
			_present = new Show( events[0] );
		}

		DEL(_following);
		if (events.size() > 1 && events[1]) {
			_following = new Show( events[1] );
		}

		//	Notify that current show changed!
		_mgr->onUpdateChannel( this, false );
	}
	else if (PSI_TID_EIT_SBF_ACTUAL <= eit->tableID() && eit->tableID() <= PSI_TID_EIT_SBL_ACTUAL ) {
		//	Process schedule information
		int id = eit->tableID() - PSI_TID_EIT_SBF_ACTUAL;

		//	Clean old events
		CLEAN_ALL( Show *, _shows[id] );

		//	Add new events
		BOOST_FOREACH(tuner::eit::Event *evt, events) {
			if (evt) {
				_shows[id].push_back(new Show(evt));
			}
		}
	}
	else {
		LWARN( "Channel", "EIT table out of range" );
	}
}

Show *Channel::getCurrentShow() const {
	LDEBUG( "Channel", "Get current show" );
	tuner::pt::ptime now;
	tuner::time::clock.get( now );
	const bpt::time_period period( now, bpt::seconds(1) );
	Shows progs;
	getShowsBetween( progs, period );
	return progs.empty() ? &impl::defaultShow : progs[0];
}

void Channel::addShow( Shows &shows, Show *show, const bpt::time_period &period ) const {
	assert( show );
	if (show->timeIntersects(period)) {
		Shows::const_iterator it = std::find_if(
			shows.begin(),
			shows.end(),
			boost::bind(&Show::eventID,_1) == show->eventID()
		);
		if (it == shows.end()) {
			show->show();
			show->lock( isProtected(show) );
			shows.push_back( show );
		}
	}
}

static std::string asString( const bpt::ptime &p ) {
	return util::format( "%d/%d %d:%d:%d",
		p.date().day().as_number(),
		p.date().month().as_number(),
		p.time_of_day().hours(),
		p.time_of_day().minutes(),
		p.time_of_day().seconds()
	);
}

void Channel::getShowsBetween( Shows &shows, const bpt::time_period &period ) const {
	//	Add present ...
	if (_present) {
		addShow( shows, _present, period );
	}

	//	Add schedule events (TODO: optimize)
	for (int i=0; i<MAX_EIT_TABLES; i++) {
		BOOST_FOREACH( Show *show, _shows[i] ) {
			addShow( shows, show, period );
		}
	}

	LDEBUG( "Channel", "Get shows: period(%s,%s), shows=%d",
		asString(period.begin()).c_str(), asString(period.end()).c_str(), shows.size() );
}

void Channel::show() const {
	LDEBUG( "Channel", "id=(%04x,%04x,%04x), channelID=%d, network=%d, name=%s, age=%02x, content=%02x",
		_id.nitID(), _id.tsID(), _id.programID(), _channelID, _network, _name.c_str(), _parentalAge, _parentalContent );
}

const std::string &Channel::name() const {
	return _name;
}

size_t Channel::network() const {
	return _network;
}

bool Channel::isProtected() const {
	assert(_mgr);
	return isProtected( getCurrentShow() );
}

bool Channel::isProtected( Show *show ) const {
	util::BYTE curContent=parental::getContent( _parentalContent, show->parentalContent() );
	parental::age::type curAge=parental::getAge( _parentalAge, show->parentalAge() );
	return _mgr->isProtected( isBlocked(), curContent, curAge );
}

parental::age::type Channel::parentalAge() const {
	return _parentalAge;
}

util::BYTE Channel::parentalContent() const {
	return _parentalContent;
}

bool Channel::isOneSeg() const {
	return type() == tuner::service::type::one_seg;
}

tuner::service::Type Channel::type() const {
	return _type;
}

const std::string Channel::channel() const {
	return _channel;
}

//	Update info
void Channel::update( Channel *ch ) {
	bool persist = false;
	
	if (_name != ch->_name) {
		_name = ch->_name;
		persist = true;
	}

	if (_type != ch->_type) {
		_type = ch->_type;
		persist = true;
	}

	if (_channel != ch->_channel) {
		_channel = ch->_channel;
		persist = true;
	}
	
	_parentalAge = ch->_parentalAge;
	_parentalContent = ch->_parentalContent;;
	delete ch;
	
	_mgr->onUpdateChannel( this, persist );	
}

}
}

