/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "channelmanager.h"
#include "channel.h"
#include "tunerlistener.h"
#include "../tuner/tuner.h"
#include "../control/service.h"
#include "../plugin/pluginmanager.h"
#include <mpegparser/service/extension/epg/epgextension.h>
#include <mpegparser/demuxer/psi/eit.h>
#include <mpegparser/time.h>
#include <util/cfg/cfg.h>
#include <util/log.h>
#include <util/mcr.h>

namespace zapper {
namespace channel {

static void myTimeCallback( plugin::PluginManager *mgr, const std::string &name, bool isValid ) {
	mgr->runMethod<ChannelManager>( name, boost::bind(&ChannelManager::onTimeValid,_1,isValid) );
}

ChannelManager::ChannelManager( bool useEPG )
	: plugin::ServiceImpl<ChannelManager>( "ChannelManager" ), _withEPG(useEPG)
{
	_channelID = 0;
	_showOneSeg = util::cfg::getValue<bool>("zapper.showOneSeg");
	_control = NULL;
}

ChannelManager::~ChannelManager()
{
	CLEAN_ALL(Channel *,_channels);
}

//	Scan
void ChannelManager::onBeginScan() {
	//	Invalidate list of channels
	_channelID=0;
	removeAll();
	mgr()->save( "channelmanager.channels", _channels );
}

struct FindChannel {
	FindChannel( Channel *ch ) : _toFind(ch) {}
	
	bool operator()( Channel *ch ) const {
		return _toFind->id() == ch->id() && _toFind->network() == ch->network();
	}
	
	Channel *_toFind;
};
	
void ChannelManager::onChannel( Channel *ch, bool inScan ) {
	assert(ch);

	//	Check if ch already exist
	Channels::const_iterator it=std::find_if(
		_channels.begin(),
		_channels.end(),
		FindChannel( ch )
	);
	bool found=it != _channels.end();
	if (inScan && !found) {	//	Not exist? in scan? add channel ...
		//	Attach channel
		ch->attach( this );
		ch->channelID( _channelID++ );

		LINFO( "ChannelManager", "Add channel: name=%s", ch->name().c_str() );
		_channels.push_back( ch );
		mgr()->save( "channelmanager.channels", _channels );

		//	Notify a new channel found
		if (isVisible(ch)) {
			_onNewChannel( ch );
		}
	}
	else if (found) {
		LDEBUG( "ChannelManager", "Update channel: name=%s", ch->name().c_str() );
		//	Try update channel
		(*it)->update( ch );
	}
}

void ChannelManager::onUpdateChannel( Channel *ch, bool needPersist ) {
	if (needPersist) {
		mgr()->save( "channelmanager.channels", _channels );
	}
	_onChannelChanged( ch );
}

void ChannelManager::onTimeValid( bool isValid ) {
	LDEBUG( "ChannelManager", "Time is valid: isValid=%d", isValid );
	BOOST_FOREACH( Channel *ch, _channels ) {
		onUpdateChannel( ch, false );
	}	
}

bool ChannelManager::isProtected( bool isBlocked, util::BYTE pContent, parental::age::type pAge ) const {
	//	Check if user is authenticated
	if (!_control->sessionIsOpen()) {
		return isBlocked || parental::isProtected(
			pContent,
			_control->parentalContent(),
			pAge,
			_control->parentalAge() );
	}
	return false;
}

//	Channel modification
OnNewChannelSignal &ChannelManager::onNewChannel() {
	return _onNewChannel;
}

OnChannelRemovedSignal &ChannelManager::onChannelRemoved() {
	return _onChannelRemoved;
}

OnChangedSignal &ChannelManager::onChannelChanged() {
	return _onChannelChanged;
}

void ChannelManager::removeAll() {
	for (size_t i=0; i<_channels.size(); i++) {
		erase( _channels[i], true );
	}
	_channels.clear();
}

bool ChannelManager::remove( int channelID ) {
	bool result=false;
	Channels::iterator it=std::find_if(
		_channels.begin(), _channels.end(), boost::bind(&Channel::channelID,_1) == channelID );
	if (it != _channels.end()) {
		Channel *ch=(*it);
		_channels.erase( it );
		erase( ch, false );
		mgr()->save( "channelmanager.channels", _channels );
		result=true;
	}
	return result;
}

void ChannelManager::erase( Channel *ch, bool removeAll ) {
	LDEBUG( "ChannelManager", "Remove channel: channelID=%d, name=%s",
		ch->channelID(), ch->name().c_str() );
	_onChannelRemoved( ch, removeAll );
	delete ch;
}

void ChannelManager::unblockAll() {
	BOOST_FOREACH( Channel *ch, _channels ) {
		ch->unblock();
	}
	mgr()->save( "channelmanager.channels", _channels );
}

//	Iterators
bool ChannelManager::showOneSeg() const {
	return _showOneSeg;
}

void ChannelManager::showOneSeg( bool show ) {
	_showOneSeg = show;
	mgr()->save( "channelmanager.showOneSeg", _showOneSeg );	
}

bool ChannelManager::toggleOneSeg() {
	_showOneSeg = !_showOneSeg;
	mgr()->save( "channelmanager.showOneSeg", _showOneSeg );
	return _showOneSeg;
}

int ChannelManager::nextChannel( int first, int factor ) const {
	return nextChannel( first, factor, WithoutProtectedIterator() );
}

int ChannelManager::nextFavorite( int first, int factor ) const {
	return nextChannel( first, factor, FavoriteIterator() );
}

//	Getters
void ChannelManager::getAll( Channels &channels ) const {
	channels.clear();
	for (size_t i=0; i<_channels.size(); ++i) {
		Channel *ch=_channels[i];
		if (isVisible( ch )) {
			channels.push_back( ch );
		}
	}
}

bool ChannelManager::haveChannels() const {
	return _channels.empty() ? false : true;
}

int ChannelManager::channelCount() const {
	return _channels.size();
}

Channel *ChannelManager::get( int chID ) const {
	Channel *ptr=NULL;
	Channels::const_iterator it=std::find_if(
		_channels.begin(), _channels.end(), boost::bind(&Channel::channelID,_1) == chID );
	if (it != _channels.end()) {
		ptr=(*it);
	}
	return ptr;
}

int ChannelManager::getID( Channel *ch ) const {
	return ch->channelID();
}

int ChannelManager::getIndex( int chID ) const {
	int result=-1;

	//	Channel is valid?
	if (chID != INVALID_CHANNEL) {
		//	Find channelID in list
		for (size_t p=0; p<_channels.size(); p++) {
			Channel *ch=_channels[p];
			if (isVisible(ch) && ch->channelID() == chID) {
				result=(int)p;
			}
		}
	}
	if (result < 0) {
		result=0;
	}
	return result;
}

bool ChannelManager::isVisible( Channel *ch ) const {
	return !(!_showOneSeg && ch->isOneSeg());
}

//	EPG methods
void ChannelManager::processEit( const boost::shared_ptr<tuner::Eit> &eit ) {
	ServiceID id( eit->nitID(), eit->tsID(), eit->serviceID() );

	//	Find channel
	Channels::const_iterator it=std::find_if(
		_channels.begin(), _channels.end(), boost::bind(&Channel::id,_1) == id );
	if (it != _channels.end()) {
		//	Process eit
		(*it)->process( eit.get() );
	}
	else {
		eit->show();
		
		LWARN( "ChannelManager", "Cannot update show database. Channel not found: id=(%04x,%04x,%04x)",
			id.nitID(), id.tsID(), id.programID() );
	}
}

//	Aux control
void ChannelManager::onControlEnabled( bool isEnabled ) {
	LDEBUG( "ChannelManager", "Auth control enabled changed: isEnabled=%d", isEnabled );
	if (!isEnabled) {
		unblockAll();
	}
}

//	Plugin methods
void ChannelManager::onAttach() {
	//	Add service dependencies
	std::vector<std::string> tuners=mgr()->findInstances( "Tuner" );
	BOOST_FOREACH( const std::string &tun, tuners ) {
		addSimpleDependency( tun );
	}
	addSimpleDependency( "Control" );
}

bool ChannelManager::onOnline() {
	std::vector<std::string> tuners=mgr()->findInstances( "Tuner" );
	BOOST_FOREACH( const std::string &tun, tuners ) {
		LDEBUG( "ChannelManager", "Registrate into tuner: %s", tun.c_str() );

		//	Get tuner service
		Tuner *tuner = mgr()->findService<Tuner>( tun );
		assert(tuner);

		//	Add tuner listener
		tuner->addListener( new TunerListener(this) );

		if (_withEPG) {	//	Add EPG extension
			//	Create EPG extension
			tuner::EPGExtension *extension = new tuner::EPGExtension();
			extension->onEitParsed( boost::bind( &ChannelManager::processEit, this, _1 ) );

			//	Attach extension to tuner
			Tuner::attach( mgr(), extension, tun );
		}
	}

	//	Get control service
	_control = mgr()->addRef<control::Service>( "Control" );
	assert(_control);
	_control->onEnabledChanged().connect( boost::bind(&ChannelManager::onControlEnabled,this,_1) );

	//	Get time changes
	tuner::time::clock.onTimeValidChanged().connect( boost::bind(&myTimeCallback,mgr(),name(),_1) );	

	return true;
}

void ChannelManager::onOffline() {
	LDEBUG( "ChannelManager", "On Offline" );
	RELEASE_SERVICE(_control);
}

void ChannelManager::onLoadConfig() {
	//	Load variables ...
	_showOneSeg = util::cfg::getValue<bool>("zapper.showOneSeg");
	mgr()->load( "channelmanager.showOneSeg", _showOneSeg );

	//	Load scanned channels
	mgr()->load( "channelmanager.channels", _channels );

	//	Attach to manager
	BOOST_FOREACH( Channel *ch, _channels ) {
		ch->attach( this );
	}

	LDEBUG( "ChannelManager", "Load config: showOneSeg=%d, channels=%d",
		_showOneSeg, _channels.size() );
}

void ChannelManager::onResetConfig() {
	_showOneSeg = util::cfg::getValue<bool>("zapper.showOneSeg");
	mgr()->save( "channelmanager.showOneSeg", _showOneSeg );
}

}
}

