/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "channelplayer.h"
#include "../channel/channelmanager.h"
#include "../channel/channel.h"
#include "../control/service.h"
#include "../player/mediaplayer.h"
#include "../player/playerservice.h"
#include "../plugin/pluginmanager.h"
#include <util/mcr.h>
#include <util/log.h>

namespace zapper {
namespace channel {

ChannelPlayer::ChannelPlayer()
{
	_ch      = NULL;
	_chs     = NULL;
	_control = NULL;
	_player  = NULL;
	_blocked = false;
}

ChannelPlayer::~ChannelPlayer()
{
	assert(!_ch);
	assert(!_player);
	assert(!_chs);
	assert(!_control);	
}

//	Change channel
OnChangeChannelSignal &ChannelPlayer::onChangeChannel() {
	return _onChange;
}

OnChannelBlocked &ChannelPlayer::onChannelBlocked() {
	return _onBlocked;
}

int ChannelPlayer::change( int channel ) {
	if (channel != current()) {
		//	Stop current channel
		stopCurrent();
		
		//	Get channel from channel manager
		Channel *ch = _chs->get( channel );
		if (ch) {
			LDEBUG( "ChannelPlayer", "Change to channel: chID=%d", channel );
			if (startPlayer( ch )) {
				//	Save channel
				_ch = ch;

				//	Notify
				_onChange( channel, ch );
			}
		}

		//	Always save current!
		_chs->mgr()->save( "channelplayer.current", current() );
	}
	return current();
}

int ChannelPlayer::current() const {
	return _ch ? _ch->channelID() : INVALID_CHANNEL;
}

//	Player methods
void ChannelPlayer::setFullScreen() {
	_player->setFullScreen();
}

void ChannelPlayer::resize( int x, int y, int w, int h ) {
	_player->resize( x, y, w, h );
}

void ChannelPlayer::mute( bool needMute ) {
	_player->mute( needMute );
}

void ChannelPlayer::volume( audio::Volume vol ) {
	_player->volume( vol );
}

audio::channel::type ChannelPlayer::getAudio() const {
	return _player->getAudio();
}

display::aspect::type ChannelPlayer::getAspect() const {
	return _player->getAspect();
}

display::mode::type ChannelPlayer::getResolution() const {
	return _player->getResolution();
}

//	Service notifications
void ChannelPlayer::onStart( plugin::PluginManager *mgr ) {
	//	Get manager
	_chs = mgr->addRef<channel::ChannelManager>( "ChannelManager" );
	assert(_chs);

	//	Get control service
	_control = mgr->addRef<control::Service>( "Control" );
	assert(_control);
	_control->onSessionChanged().connect( boost::bind(&ChannelPlayer::onSessionStarted,this) );

	//	Add remove listener
	_chs->onChannelRemoved().connect( boost::bind(&ChannelPlayer::onRemoved,this,_1,_2) );
	_chs->onChannelChanged().connect( boost::bind(&ChannelPlayer::onChanged,this,_1) );

	//	Create player (TODO: need be for each channel)
	_player = mgr->findService<player::PlayerService>( "Player" )->create( player::type::service );
	_player->setFullScreen();
	assert(_player);

	//	Load last channel started
	int restored = INVALID_CHANNEL;
	mgr->load( "channelplayer.current", restored );

	//	Play restored channel if valid
	if (restored != INVALID_CHANNEL) {
		//	Start channel
		change( restored );
	}
}

void ChannelPlayer::onStop( plugin::PluginManager *mgr ) {
	stopCurrent();
	mgr->findService<player::PlayerService>( "Player" )->destroy(_player);
	_player=NULL;
	RELEASE_SERVICE(_chs);
	RELEASE_SERVICE(_control);	
}

void ChannelPlayer::onStandby( plugin::PluginManager * /*mgr*/ ) {
	stopCurrent();
}

void ChannelPlayer::onWakeup( plugin::PluginManager *mgr ) {
	//	Load last channel started
	int restored = INVALID_CHANNEL;
	mgr->load( "channelplayer.current", restored );

	//	Play restored channel if valid
	if (restored != INVALID_CHANNEL) {
		//	Start channel
		change( restored );
	}	
}

void ChannelPlayer::restartPlayer( int chID ) {
	LDEBUG( "ChannelPlayer", "Restart player" );		
	_player->stop();
	Channel *ch = _chs->get( chID );
	if (ch) {
		startPlayer( ch );
	}
}

bool ChannelPlayer::startPlayer( Channel *ch ) {
	bool result;

	LDEBUG( "ChannelPlayer", "Start player" );
	
	//	Can play protected?
	bool canPlay = ch->isProtected() ? _control->sessionIsOpen() : true;
	if (!canPlay) {
		result = true;
		block( true );				
	}
	else {
		//	Start player
		result = _player->play(ch->getParameters());

		block( false );
	}

	return result;
}

void ChannelPlayer::stopCurrent() {
	LDEBUG( "ChannelPlayer", "Stop current channel" );
	if (_ch) {
		_onChange( current(), NULL );
		_player->stop();
		_ch = NULL;
		block( false );
	}
}

void ChannelPlayer::onRemoved( Channel *ch, bool removeAll ) {
	LDEBUG( "ChannelPlayer", "Channel(s) removed" );
	
	if (removeAll) {
		//	Stop current
		stopCurrent();
		_chs->mgr()->save( "channelplayer.current", INVALID_CHANNEL );
	}
	else if (ch->channelID() == current()) {	//	Is my channel removed?
		//	Get next channel not blocked
		int next=_chs->nextChannel( current(), 1 );
		change( next );
	}
}

void ChannelPlayer::onChanged( Channel *ch ) {
	int cur = current();
	if (ch->channelID() == cur) {
		LDEBUG( "ChannelPlayer", "Current channel changed: blocked=%d, protected=%d",
			_blocked, ch->isProtected() );
		
		if (ch->isProtected() != _blocked) {
			restartPlayer( cur );
		}
	}
}

void ChannelPlayer::onSessionStarted() {
	int cur = current();
	if (_blocked && cur != INVALID_CHANNEL) {
		restartPlayer( cur );
	}
}

void ChannelPlayer::block( bool value ) {
	if (_blocked != value) {
		_blocked = value;
		LINFO( "ChannelPlayer", "Channel status changed: blocked=%d", _blocked );
		_onBlocked( _blocked );
	}
}

bool ChannelPlayer::isBlocked() const {
	return _blocked;
}


int ChannelPlayer::nextVideo() {
	return _player->nextVideo();
}

int ChannelPlayer::nextAudio() {
	return _player->nextAudio();
}

int ChannelPlayer::nextSubtitle() {
	return _player->nextSubtitle();
}

int ChannelPlayer::subtitleCount() {
	return _player->subtitleCount();
}

}
}

