/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "displayservice.h"
#include "layer.h"
#include "zapperlayer.h"
#include "../plugin/pluginmanager.h"
#include "generated/config.h"
#include <canvas/system.h>
#include <util/log.h>
#include <boost/lexical_cast.hpp>
#include <boost/math/special_functions/round.hpp>
#include <algorithm>

namespace zapper {
namespace display {

static std::string getDisplayName( connector::type conn ) {
	std::string tmp="zapper.";
	tmp += boost::lexical_cast<std::string>(conn);
	tmp += ".displayMode";
	return tmp;
}

DisplayService::DisplayService( canvas::System *sys )
	: plugin::ServiceImpl<DisplayService>( "Display" )
{
	_sys          = sys;
	_transparency = -1;
	_connector    = connector::LAST_CONNECTOR;
	_aspect       = aspect::LAST_ASPECT;
	_modulator    = modulator::LAST_MODULATOR;
	for (int i=0; i<connector::LAST_CONNECTOR; i++) {
		_mode[i] = mode::LAST_MODE;
	}
}

DisplayService::~DisplayService( void )
{
}

//	Connector aux
bool DisplayService::setActive( connector::type conn ) {
	bool result=false;
	if (conn != _connector) {
		mgr()->save( "zapper.activeConnector", conn );
		_connector=conn;
		result=true;
	}
	return result;
}

std::vector<connector::type> DisplayService::supportedConnectors() const {
	std::vector<connector::type> tmp;
	tmp.push_back( connector::vga );
	tmp.push_back( connector::composite );
	tmp.push_back( connector::svideo );
	tmp.push_back( connector::component );
	tmp.push_back( connector::dvi );
	tmp.push_back( connector::hdmi );
	return tmp;
}

connector::type DisplayService::getDefaultActiveConnector() const {
	return connector::composite;
}

connector::type DisplayService::getActiveConnector() const {
	return _connector;
}

bool DisplayService::connectorIsValid( connector::type conn ) const {
	bool result=false;

	if (conn < connector::LAST_CONNECTOR) {
		std::vector<connector::type> valid=supportedConnectors();
		std::vector<connector::type>::const_iterator it=std::find(
			valid.begin(), valid.end(), conn );
		result=(it != valid.end());
	}

	if (!result) {
		LWARN( "DisplayService", "Video display connector not supported: %d", conn );
	}
	return result;
}

//	Video modes
std::vector<mode::type> DisplayService::supportedModes( connector::type conn ) const {
	std::vector<mode::type> tmp;

	switch (conn) {
		case connector::composite:
		case connector::svideo: {
			tmp.push_back( mode::m576i_50 );
			tmp.push_back( mode::mNTSC );
			break;
		}
		case connector::vga:
		case connector::component:
		case connector::dvi:
		case connector::hdmi: {
			tmp.push_back( mode::m576i_50 );
			tmp.push_back( mode::m576p_50 );

			tmp.push_back( mode::m720p_50 );
			tmp.push_back( mode::m720p_60 );

			tmp.push_back( mode::m720i_50 );
			tmp.push_back( mode::m720i_60 );

			tmp.push_back( mode::m1080i_25 );
			tmp.push_back( mode::m1080i_30 );
			tmp.push_back( mode::m1080i_50 );
			tmp.push_back( mode::m1080i_60 );

			tmp.push_back( mode::m1080p_24 );
			tmp.push_back( mode::m1080p_25 );
			tmp.push_back( mode::m1080p_30 );
			tmp.push_back( mode::m1080p_50 );
			tmp.push_back( mode::m1080p_60 );
			break;
		}
		default: {
			LWARN( "DisplayService", "Invalid type of connector: conn=%02x", conn );
		}
	};
	return tmp;
}

mode::type DisplayService::getCurrentVideoMode( connector::type conn ) const {
	if (conn >= connector::LAST_CONNECTOR) {
		conn = connector::unknown;
	}
	return _mode[conn];
}

bool DisplayService::setMode( connector::type conn, mode::type vMode, bool doActive/*=true*/ ) {
	if (!connectorIsValid( conn )) {
		LWARN( "DisplayService", "Output connector not valid: conn=%d", conn );
		return false;
	}

	bool result=false;
	if (vMode != _mode[conn]) {
		std::vector<mode::type> modes = supportedModes( conn );
		std::vector<mode::type>::const_iterator it=std::find( modes.begin(), modes.end(), vMode );
		if (it != modes.end()) {
			Mode mode = getMode( vMode );

			LDEBUG( "DisplayService", "Set video display mode: conn=%d, mode=%d, name=%s, type=%d, width=%d, height=%d, freq=%d",
				conn, vMode, mode.name, mode.type,
				mode.width, mode.height, mode.frequency );

			if (setVideoMode( conn, vMode, doActive )) {
				_mode[conn] = vMode;
				mgr()->save( getDisplayName(conn), _mode[conn] );
				if (doActive) {
					setActive( conn );
					_onModeChanged( vMode );
				}
				result=true;
			}
			else {
				LWARN( "DisplayService", "Cannot set video display mode: %d", vMode );
			}
		}
		else {
			LWARN( "DisplayService", "Video display mode not supported: %d", vMode );
		}
	}
	else if (doActive && conn != getActiveConnector()) {
		LDEBUG( "DisplayService", "Set video display mode: conn=%d, mode=%d",
			conn, vMode );
		
		//	Do output active
		setActive( conn );

		if (!supportMultipleModes()) {
			//	Force mode
			result=setVideoMode( conn, vMode, true );
		}
		else {
			result=true;
		}

		if (result) {
			_onModeChanged( vMode );
		}
	}
	
	return result;
}

mode::type DisplayService::defaultVideoMode( connector::type /*conn*/ ) const {
	return mode::m576i_50;
}

bool DisplayService::supportMultipleModes() const {
	return true;
}

bool DisplayService::setVideoMode( connector::type /*conn*/, mode::type /*mode*/, bool /*doActive*/ ) {
	//	Do nothing ...
	return true;
}

DisplayService::ModeChangedSignal &DisplayService::onModeChanged( void ) {
	return _onModeChanged;
}

//	Video aspect ratio
std::vector<aspect::type> DisplayService::supportedAspects() const {
	std::vector<aspect::type> tmp;
	tmp.push_back( aspect::automatic );
	return tmp;
}

bool DisplayService::setAspect( aspect::type vAspect ) {
	LDEBUG( "DisplayService", "Set video display aspect ratio: %d", vAspect );
	if (vAspect != _aspect) {
		std::vector<aspect::type> aspects = supportedAspects();
		std::vector<aspect::type>::const_iterator it=std::find( aspects.begin(), aspects.end(), vAspect );
		if (it != aspects.end()) {
			if (setVideoAspect( vAspect )) {
				_onAspectChanged( vAspect );
				_aspect = vAspect;
				mgr()->save( "zapper.displayAspect", _aspect );
				return true;
			}
			else {
				LWARN( "DisplayService", "Cannot set video aspect mode: %d", vAspect );
			}
		}
		else {
			LWARN( "DisplayService", "Video aspect mode not supported: %d", vAspect );
		}
	}
	return false;
}

aspect::type DisplayService::getCurrentAspect( void ) const {
	return _aspect;
}

aspect::type DisplayService::defaultVideoAspect() const {
	return aspect::automatic;
}

bool DisplayService::setVideoAspect( aspect::type /*vAspect*/ ) {
	//	Do nothing
	return true;
}

DisplayService::AspectChangedSignal &DisplayService::onAspectChanged() {
	return _onAspectChanged;
}

//	Video modulator
std::vector<modulator::type> DisplayService::supportedModulators() const {
	std::vector<modulator::type> tmp;
	tmp.push_back( modulator::mNone );
	return tmp;
}

bool DisplayService::setModulator( modulator::type mod ) {
	LDEBUG( "DisplayService", "Set video modulator: %d", mod );
	if (mod != _modulator) {
		std::vector<modulator::type> modulators = supportedModulators();
		std::vector<modulator::type>::const_iterator it=std::find( modulators.begin(), modulators.end(), mod );
		if (it != modulators.end()) {
			if (setVideoModulator( mod )) {
				_modulator = mod;
				mgr()->save( "zapper.displayModulator", _modulator );
				return true;
			}
			else {
				LWARN( "DisplayService", "Cannot set modulator mode: %d", mod );
			}
		}
		else {
			LWARN( "DisplayService", "Video modulator not supported: %d", mod );
		}
	}
	return false;
}

modulator::type DisplayService::getCurrentModulator( void ) const {
	return _modulator;
}

modulator::type DisplayService::defaultVideoModulator() const {
	return modulator::mNone;
}

bool DisplayService::setVideoModulator( modulator::type /*mod*/ ) {
	//	Do nothing
	return true;
}


//	Transparency
bool DisplayService::setTransparency( int value ) {
	LDEBUG( "DisplayService", "Set zapper layer transparency: value=%d", value );

	if (_transparency != value) {
		if(value<ZAPPER_MIN_TRANSPARENCY || ZAPPER_MAX_TRANSPARENCY<value) {
			LWARN( "DisplayService", "Transparency value out of range: value=%d", value );
			return false;
		}

		if (setDisplayTransparency( value )) {
			_transparency = value;
			mgr()->save( "zapper.transparency", value );
			return true;
		}
		else {
			LWARN( "DisplayService", "Error in setup zapper layer transparency: value=%d", value );
		}
	}

	return false;
}

//	Transparency aux
bool DisplayService::setDisplayTransparency( int value ) {
	int range = 255 - ZAPPER_FIRST_TRANSPARENCY;
	int alpha = boost::math::iround( ((value * range) / ZAPPER_MAX_TRANSPARENCY) + ZAPPER_FIRST_TRANSPARENCY );
	return getLayer( layer::zapper )->setAlpha( alpha );
}

int DisplayService::getTransparency( void ) const {
	return _transparency;
}

int DisplayService::defaultTransparency() const {
	return 8;
}

//	Layers
bool DisplayService::createLayers( void ) {
	//	Create layers
	Layer *layer;
	for (int i=0; i<layer::LAST_LAYER; i++) {
		//	Create layer
		layer=createLayer( (layer::type)i );
		if (!layer) {
			LWARN( "DisplayService", "Cannot create layer: %d", i );
			destroyLayers();
			return false;
		}

		//	Open layer
		if (!layer->initialize()) {
			LWARN( "DisplayService", "Cannot initialize layer: %d", i );
			delete layer;
			destroyLayers();
			return false;
		}

		//	Add to list
		_layers.push_back( layer );
	}
	return true;
}

void DisplayService::destroyLayers( void ) {
	Layer *layer;
	for (int i=0; i<layer::LAST_LAYER; i++) {
		layer=_layers[i];
		layer->finalize();
		delete layer;
	}
	_layers.clear();
}

bool DisplayService::setupLayers( void ) {
	return true;
}

Layer *DisplayService::getLayer( layer::type layer ) {
	assert( layer < layer::LAST_LAYER );
	return _layers[layer];
}

Layer *DisplayService::createLayer( layer::type layerID ) const {
	if (layerID == layer::middleware) {
		return new Layer( layerID );
	} else if (layerID == layer::zapper) {
		return new ZapperLayer( _sys->canvas() );
	}
	return new Layer( layerID );
}

//	Service virtual methods
bool DisplayService::onOnline( void ) {
	//	Create layers
	if (!createLayers()) {
		LWARN( "DisplayService", "Cannot create display layers" );
		return false;
	}

	//	Setup layers
	if (!setupLayers()) {
		LWARN( "DisplayService", "Cannot create display layers" );
		destroyLayers();
		return false;
	}

	return true;
}

void DisplayService::onOffline( void ) {
	destroyLayers();
}

void DisplayService::onLoadConfig() {
	{	//	Setup active connector
		connector::type conn = getDefaultActiveConnector();
		mgr()->load( "zapper.activeConnector", conn );
		setActive( conn );
	}

	{	//	Setup display resolution
		std::vector<connector::type> supported=supportedConnectors();
		connector::type conn;
		for (size_t i=0; i<supported.size(); i++) {
			conn=supported[i];
			mode::type mode = defaultVideoMode( conn );
			mgr()->load( getDisplayName( conn ), mode );
			setMode( conn, mode, getActiveConnector() == conn );
		}
	}

	{	//	Setup display aspect ratio
		aspect::type aspect = defaultVideoAspect();
		mgr()->load( "zapper.displayAspect", aspect );
		setAspect( aspect );
	}

	{	//	Setup display modulator
		modulator::type mod = defaultVideoModulator();
		mgr()->load( "zapper.displayModulator", mod );
		setModulator( mod );
	}

	{	//	Setup transparency
		int value = defaultTransparency();
		mgr()->load( "zapper.transparency", value );
		setTransparency( value );
	}
}

void DisplayService::onResetConfig( void ) {
	//	Set active connector when set mode
	
	{	//	Video display resolution
		std::vector<connector::type> supported=supportedConnectors();
		for (size_t i=0; i<supported.size(); i++) {
			connector::type conn=supported[i];
			setMode( conn, defaultVideoMode( conn ), getDefaultActiveConnector() == conn );
		}
	}

	//	Aspect ratio
	setAspect( defaultVideoAspect() );

	//	Display modulator
	setModulator( defaultVideoModulator() );

	//	Transparency
	setTransparency( defaultTransparency() );
}

}
}

