/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "inputservice.h"
#include "../plugin/pluginmanager.h"
#include <util/mcr.h>
#include <util/log.h>
#include <boost/bind.hpp>
#include <stdlib.h>

namespace zapper {
namespace input {

class InputService::Listener {
public:
	Listener( const std::string &id, const Callback &callback, const Keys &keys, int priority )
		: _id( id ), _callback( callback ), _keys( keys ), _priority( priority ) {}
	virtual ~Listener( void ) {}

	const std::string &id() const { return _id; }
	bool hasReserved( util::key::type key ) const {	return std::find( _keys.begin(), _keys.end(), key ) != _keys.end(); }
	int priority() const { return _priority; }
	void priority( int priority ) { _priority = priority; }
	void dispatchKey( util::key::type key, bool state ) { if (hasReserved(key)) { _callback( key, state ); } }
	void reserveKeys( const Keys &keys ) { _keys = keys; }

private:
	std::string _id;	
	Callback _callback;
	Keys _keys;
	int _priority;
};


InputService::InputService( void )
	: plugin::ServiceImpl<InputService>( "Input" )
{
}

InputService::~InputService( void )
{
	CLEAN_ALL( Listener *,_listeners );
}

void InputService::reserveKeys( const std::string &id, const Callback &callback, const Keys &keys, int priority ) {
	Listener *listener;

	LDEBUG( "InputService", "Reserve keys: listeners=%d, id=%s, size=%d, priority=%d",
		_listeners.size(), id.c_str(), keys.size(), priority );
	
	Listeners::iterator it=std::find_if(
		_listeners.begin(), _listeners.end(), boost::bind(&Listener::id,_1) == id );
	if (it == _listeners.end()) {
		listener = new Listener( id, callback, keys, priority );
		_listeners.push_back( listener );
	}
	else {
		//	Listener found, process keys
		listener = (*it);
		if (keys.size() > 0) {
			listener->reserveKeys( keys );
		}
		else {
			//	Remove listener
			_listeners.erase( it );
			delete listener;

			//	Remove all keys dispatched from this listener
			Dispatched::iterator it=_dispatched.begin();
			while (it != _dispatched.end()) {
				if (it->second == listener) {
					_dispatched.erase(it++);
				}
				else {
					it++;					
				}
			}
		}
	}
}

void InputService::renice( const std::string &id, int priority ) {
	Listeners::const_iterator it=std::find_if(
		_listeners.begin(), _listeners.end(), boost::bind(&Listener::id,_1) == id );
	if (it != _listeners.end()) {
		(*it)->priority( priority );
	}
}

bool InputService::dispatchKey( util::key::type key, bool isUp ) {
	Listener *notify=NULL;
	int priority = 0;

	//	If key press ... find listener with more priority
	if (!isUp) {
		BOOST_FOREACH( Listener *listener, _listeners ) {
			if (listener->hasReserved( key ) && priority < listener->priority()) {
				priority = listener->priority();
				notify = listener;
			}
		}
		pushListener( notify, key );
	}
	else {
		//	Get listener stored with key press
		notify=popListener( key );
	}

	//	Dispatch key
	if (notify) {
		bool dispatch = true;
		if (mgr()->inStandby()) {
			dispatch = (notify->priority() == MAX_KEY_PRIORITY);
		}

		if (dispatch) {
			notify->dispatchKey( key, isUp );
		}
	}
	
	return notify != NULL;
}

//	Button handling
void InputService::reserveButtons( const ButtonCallback &callback ) {
	_buttonCallback = callback;
}

void InputService::dispatchButton( canvas::event::ButtonEvent *evt ) {
	if (!_buttonCallback.empty()) {
		_buttonCallback( evt );
	}
}

//	State
void InputService::pushListener( Listener *l, util::key::type key ) {
	_dispatched[key] = l;
}

InputService::Listener *InputService::popListener( util::key::type key ) {
	Dispatched::iterator it=_dispatched.find( key );
	return (it == _dispatched.end()) ? NULL : it->second;
}

}
}
