/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "mediaplayer.h"
#include <util/cfg/configregistrator.h>
#include <util/url.h>
#include <util/log.h>
#include <util/mcr.h>

#define CHECK_INFO( t, v, n, m)	  \
	if (isPlaying() && v == t::LAST_##n) { m(); } \
	return (v == t::LAST_##n) ? t::unknown : v;

namespace zapper {
namespace player {

namespace impl {

static const char *audioStrings[] = {
	"audioPID",
	"audioType"
};

static const char *videoStrings[] = {
	"videoPID",
	"videoType"
};

}

MediaPlayer::MediaPlayer()
{
	_audio = audio::channel::LAST_CHANNEL;
	_aspect = display::aspect::LAST_ASPECT;
	_display = display::mode::LAST_MODE;
	_pause = false;
	_allService = false;
	_srv = NULL;
}

MediaPlayer::~MediaPlayer()
{
	assert(!_srv);
}

bool MediaPlayer::isPlaying() const {
	return _url.empty() ? false : true;
}

bool MediaPlayer::play( const std::string &url ) {
	stop();	

	if (startPlay( url )) {
		//	Reset player information
		_audio = audio::channel::LAST_CHANNEL;
		_aspect = display::aspect::LAST_ASPECT;
		_display = display::mode::LAST_MODE;
		_url = url;
		_pause = false;
		_allService = false;		
		return true;
	}
	return false;
}

void MediaPlayer::stop( void ) {
	if (isPlaying()) {
		stopPlay();
		_url.clear();
		_onPlayStop();
	}
}

OnPlayStop &MediaPlayer::onPlayStop() {
	return _onPlayStop;
}

void MediaPlayer::restart() {
	std::string tmp=_url;
	stop();
	play( tmp );
}

void MediaPlayer::pause( bool pause ) {
	if (isPlaying() && _pause != pause) {
		pausePlay( pause );
		_pause = pause;
	}
}

void MediaPlayer::pausePlay( bool /*pause*/ ) {
	//	Do nothing
}

const std::string &MediaPlayer::url() const {
	return _url;
}

const canvas::Rect &MediaPlayer::rect() const {
	return _rect;
}

void MediaPlayer::setFullScreen() {
	util::cfg::PropertyNode &sizeNode = util::cfg::get("gui.canvas.size");
	int w = sizeNode.get<int>( "width" );
	int h = sizeNode.get<int>( "height" );
	_rect = canvas::Rect(0,0,w,h);
	if (isPlaying()) {
		doResize();
	}
}

void MediaPlayer::resize( int x, int y, int w, int h ) {
	//	Do nothing
	_rect = canvas::Rect(x,y,w,h);
	if (isPlaying()) {
		doResize();
	}
}

void MediaPlayer::doResize() {
	//	Do nothing
}

int MediaPlayer::nextAudio() {
	//	Do nothing
	return -1;
}

int MediaPlayer::audioCount() const {
	return 0;
}

int MediaPlayer::nextVideo() {
	//	Do nothing
	return -1;
}

int MediaPlayer::videoCount() const {
	return 0;
}

int MediaPlayer::nextSubtitle() {
	//	Do nothing
	return -1;
}

int MediaPlayer::subtitleCount() const {
	//	Do nothing	
	return 0;
}

//	Volume
audio::channel::type MediaPlayer::getAudio() {
	CHECK_INFO( audio::channel, _audio, CHANNEL, refreshAudio );
}

display::aspect::type MediaPlayer::getAspect() {
	CHECK_INFO( display::aspect, _aspect, ASPECT, refreshAspect );
}

display::mode::type MediaPlayer::getResolution() {
	CHECK_INFO( display::mode, _display, MODE, refreshDisplay );
}

void MediaPlayer::refreshAudio() {
}

void MediaPlayer::refreshAspect() {
}

void MediaPlayer::refreshDisplay() {
}

void MediaPlayer::setAudio( audio::channel::type audio ) {
	LDEBUG( "MediaPlayer", "Set audio: %d", audio );
	_audio = audio;
}

void MediaPlayer::setAspect( display::aspect::type aspect ) {
	LDEBUG( "MediaPlayer", "Set aspect: %d", aspect );
	_aspect = aspect;
}

void MediaPlayer::setDisplay( display::mode::type display ) {
	LDEBUG( "MediaPlayer", "Set media resolution: %d", display );
	_display = display;
}

//	Service specific
void MediaPlayer::beginStartService() {
	assert(!_srv);
	_srv = new util::Url();
	_srv->type( "srvdtv" );
	_allService = true;	
}

void MediaPlayer::endStartService() {
	assert(_srv);
	play( _srv->id() );
	_allService = false;		
}

void MediaPlayer::beginStopService() {
	stop();
	assert(_srv);
	_allService = true;		
}

void MediaPlayer::endStopService() {
	assert(_srv);
	DEL(_srv);
	_allService = false;
}

bool MediaPlayer::startPCR( tuner::ID pid ) {
	assert(_srv);
	_srv->addParam( "pcrPID", pid );
	return true;
}

bool MediaPlayer::startStream( bool isVideo, tuner::ID pid, tuner::ID type ) {
	assert(_srv);

	//	Add stream params
	const char **ptr = isVideo ? impl::videoStrings : impl::audioStrings;
	_srv->addParam( ptr[0], pid );
	_srv->addParam( ptr[1], type );

	//	All service is starting?
	if (!_allService) {
		play( _srv->id() );
	}
	
	return true;
}

void MediaPlayer::stopStream( bool isVideo ) {
	assert(_srv);

	//	All service is stopping?
	if (!_allService) {
		stop();
	}

	//	Remove stream
	const char **ptr = isVideo ? impl::videoStrings : impl::audioStrings;
	_srv->rmParam( ptr[0] );
	_srv->rmParam( ptr[1] );
}

//	Display notifications
void MediaPlayer::onDisplayChanged( void ) {
}

}
}

