/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "captionplayer.h"
#include "captionviewer.h"
#include "caption.h"
#include "../streaminfo.h"
#include "../../../tuner/tuner.h"
#include <mpegparser/stream/basicpipe.h>
#include <mpegparser/demuxer/pes/caption/captiondemuxer.h>
#include <util/log.h>
#include <util/mcr.h>
#include <boost/tokenizer.hpp>
#include <stdio.h>

#define MIN_WAIT_INTER_CAPTION  100
#define MAX_WAIT_INTER_CAPTION  1500

namespace zapper {
namespace player {
namespace mpeg {

CaptionPlayer::CaptionPlayer( MediaPlayer *mp, CaptionViewer *view )
	: StreamPlayer( mp, tuner::pes::subtitle ), _view( view )
{
	_pipe = new tuner::stream::BasicPipe( "cc", 100, 1024 );

	{	//	ARIB STD-B24 Caption
		StreamInfo *i = new StreamInfo( 0x06, "ARIB STD-B24 Caption" );		
		i->addTag( 0x30, 0x3F );
		i->addTag( 0x87, 0x88 );
		i->addTag( 0x00, 0x00 );
		addStream( i );
	}

	_exit=false;
	_running=false;
	_thread = boost::thread( boost::bind(&CaptionPlayer::playCaptions,this) );
}

CaptionPlayer::~CaptionPlayer( void )
{
	_mutex.lock();
	_exit=true;
	_running=false;
	_mutex.unlock();
	_cWakeup.notify_all();
	
	_pipe->exit();
	_thread.join();

	delete _pipe;
	delete _view;
	CLEAN_ALL( tuner::arib::Statement *, _sts );
}

//	Stream player methods
bool CaptionPlayer::start( const Stream *sInfo ) {
	LDEBUG( "CaptionPlayer", "Start stream: pid=%04x", sInfo->pid );

	_pipe->reset();
	
	//	Clean lost data
	_mutex.lock();
	_running=true;
	_mutex.unlock();

	return tuner()->startFilter( sInfo->pid, type(), pipe() );
}

void CaptionPlayer::stop( const Stream *sInfo ) {
	LDEBUG( "CaptionPlayer", "Stop" );

	tuner()->stopFilter( sInfo->pid );

	_mutex.lock();
	_running=false;
	CLEAN_ALL( tuner::arib::Statement *, _sts );
	_mutex.unlock();
	_cWakeup.notify_all();

	_pipe->reset();

	_view->hide();
}

tuner::stream::Pipe *CaptionPlayer::pipe() const {
	return _pipe;
}

int CaptionPlayer::first() const {
	return -1;
}

bool CaptionPlayer::supportOff() const {
	return true;
}

//	Aux Demuxer
void CaptionPlayer::onStatement( tuner::arib::Statement *st ) { // Service thread
#ifdef DEBUG_CAPTION
	st->show();
#endif
	bool notify;
	_mutex.lock();
	if (_running) {
		notify=true;
		_sts.push_back( st );
	}
	else {
		notify=false;
		delete st;
	}
	_mutex.unlock();
	if (notify) {
		_cWakeup.notify_one();
	}
}

void CaptionPlayer::onManagement( tuner::arib::CaptionDemuxer *caption, tuner::arib::Management *mgt ) { // Service thread
	if (mgt->langs().size() > 0) {
		//	Clear management callback
		tuner::arib::CaptionDemuxer::ManagementCallback callback;
		caption->onManagement( callback );

		//	Set filter on statement
		caption->onStatement( boost::bind(&CaptionPlayer::onStatement,this,_1) );

		//	Setup group to filter: TODO: if group == 0x20 ? have 2 captions!
		util::BYTE group= (mgt->group() == 0) ? 1 : 0x21;
		caption->filterGroup( group );
	}

	mgt->show();
	delete mgt;
}

void CaptionPlayer::playCaptions() {
	tuner::arib::Statement *st;

	//	Setup last show
	pt::ptime lastShow = pt::microsec_clock::local_time();

	//	Create PES demuxer
	tuner::arib::CaptionDemuxer *demux=new tuner::arib::CaptionDemuxer();
	demux->onManagement( boost::bind(&CaptionPlayer::onManagement,this,demux,_1) );

	LINFO( "CaptionPlayer", "Begin thread" );
	while (!_exit) {
		//	Get and process pes packets
		util::Buffer *pesPkt = _pipe->popBuffer();
		if (pesPkt) {
			demux->parse( (util::BYTE *)pesPkt->buffer(), pesPkt->length() );
			_pipe->free( pesPkt );
		}

		//	Process statements
		while (!_exit && !_sts.empty()) {
			//	Get statement
			st = _sts.front();
			_sts.pop_front();

			{	//	Process statement
				const util::Buffer &buf=st->dataUnit();
				if (buf.length() > 0 && wait( lastShow )) {
				#ifdef DEBUG_CAPTION
					_debugData.clear();
				#endif
					_data.clear();
					
					processUnit( (util::BYTE *)buf.buffer(), buf.length() );
					cleanData();
					_view->show( _data );	//	If empty, show to clean the text!

					lastShow = pt::microsec_clock::local_time();
					
				#ifdef DEBUG_CAPTION					
					LDEBUG( "CaptionPlayer", "text=%s, sub=%s", _data.c_str(), _debugData.c_str() );
				#endif
				}
				
				delete st;
			}
		}
	}

	delete demux;	
	LINFO( "CaptionPlayer", "End thread" );
}

typedef boost::tokenizer< boost::char_separator<char> > tokenizer;
void CaptionPlayer::cleanData() {
	std::vector<std::string> lines;
	std::string text;
	boost::char_separator<char> sep("\n");
	tokenizer tok(_data,sep);
	for (tokenizer::iterator beg=tok.begin(); beg!=tok.end(); ++beg) {
		std::vector<std::string>::const_iterator it=std::find(
			_lines.begin(), _lines.end(), *beg );
		if (it == _lines.end()) {
			text += *beg;
		}
		lines.push_back( *beg );
	}
	_data = text;
	_lines = lines;
}

bool CaptionPlayer::wait( const pt::ptime &start ) {
	bool process=false;
	bool end=false;

	while (!end) {
		pt::ptime now = pt::microsec_clock::local_time();
		int elapsed = (int)(now - start).total_milliseconds();
		if (elapsed < MAX_WAIT_INTER_CAPTION) {
			boost::unique_lock<boost::mutex> lock( _mutex );		
			end = _cWakeup.timed_wait( lock, boost::posix_time::milliseconds(MAX_WAIT_INTER_CAPTION-elapsed) ) == false;
			process = end && !_exit && _running;
		}
		else {
			end=true;
			process = (!_exit && _running);
		}
	}
	return process;
}

void CaptionPlayer::processUnit( util::BYTE *data, int len ) {
	int skip;

	//LDEBUG( "CaptionPlayer", "Process Unit: len=%d", len );

	while (len > 0) {
		skip=0;
		switch ((*data >> 5) & 0x3) {
			case 0x3:
			case 0x2:
			case 0x1:
				skip = processChar( data, len );
				break;
			case 0x0: // C0/C1
				skip = processControlCode(data, len);
				break;
		}
		len -= skip;
		data += skip;
	}

	//LDEBUG( "CaptionPlayer", "Process Unit end!!!" );
}

int CaptionPlayer::processControlCode( util::BYTE *data, int len ) {
	int pos;
	switch (data[0]) {
		case CTRL_CODE_APS: {   //	Set active position
			if (data[1] >= 0x40 && data[1] <= 0x7F &&
				data[2] >= 0x40 && data[2] <= 0x7F)
			{
				setActivePosition( data[1] & 0x3F, data[2] & 0x3F );
			}
			pos=3;
			break;
		}
		case CTRL_CODE_COL: {  //	Color
			setColour( data[1] );
			pos=2;
			break;
		}
		case CTRL_CODE_SSZ: { //	Specifies the character size is small
			pos=1;
			setCharacterSmallSize();
			break;
		}
		case CTRL_CODE_MSZ: {  //	Specifies the character size is middle
			pos=1;
			setCharacterMiddleSize();
			break;
		}
		case CTRL_CODE_NSZ: {  //	Specifies the character size is normal
			pos=1;
			setCharacterNormalSize();
			break;
		}
		case CTRL_CODE_WHF: {  //	Foreground color: White
			setForegroundColour( 0, 0xFF, 0xFF, 0xFF );
			pos=1;
			break;
		}
		case CTRL_CODE_CSI:
			pos=processCSI( data, len );
			break;
		case CTRL_CODE_CS: {   //	Clear screen
			pos=1;
			_data.clear();
			break;
		}
		case CTRL_CODE_APF: {   //	Active position forward
			setActivePositionForward();
			pos=1;
			break;
		}
		case CTRL_CODE_APB: {   //	Active position backward
			setActivePositionBackward();
			pos=1;
			break;
		}
		case CTRL_CODE_APD: {	//	Active position down
			setActivePositionDown();
			pos=1;
			break;
		}
		case CTRL_CODE_APU: {	//	Active position up
			setActivePositionUp();
			pos=1;
			break;
		}
		case CTRL_CODE_NULL:   //	Null
		case CTRL_CODE_BEL:    //	Bell
		case CTRL_CODE_APR:
		case CTRL_CODE_LS1:
		case CTRL_CODE_LS0:
		case CTRL_CODE_PAPF:
		case CTRL_CODE_CAN:
		case CTRL_CODE_SS2:
		case CTRL_CODE_ESC:
		case CTRL_CODE_SS3:
		case CTRL_CODE_RS:
		case CTRL_CODE_US:
		case CTRL_CODE_BKF:
		case CTRL_CODE_RDF:
		case CTRL_CODE_GRF:
		case CTRL_CODE_YLF:
		case CTRL_CODE_BLF:
		case CTRL_CODE_MGF:
		case CTRL_CODE_CNF:
		case CTRL_CODE_SZX:
		case CTRL_CODE_FLC:
		case CTRL_CODE_CDC:
		case CTRL_CODE_POL:
		case CTRL_CODE_WMM:
		case CTRL_CODE_MACRO:
		case CTRL_CODE_HLC:
		case CTRL_CODE_RPC:
		case CTRL_CODE_SPL:
		case CTRL_CODE_STL:
		case CTRL_CODE_TIME:
		default: {
			pos=len;
			LWARN( "CaptionDemuxer", "Unhanded ctrl code: code=%02x", data[0] );
			break;
		}
	};

	return pos;
}

#define CHECK_PARAMS_EXACT(count) (current_params > count)
#define CHECK_PARAMS_RANGE(min,max) (current_params > min && current_params < max)
int CaptionPlayer::processCSI( util::BYTE *data, int len ) {
	int params[CSI_MAX_PARAMS];

	//	CSI Command sequence: CSI P1 I1 P2 I1 P3 ... I1 PN I2 F
	//	Examples:
	//		9B:32:20:58
	//		9B:36:30:38:3B:33:39:30:20:56

	int pos=1;	//	Skip CSI code
	int current_params=0;
	while (data[pos] != CSI_I2 && pos < len) {
		params[current_params] = 0;
		while (data[pos] != CSI_I1 && data[pos] != CSI_I2 && pos < len) {
			params[current_params] *= 10;
			params[current_params] += data[pos] & 0x0F;
			pos++;
		}
		if (data[pos] == CSI_I1) {	//	Skip I1
			pos++;
		}
		current_params++;
		assert(current_params <= CSI_MAX_PARAMS);
	}
	pos++;	//	Skip I2

	//	Check len
	if (pos > len) {
		return len;
	}

	switch (data[pos]) {
		case CSI_SWF: {
			if (CHECK_PARAMS_RANGE(0,4)) {
				setWritingFormat( params, current_params );
			}
			break;
		}
		case CSI_SDP:
			if (CHECK_PARAMS_EXACT(2)) {
				setDisplayPosition( params[0], params[1] );
			}
			break;
		case CSI_SDF:
			if (CHECK_PARAMS_EXACT(2)) {
				setDisplayFormat( params[0], params[1] );
			}
			break;
		case CSI_SSM:
			if (CHECK_PARAMS_EXACT(2)) {
				setCharacterComposition( params[0], params[1] );
			}
			break;
		case CSI_SHS:
			if (CHECK_PARAMS_EXACT(1)) {
				setHorizontalSpacing( params[0] );
			}
			break;
		case CSI_SVS:
			if (CHECK_PARAMS_EXACT(1)) {
				setVerticalSpacing( params[0] );
			}
			break;
		case CSI_RCS:
			if (CHECK_PARAMS_EXACT(1)) {
				setRasterColour( params[0] );
			}
			break;
		default:
			LWARN( "CaptionDemuxer", "Unhandled CSI extension: code=%02x, params=%d",
				data[pos], current_params );
			break;
	}

	return pos+1;	//	Skip command
}

int CaptionPlayer::processChar( util::BYTE *data, int /*len*/ ) {
	if ((*data & 0x7F) == CTRL_CODE_SP) {
#ifdef DEBUG_CAPTION
		_debugData += ' ';
#endif
		if (!_data.empty()) {
			char last=_data[_data.size()-1];
			if (last != ' ' && last != '\n') {
				_data += data[0];
			}
		}
	}
	else if ((*data & 0x7F) == CTRL_CODE_DEL) {
		if (_data.size () > 0) {
			_data.resize( _data.size()-1 );
		}
	}
	else {
#ifdef DEBUG_CAPTION
		_debugData += data[0];
#endif

		bool copy=true;
		if (_data.empty() && (data[0] == ' ' || data[0] == '\n')) {
			copy=false;
		}

		if (copy) {
			_data += data[0];
		}
	}
	return 1;
}

//	Commands
void CaptionPlayer::setActivePosition( int p1, int p2 ) {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<APS(%d,%d)>\n", p1, p2 );
	_debugData += tmp;
#else
	UNUSED(p1);
	UNUSED(p2);
#endif
	if (!_data.empty() && _data[_data.size()-1] != '\n') {
		_data += '\n';
	}
}

void CaptionPlayer::setActivePositionForward() {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<APF>\n" );
	_debugData += tmp;
#endif
}

void CaptionPlayer::setActivePositionBackward() {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<APB>\n" );
	_debugData += tmp;
#endif
}

void CaptionPlayer::setActivePositionDown() {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<APD>\n" );
	_debugData += tmp;
#endif
}

void CaptionPlayer::setActivePositionUp() {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<APU>\n" );
	_debugData += tmp;
#endif
}

void CaptionPlayer::setForegroundColour( util::BYTE r, util::BYTE g, util::BYTE b, util::BYTE alpha/*=255*/ ) {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<fg color: rgba=%02x,%02x,%02x,%02x>\n", r, g, b, alpha );
	_debugData += tmp;
#else
	UNUSED(r);
	UNUSED(g);
	UNUSED(b);
	UNUSED(alpha);
#endif
}

void CaptionPlayer::setBackgroundColour( util::BYTE r, util::BYTE g, util::BYTE b, util::BYTE alpha/*=255*/ ) {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<bg color: rgba=%02x,%02x,%02x,%02x>\n", r, g, b, alpha );
	_debugData += tmp;
#else
	UNUSED(r);
	UNUSED(g);
	UNUSED(b);
	UNUSED(alpha);
#endif
}

void CaptionPlayer::setColour( int p1 ) {
	switch (p1) {
		case 0x50:
		case 0x70:	//	Black
			setBackgroundColour( 0, 0, 0 );
			break;
		case 0x51:
		case 0x71:	//	Red
			setBackgroundColour( 0xFF, 0, 0 );
			break;
		case 0x52:
		case 0x72:	//	Green
			setBackgroundColour( 0, 0xFF, 0 );
			break;
		case 0x53:
		case 0x73:	//	Yellow
			setBackgroundColour( 0xFF, 0xFF, 0 );
			break;
		case 0x54:
		case 0x74:	//	Blue
			setBackgroundColour( 0, 0xFF, 0xFF );
			break;
		case 0x55:
		case 0x75:	//	Magenta
			setBackgroundColour( 0xFF, 0, 0xFF );
			break;
		case 0x56:
		case 0x76:	//	Cyan
			setBackgroundColour( 0, 0xFF, 0xFF );
			break;
		case 0x57:
		case 0x77:	//	White
			setBackgroundColour( 0xFF, 0xFF, 0xFF );
			break;
		case 0x58:
		case 0x78:	//	Transparent
			setBackgroundColour( 0, 0, 0, 0 );
			break;
		case 0x59:
		case 0x79:	//	Half red
			setBackgroundColour( 0x88, 0, 0 );
			break;
		case 0x5A:
		case 0x7A:	//	Half green
			setBackgroundColour( 0, 0x88, 0 );
			break;
		case 0x5B:
		case 0x7B:	//	Half yellow
			setBackgroundColour( 0x88, 0x88, 0 );
			break;
		case 0x5C:
		case 0x7C:	//	Half blue
			setBackgroundColour( 0, 0, 0x88 );
			break;
		case 0x5D:
		case 0x7D:	//	Half magenta
			setBackgroundColour( 0x88, 0, 0x88 );
			break;
		case 0x5E:
		case 0x7E:	//	Half cyan
			setBackgroundColour( 0, 0x88, 0x88 );
			break;
		case 0x5F:
		case 0x7F:	//	Half white
			setBackgroundColour( 0x88, 0x88, 0x88 );
			break;
		case 0x60:	//	Black
			setForegroundColour( 0, 0, 0 );
			break;
		case 0x61:	//	Red
			setForegroundColour( 0xFF, 0, 0 );
			break;
		case 0x62:	//	Green
			setForegroundColour( 0, 0xFF, 0 );
			break;
		case 0x63:	//	Yellow
			setForegroundColour( 0xFF, 0xFF, 0 );
			break;
		case 0x64:	//	Blue
			setForegroundColour( 0, 0, 0xFF );
			break;
		case 0x65:	//	Magenta
			setForegroundColour( 0xFF, 0, 0xFF );
			break;
		case 0x66:	//	Cyan
			setForegroundColour( 0, 0xFF, 0xFF );
			break;
		case 0x67:	//	White
			setForegroundColour( 0xFF, 0xFF, 0xFF );
			break;
		case 0x48:
		case 0x68:	//	Transparent
			setForegroundColour( 0, 0, 0, 0 );
			break;
		case 0x49:
		case 0x69:	//	Half red
			setForegroundColour( 0x88, 0, 0 );
			break;
		case 0x4A:
		case 0x6A:	//	Half green
			setForegroundColour( 0, 0x88, 0 );
			break;
		case 0x4B:
		case 0x6B:	//	Half Yellow
			setForegroundColour( 0x88, 0x88, 0 );
			break;
		case 0x4C:
		case 0x6C:	//	Half Blue
			setForegroundColour( 0, 0, 0x88 );
			break;
		case 0x4D:
		case 0x6D:	//	Half Magenta
			setForegroundColour( 0x88, 0, 0x88 );
			break;
		case 0x4E:
		case 0x6E:	//	Half Cyan
			setForegroundColour( 0, 0x88, 0x88 );
			break;
		case 0x4F:
		case 0x6F:	//	Half White
			setForegroundColour( 0x88, 0x88, 0x88 );
			break;
	};
}

void CaptionPlayer::setCharacterSmallSize() {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<SSZ>\n" );
	_debugData += tmp;
#endif
}

void CaptionPlayer::setCharacterMiddleSize() {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<MSZ>\n" );
	_debugData += tmp;
#endif
}

void CaptionPlayer::setCharacterNormalSize() {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<NSZ>\n" );
	_debugData += tmp;
#endif
}

//	CSI commands
void CaptionPlayer::setWritingFormat( int *params, int count ) {
	//	Check writing form
	if (params[0] >= 0 && params[0] <= 12) {
		setWritingForm( params[0] );
	}

	//	Check size
	if (count > 1) {
		switch (params[1]) {
			case 0x30:
				setCharacterSmallSize();
				break;
			case 0x31:
				setCharacterMiddleSize();
				break;
			case 0x33:
				setCharacterNormalSize();
				break;
		};
	}
}

void CaptionPlayer::setWritingForm( int value ) {
	//	Value can be:
	//		0: horizontal writing form in standard density
	//		1: vertical writing form in standard density
	//		2: horizontal writing form in high density
	//		3: vertical writing form in high density
	//		4: horizontal writing form in Western language
	//		5: horizontal writing form in 1920x1080
	//		6: vertical writing form in 1920x1080
	//		7: horizontal writing form in 960x540
	//		8: vertical writing form in 950x540
	//		9: horizontal writing form in 720x480
	//	   10: vertical writing form in 720x480
	//	   11: horizontal writing form in 1280x720
	//	   12: vertical writing form in 1280x720

#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<CSI_SWF(%d)>\n", value );
	_debugData += tmp;
#else
	UNUSED(value);
#endif
}

void CaptionPlayer::setDisplayPosition( int horizontal, int vertical ) {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<SDP(%d,%d)>\n", horizontal, vertical );
	_debugData += tmp;
#else
	UNUSED(horizontal);
	UNUSED(vertical);
#endif
}

void CaptionPlayer::setDisplayFormat( int horizontal, int vertical ) {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<SDF(%d,%d)>\n", horizontal, vertical );
	_debugData += tmp;
#else
	UNUSED(horizontal);
	UNUSED(vertical);
#endif
}

void CaptionPlayer::setCharacterComposition( int horizontal, int vertical ) {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<SSM(%d,%d)>\n", horizontal, vertical );
	_debugData += tmp;
#else
	UNUSED(horizontal);
	UNUSED(vertical);
#endif
}

void CaptionPlayer::setHorizontalSpacing( int value ) {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<SHS(%d)>\n", value );
	_debugData += tmp;
#else
	UNUSED(value);
#endif
}

void CaptionPlayer::setVerticalSpacing( int value ) {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<SVS(%d)>\n", value );
	_debugData += tmp;
#else
	UNUSED(value);
#endif
}

void CaptionPlayer::setRasterColour( int value ) {
#ifdef DEBUG_CAPTION
	char tmp[50];
	snprintf( tmp, sizeof(tmp), "<RCS(%d)>\n", value );
	_debugData += tmp;
#else
	UNUSED(value);
#endif
}

}
}
}

