/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "streamplayer.h"
#include "streaminfo.h"
#include "../mediaplayer.h"
#include "../../tuner/tuner.h"
#include <util/log.h>

namespace zapper {
namespace player {
namespace mpeg {

bool TagPriority( const Stream &i1, const Stream &i2 ) {
	return i1.tag < i2.tag;
}

StreamPlayer::StreamPlayer( MediaPlayer *mp, tuner::pes::type type, const std::vector<StreamInfo *> &streams )
	: _mp(mp), _pesType( type ), _supported( streams )
{
	_current = -1;
	_tuner = NULL;
}

StreamPlayer::StreamPlayer( MediaPlayer *mp, tuner::pes::type type )
	: _mp(mp), _pesType( type )
{
	_current = -1;
	_tuner = NULL;
}

StreamPlayer::~StreamPlayer( void )
{
	CLEAN_ALL( StreamInfo *, _supported );
}

bool StreamPlayer::initialize( Tuner *tuner ) {
	_tuner = tuner;
	return true;
}

void StreamPlayer::finalize() {
	_tuner = NULL;
}

void StreamPlayer::addStream( StreamInfo *info ) {
	_supported.push_back( info );
}

void StreamPlayer::start() {
	//	Sort streams by tag
	std::sort( _streams.begin(), _streams.end(), TagPriority );

	start( first() );
}

void StreamPlayer::stop( bool clean ) {
	if (_current >= 0) {
		LINFO( "mpeg::StreamPlayer", "Stop stream: type=%d", _pesType );
		const mpeg::Stream *sInfo = getStream(_current);
		assert(sInfo);
		stop( sInfo );
		_current = -1;
	}

	if (clean) {
		_streams.clear();
	}
}

int StreamPlayer::first() const {
	return _streams.size() > 0 ? 0 : -1;
}

int StreamPlayer::next() {
	//	Get next stream
	int nextID = next( _current, 1, _streams.size(), supportOff() );
	LDEBUG( "mpeg::StreamPlayer", "Next stream: type=%d, _current=%d, next=%d", _pesType, _current, nextID );
	if (nextID != _current) {
		//	Stop current stream
		stop( false );

		//	Start new
		start( nextID );
	}
	return _current;
}

int StreamPlayer::count() const {
	return static_cast<int>(_streams.size());
}

tuner::pes::type StreamPlayer::type() const {
	return _pesType;
}

bool StreamPlayer::collectStream( const tuner::ElementaryInfo &info ) {
	//	Get elementary stream info for that type
	StreamInfo *sInfo = getStreamInfo( info.streamType );
	if (sInfo) {
		//	Get tag from descriptor
		util::BYTE tag=tuner::desc::getStreamIdentifier( info.descriptors );

		//	Check if tag is in range
		if (!sInfo->checkTag( tag )) {
			LWARN( "mpeg::StreamPlayer", "Component tag out of range: pid=%04x, type=%04x, tag=%02x",
				info.pid, info.streamType, tag );
		}

		{	//	Always add stream, the component tag is only recomended
			Stream stream;
			stream.pid = info.pid;
			stream.type = info.streamType;
			stream.tag = tag;
			_streams.push_back( stream );

			LINFO( "mpeg::StreamPlayer", "Add stream: pid=%04x, type=%04x, name=%s",
				info.pid, info.streamType, sInfo->name().c_str() );
		}

		return true;
	}

	return false;
}

tuner::stream::Pipe *StreamPlayer::pipe() const {
	return NULL;
}

void StreamPlayer::start( int id ) {
	if (id >= 0) {
		const Stream *sInfo = getStream(id);
		assert(sInfo);
		LINFO( "mpeg::StreamPlayer", "Start stream: type=%d, id=%d, pid=%04x", _pesType, id, sInfo->pid );
		if (start( sInfo )) {
			_current = id;
		}
		else {
			LWARN( "mpeg::StreamPlayer", "Cannot start stream: type=%d, id=%d, pid=%04x", _pesType, id, sInfo->pid );			
		}
	}
}

bool StreamPlayer::start( const Stream *sInfo ) {
	if (!_mp->startStream( _pesType == tuner::pes::video, sInfo->pid, sInfo->type )) {
		LWARN( "mpeg::StreamPlayer", "Cannot start stream: pid=%04x", sInfo->pid );		
		return false;
	}
	
	//	Start filter
	return tuner()->startFilter( sInfo->pid, _pesType, pipe() );
}

void StreamPlayer::stop( const Stream *sInfo ) {
	_mp->stopStream( _pesType == tuner::pes::video );
	
	tuner()->stopFilter( sInfo->pid );
}

bool StreamPlayer::supportOff() const {
	return false;
}

StreamInfo *StreamPlayer::getStreamInfo( tuner::ID type ) const {
	std::vector<StreamInfo *>::const_iterator it=std::find_if(
		_supported.begin(),
		_supported.end(),
		boost::bind(&StreamInfo::type,_1) == type );
	return (it == _supported.end()) ? NULL : (*it);
}

const Stream *StreamPlayer::getStream( int id ) const {
	if (id >= 0 && id < static_cast<int>(_streams.size())) {
		return &_streams[id];
	}
	else {
		return NULL;
	}
}

int StreamPlayer::next( int pos, int factor, int size, bool supportOff ) const {
	if (!size || (supportOff && (pos >= 0 && pos == size-1))) {
		return -1;
	}
	else {
		int a = (pos + factor);
		return (int)(a - floor(a/(double)size)*size);
	}
}

Tuner *StreamPlayer::tuner() const {
	assert(_tuner);
	return _tuner;
}

}
}
}

