/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "pluginmanager.h"
#include "plugin.h"
#include <util/task/dispatcher.h>
#include <util/mcr.h>
#include <util/log.h>
#include <boost/bind.hpp>
#include <boost/utility.hpp>

namespace zapper {
namespace plugin {

PluginManager::PluginManager( util::task::Dispatcher *disp )
	: _tasks( disp ), _status( status::idle )
{
	_storage = NULL;
}

PluginManager::~PluginManager()
{
	finalize();

	//	Clean all plugins
	CLEAN_ALL(Plugin *,_plugins);

	assert(!_storage);
}

//	Add plugin
void PluginManager::add( Plugin *plugin ) {
	assert(plugin);
	_plugins.push_back( plugin );
}

bool PluginManager::isPresent( const std::string &plugin ) const {
	return findPlugin( plugin ) ? true : false;
}

bool PluginManager::isEnabled( const std::string &plugin ) const {
	Plugin *p = findPlugin( plugin );
	return p ? p->isEnabled() : false;
}

void PluginManager::enable( const std::string &plugin, bool state ) {
	Plugin *p = findPlugin( plugin );
	if (p) {
		p->enable( state );
	}
	else {
		LWARN( "PluginManager", "Plugin not found: name=%s", plugin.c_str() );
	}
}

//	Initialization
void PluginManager::initialize( const std::string &dbDir ) {
	if (_status == status::idle) {
		//	Load configuration from storage
		_storage = new util::storage::bdb::Storage( dbDir );
		if (!_storage->initialize()) {
			DEL(_storage);
			LWARN("PluginManager", "Could not initialize storage");
		}

		_status = status::initialized;
		BOOST_FOREACH( Plugin *plugin, _plugins ) {
			plugin->initialize( this );
		}

		//	Register into dispatcher
		_tasks->registerTarget( this );
	}
}

void PluginManager::finalize() {
	if (_status > status::idle) {
		//	Un-register into dispatcher
		_tasks->unregisterTarget( this );
		
		//	Stop services
		stop();

		//	Finalize all plugins
		BOOST_FOREACH( Plugin *plugin, _plugins ) {
			plugin->finalize();
		}

		//	Un-register all services
		cleanServices();

		//	Save configuration in storage
		_storage->finalize();
		DEL(_storage);

		_status = status::idle;
	}
}

bool PluginManager::isInitialized() const {
	return _status == status::initialized;
}

//	Start/Stop services
bool PluginManager::isStarted() const {
	return _status == status::started;
}

void PluginManager::start() {
	if (_status == status::initialized) {
		LDEBUG( "PluginManager", "Start: services=%d", _services.size() );
		
		_status = status::started;
		BOOST_FOREACH( Service *srv, _services ) {
			start( srv );
		}
	}
}

void PluginManager::start( Service *srv ) {
	LDEBUG( "PluginManager", "Start: name=%s", srv->name().c_str() );
	srv->start();
}

void PluginManager::stop() {
	if (_status == status::started) {
		LDEBUG( "PluginManager", "Stop: services=%d", _services.size() );
		
		BOOST_FOREACH( Service *srv, _services ) {
			srv->onStopFired();
		}
		
		_status = status::initialized;
		BOOST_FOREACH( Service *srv, _services ) {
			stop( srv );
		}
	}
}

void PluginManager::stop( Service *srv ) {
	LDEBUG( "PluginManager", "Stop service: name=%s", srv->name().c_str() );
	srv->stop();
}

//	Standby
bool PluginManager::inStandby() const {
	return _status == status::standby;
}

void PluginManager::standby() {
	if (_status == status::started) {
		LDEBUG( "PluginManager", "Standby: services=%d", _services.size() );
		
		_status = status::standby;
		BOOST_FOREACH( Service *srv, _services ) {
			srv->standby();
		}
	}
}

void PluginManager::wakeup() {
	if (_status == status::standby) {
		LDEBUG( "PluginManager", "Wakeup: services=%d", _services.size() );
		
		_status = status::started;
		BOOST_FOREACH( Service *srv, _services ) {
			srv->wakeup();
		}
	}
}

//	Configuration
void PluginManager::resetConfig() {
	//	Reset services config
	BOOST_FOREACH( Service *srv, _services ) {
		srv->resetConfig();
	}
	//	Reset plugin config
	BOOST_FOREACH( Plugin *plugin, _plugins ) {
		plugin->resetConfig();
	}
}

//	Service registration
std::vector<std::string> PluginManager::findInstances( const std::string &className ) const {
	std::vector<std::string> tmp;
	Services::const_iterator it=_services.begin();
	while (it != _services.end()) {
		if ((*it)->className() == className) {
			tmp.push_back( (*it)->name() );
		}
		it++;
	}
	return tmp;
}

void PluginManager::cleanServices() {
	Services::iterator it=_services.begin();
	while (it != _services.end()) {
		unregisterService( (*it) );
		it = _services.erase( it );
	}
	_services.clear();
}

OnRegisterChanged &PluginManager::onRegisterChanged() {
	return _onRegisterChanged;
}

const std::vector<Service *> &PluginManager::services() const {
	return _services;
}

Service *PluginManager::getService( const std::string &name ) const {
	Services::const_iterator it=std::find_if( _services.begin(), _services.end(), boost::bind(&Service::name,_1) == name );
	if (it != _services.end()) {
		return (*it);
	}
	return NULL;
}

bool PluginManager::isServiceRegistered( const std::string &name ) const {
	return getService( name ) ? true : false;
}

bool PluginManager::registerService( Service *srv ) {
	bool result=false;

	//	Only register service if manager is initialized ...
	if (_status >= status::initialized && !isServiceRegistered( srv->name() )) {
		LDEBUG( "PluginManager", "Register service: name=%s", srv->name().c_str() );
			
		//	Add to list of services registered
		_services.push_back( srv );

		//	Attach service to manager and notify
		srv->attach( this );

		//	Notify a service was registered
		_onRegisterChanged( srv, true );

		//	Try start if necesary
		if (_status == status::started) {
			start( srv );
		}
		result=true;
	}
	return result;
}

void PluginManager::unregisterService( const std::string &name ) {
	Services::iterator it=std::find_if( _services.begin(), _services.end(), boost::bind(&Service::name,_1) == name );
	if (it != _services.end()) {
		unregisterService( (*it) );
		_services.erase( it );
	}
}

void PluginManager::unregisterService( Service *srv ) {
	LDEBUG( "PluginManager", "Unregister service: name=%s", srv->name().c_str() );

	_onRegisterChanged( srv, false );
	
	//	Detach and delete service
	srv->dettach();
	if (!srv->haveReferences()) {
		delete srv;
	}
}

//	Service notifications
void PluginManager::onServiceChanged( Service *srv ) {
	//	Notify to all services that a service was changed
	BOOST_FOREACH( Service *s, _services ) {
		if (s != srv) {	//	Only notify to other services
			s->onServiceChanged( srv );
		}
	}
}

//	Aux
Plugin *PluginManager::findPlugin( const std::string &name ) const {
	std::vector<Plugin *>::const_iterator it=std::find_if( _plugins.begin(), _plugins.end(), boost::bind(&Plugin::name,_1) == name );
	return (it != _plugins.end()) ? (*it) : NULL;
}

//	Enqueue method to execute in service thread
void PluginManager::post( const util::task::Type &task ) {
	_tasks->post( this, task );
}

}
}

