/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "service.h"
#include "pluginmanager.h"
#include "dependency.h"
#include "dependencyhandler.h"
#include <util/mcr.h>
#include <util/log.h>
#include <boost/bind.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>

namespace zapper {
namespace plugin {

Service::Service()
	: _name( "Change this!" )
{
	_registration = service::status::offline;
	_run = service::status::stopped;
	_mgr = NULL;
	_dependencies = new DependencyHandler();
	_references = 0;
}

Service::Service( const std::string &name )
	: _name( name )
{
	_registration = service::status::offline;
	_run = service::status::stopped;
	_mgr = NULL;
	_dependencies = new DependencyHandler();
	_references = 0;
}

Service::~Service()
{
	assert(!_references);
	delete _dependencies;
}

//	Status
void Service::status( service::status::registration reg ) {
	if (_registration != reg) {
		bool needNotify = (_registration > service::status::offline);
		_registration = reg;
		if (needNotify) {
			mgr()->onServiceChanged( this );
		}
	}
}

void Service::status( service::status::run st ) {
	if (_run != st) {
		bool needNotify = (_registration > service::status::offline);
		_run = st;
		if (needNotify) {
			mgr()->onServiceChanged( this );
		}
	}
}

//	Getters
PluginManager *Service::mgr() const {
	return _mgr;
}

bool Service::isOnline() const {
	return _registration == service::status::online;
}

void Service::setName( const std::string &name ) {
	_name = name;
}

void Service::setName( const std::string &name, int id ) {
	std::string tmp = name;
	tmp += boost::lexical_cast<std::string>( id );
	setName( tmp );
}

const std::string &Service::name() const {
	return _name;
}

const std::string Service::className() const {
	return name();
}

//	Start/Stop
void Service::start() {
	//	If not running and all dependencies are running
	if (_run == service::status::stopped) {
		//	Change to starting
		status( service::status::starting );

		//	Try start
		checkStatus();
	}
}

void Service::tryStart() {
	if (_registration == service::status::online) {
		//	Load service configuration
		onLoadConfig();

		//	Try start service
		service::status::run st;
		if (onStart()) {
			st = service::status::started;
		}
		else {
			LWARN( "Service", "Cannot start service: name=%s", name().c_str() );
			st = service::status::stopped;
		}
		status( st );
	}
}

void Service::stop() {
	if (_run > service::status::stopped) {
		bool needStop = (_run == service::status::started);
		status( service::status::stopped );
		if (needStop) {
			onStop();
		}
	}
}

void Service::onStopFired() {
}

bool Service::isStarted() const {
	return _run == service::status::started;
}

bool Service::onStart() {
	return true;
}

void Service::onStop() {
	//	Do nothing
}

//	Standby
void Service::standby() {
	LDEBUG( "Service", "Standby not implemented: name=%s", name().c_str() );
}

void Service::wakeup() {
	LDEBUG( "Service", "Wakeup not implemented: name=%s", name().c_str() );	
}

//	Dependent
void Service::addDependency( Dependency *dep ) {
	assert(_mgr);
	assert(dep);
	_dependencies->add( dep );
}

void Service::addSimpleDependency( const std::string &name ) {
	addDependency( new plugin::SimpleDependency( name ) );	
}

void Service::addOptionalDependency( const std::string &name ) {
	addDependency( new plugin::SimpleDependency( name, true, true ) );	
}

//	Aux config
void Service::onLoadConfig() {
	//	Do nothing
}

void Service::resetConfig() {
	if (_registration == service::status::online) {
		onResetConfig();
	}
}

void Service::onResetConfig() {
}

//	PluginManager methods
void Service::attach( PluginManager *mgr ) {
	if (_registration == service::status::offline) {
		assert(mgr);
		_mgr = mgr;
		_references = 0;

		//	Do custom attach, setup dependencies
		onAttach();
		status( service::status::registrated );

		//	Setup dependencies
		_dependencies->initialize( mgr );

		checkStatus();
	}
}

void Service::dettach() {
	if (_registration > service::status::offline) {
		//	If service is running, stop
		stop();

		if (_registration == service::status::online) {
			onOffline();
			status( service::status::offline );			
		}
		
		//	Do custom dettach
		onDettach();

		//	Check references
		assert( !_references );
		_mgr = NULL;

		//	Clean all dependencies
		_dependencies->finalize();
	}
}

void Service::onAttach() {
	//	Do nothing
}

void Service::onDettach() {
	//	Do nothing
}

//	Aux online
bool Service::onOnline() {
	//	Do nothing
	return true;
}

void Service::onOffline() {
	//	Do nothing
}

//	PluginManager notifications
void Service::onServiceChanged( Service *srv ) {
	//	Update dependencies
	_dependencies->update( srv );

	//	Try change status
	checkStatus();
}

void Service::checkStatus() {
	bool depRunning = _dependencies->areRunning();
	bool depOnline = _dependencies->areOnline();

	//	Need stop?
	if (!depRunning && _run == service::status::started) {
		status( service::status::starting );
		onStop();
	}

	//	Need go online/offline?
	if (!depOnline && _registration == service::status::online) {
		status( service::status::registrated );
		onOffline();
	}
	else if (depOnline && _registration == service::status::registrated) {
		if (onOnline()) {
			status( service::status::online );
		}
	}

	//	Can start
	if (depRunning && _run == service::status::starting) {
		tryStart();
	}
}

//	Add/release use reference
void Service::addRef() {
	_references++;
}

void Service::release() {
	_references--;
}

bool Service::haveReferences() const {
	return _references > 0;
}

}
}
