/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "tuner.h"
#include "tunerlistener.h"
#include "servicefilter.h"
#include "scanextension.h"
#include "../channel/channel.h"
#include "../plugin/pluginmanager.h"
#include "generated/config.h"
#include <mpegparser/service/serviceprovider.h>
#include <mpegparser/service/extension/extension.h>
#include <mpegparser/provider/provider.h>
#include <mpegparser/resourcemanager.h>
#include <util/buffer.h>
#include <util/mcr.h>
#include <util/log.h>
#include <boost/filesystem.hpp>
#include <boost/bind.hpp>
#include <assert.h>

#define NOTIFY(meth)	\
	BOOST_FOREACH( TunerListener *listener, _listeners ) { listener->meth; }

namespace zapper {

namespace fs = boost::filesystem;

Tuner::Tuner( int id, tuner::Provider *provider, ServiceFilter *filter, const std::string &ramDisk )
{
	//	Set name
	setName( Tuner::className(), id );

	//	Setup provider
	assert( provider );
	_provider = provider;
	_provider->onStatusChanged( boost::bind( &Tuner::onState, this, _1, _2 ) );

	{
		//	Make resources working directory
		fs::path tmp = ramDisk;
		tmp /= "resources";

		//	Setup service manager
		tuner::ResourceManager *resMgr = new tuner::ResourceManager( tmp.string(), ZAPPER_RESOURCES_MEMORY_BLOCKS, ZAPPER_RESOURCES_MAX_MODULES, ZAPPER_RESOURCES_MAX_MODULE_SIZE );
		_services = new tuner::ServiceProvider( provider, resMgr );
	}

	_filter = filter;
	_inScan = false;
}

Tuner::~Tuner()
{
	assert(!_inScan);
	CLEAN_ALL( TunerListener *, _listeners );
	delete _provider;
	delete _services;
	delete _filter;
}

const std::string Tuner::className() const {
	return "Tuner";
}

void Tuner::onStopFired() {
	stopScan();
}

//	Start/Stop
bool Tuner::onOnline() {
	//	Create scan extension and attach to tuner
	attach( new ScanExtension( this, _filter ) );
	return true;
}

bool Tuner::onStart() {
	_services->initialize();
	return _provider->start();
}

void Tuner::onStop() {
	//	Stop service provider
	_services->stop();
	
	//	Stop provider
	_provider->stop();

	_services->finalize();	
}

void Tuner::standby() {
	stopScan();
}

//  Listener
void Tuner::addListener( TunerListener *listener ) {
	_listeners.push_back( listener );
}

//	Provider notification
void Tuner::onState( tuner::state::type st, tuner::signal::type signal ) {
	NOTIFY( onStateChanged( st, signal ) );
}

tuner::state::type Tuner::state() const {
	return _provider->state();
}

//	Tune network and start srvID
bool Tuner::startService( size_t net, tuner::ID srvID ) {
	assert(!_inScan);
	if (startNetwork( net )) {
		_services->startService( srvID );
		return true;
	}
	return false;
}

void Tuner::stopService( tuner::ID srvID ) {
	_services->stopService( srvID );
}

bool Tuner::startFilter( tuner::ID pid, tuner::pes::type pesType, tuner::stream::Pipe *pipe ) {
	return _services->startFilter( pid, pesType, pipe );
}

void Tuner::stopFilter( tuner::ID pid ) {
	_services->stopFilter( pid );
}

//	Attach extension
void Tuner::attach( tuner::Extension *ext ) {
	assert(ext);
	_services->attach( ext );
}

void Tuner::detach( tuner::Extension *ext ) {
	assert(ext);
	_services->detach( ext );	
}

//	Start/stop network
bool Tuner::startNetwork( size_t net ) {
	bool result=true;

	LDEBUG( "Tuner", "Start network: net=%d", net );
	if (net != _provider->currentNetwork()) {
		//	Stop service provider
		_services->stop();
	
		//	Try set new network
		result=_provider->setNetwork( net );
		if (result) {
			//	Start service provider
			_services->start();
		}
	}
	return result;
}

//	Scan
bool Tuner::isScanning() const {
	return _inScan;
}

void Tuner::startScan() {
	if (!_inScan) {
		LDEBUG( "Tuner", "Start scan begin" );		

		_inScan = true;

		//	Notify iter to begin scan
		NOTIFY( onBeginScan( _provider->getNetworksCount() ) );

		//	Stop service provider
		_services->stop();

		//	Attach to end network scan callback
		_services->onEndScan( boost::bind(&Tuner::endNetScan,this) );

		//	Begin scan of networks
		nextScan( 0 );
	}
}

void Tuner::stopScan() {
	LDEBUG( "Tuner", "Stop scan: inScan=%d", _inScan );		
	if (_inScan) {
		_provider->cancelScan();
		_services->stop();
		_services->onEndScan( NULL );
		_inScan = false;
		NOTIFY( onEndScan() );
	}
}

void Tuner::nextScan( size_t networkToScan ) {
	if (_inScan) {
		LDEBUG( "Tuner", "Next network scan: network=%d", networkToScan );		
		if (networkToScan < _provider->getNetworksCount()) {
			//	Notify network to scan
			NOTIFY( onNetwork( _provider->getNetworkName(networkToScan) ) );

			//	Try scan
			if (_provider->scan( networkToScan )) {
				LDEBUG( "Tuner", "Begin network scan" );
				_services->start( true );
			}
			else {
				//	Enqueue task, so we can execute a cancelScan task
				mgr()->runMethod<Tuner>( name(), boost::bind(&Tuner::nextScan,_1,networkToScan+1) );
			}

		}
		else {
			stopScan();
		}
	}
}

void Tuner::endNetScan() {
	if (_inScan) {
		LDEBUG( "Tuner", "End network scan!" );

		//	Stop service scan
		_services->stop();

		//	Iterate to next network
		nextScan( _provider->currentNetwork()+1 );
	}
}

void Tuner::onChannel( channel::Channel *ch ) {
	//	Set network
	NOTIFY( onChannel( ch, _inScan ) );
}

void Tuner::onStartService( tuner::Service *srv ) {
	LDEBUG( "Tuner", "On start service" );
	_onServiceStarted( srv, true );
}

void Tuner::onStopService( tuner::Service *srv ) {
	LDEBUG( "Tuner", "On stop service" );	
	_onServiceStarted( srv, false );	
}

OnServiceStarted &Tuner::onServiceStarted() {
	return _onServiceStarted;
}

//	Attach/detach extension
void Tuner::attach( plugin::PluginManager *mgr, tuner::Extension *ext, const std::string &name/*="Tuner0"*/ ) {
	//	Attach extension
	Tuner *tuner = mgr->findService<Tuner>( name );
	assert(tuner);
	tuner->attach( ext );
}

void Tuner::detach( plugin::PluginManager *mgr, tuner::Extension *ext, const std::string &name/*="Tuner0"*/ ) {
	//	Attach extension
	Tuner *tuner = mgr->findService<Tuner>( name );
	assert(tuner);
	tuner->detach( ext );
}

}
