/*******************************************************************************

  Copyright (C) 2010, 2013 LIFIA - Facultad de Informatica - Univ. Nacional de La Plata

********************************************************************************

  This file is part of DTV-zapper implementation.

    DTV-zapper is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the Free
  Software Foundation, either version 2 of the License.

    DTV-zapper is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License along with
  this program. If not, see <http://www.gnu.org/licenses/>.

********************************************************************************

  Este archivo es parte de la implementación de DTV-zapper.

    DTV-zapper es Software Libre: Ud. puede redistribuirlo y/o modificarlo
  bajo los términos de la Licencia Pública General Reducida GNU como es publicada por la
  Free Software Foundation, según la versión 2 de la licencia.

    DTV-zapper se distribuye esperando que resulte de utilidad, pero SIN NINGUNA
  GARANTÍA; ni siquiera la garantía implícita de COMERCIALIZACIÓN o ADECUACIÓN
  PARA ALGÚN PROPÓSITO PARTICULAR. Para más detalles, revise la Licencia Pública
  General Reducida GNU.

    Ud. debería haber recibido una copia de la Licencia Pública General Reducida GNU
  junto a este programa. Si no, puede verla en <http://www.gnu.org/licenses/>.

*******************************************************************************/

#include "generated/config.h"
#include "zapper.h"
#include "plugin/pluginmanager.h"
#include "plugin/plugin.h"
#include "plugin/genericservice.h"
#include "input/inputservice.h"
#include "mount/mountservice.h"
#include "control/service.h"
#include "audio/service.h"
#include "audio/mixer.h"
#include "channel/channelmanager.h"
#include "channelplayer/channelservice.h"
#include "player/playerservice.h"
#include "player/impl/player.h"
#include "update/updateplugin.h"
#include "tuner/tuner.h"
#include "tuner/servicefilter.h"
#include "display/displayservice.h"
#include <canvas/window.h>
#include <canvas/system.h>
#include <mpegparser/service/extension/time/timeextension.h>
#include <mpegparser/provider/provider.h>
#include <util/cfg/cfg.h>
#include <util/cfg/configregistrator.h>
#include <util/tool.h>
#include <util/fs.h>
#include <util/mcr.h>
#include <util/log.h>
#include <util/string.h>
#include <boost/filesystem.hpp>
#include <string>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <fstream>

namespace zapper {

namespace fs = boost::filesystem;

namespace impl {

static std::string getPlatformVersion() {
	std::string version("default_00");
	std::ifstream file (PLATFORM_VERSION_FILE);
	if (file.is_open()){
		getline( file, version );
		file.close();
	}
	return version;
}

}

REGISTER_INIT_CONFIG( zapper ) {
	root().addNode( "zapper" )
		.addValue( "showOneSeg", "Show one seg", false )
		.addValue( "canShutdown", "Set if zapper can shutdown", ZAPPER_CAN_SHUTDOWN ? true : false )
		.addValue( "countryCode", "ISO country code", "ARG" )			
		.addValue( "imagesPath", "Images path", "" )	
		.addValue( "useEPG", "Enable EPG", true )
		.addValue( "useOTA", "Enable Over The Air updates", true )
		.addValue( "otaMakerId", "Zapper OTA Maker ID", ZAPPER_OTA_MAKER_ID )
		.addValue( "otaModelId", "Zapper OTA Model ID", ZAPPER_OTA_MODEL_ID )
		.addValue( "otaVersion", "Zapper OTA Version", ZAPPER_OTA_VERSION )
		.addValue( "platformVersion", "Zapper platform version string", impl::getPlatformVersion() );
}

Zapper::Zapper( const std::string &name )
{
	_exitCode = 0;
	_sys = NULL;
	_mgr = NULL;
	_input = NULL;
	_initialized = false;
	_shutdown = util::cfg::getValue<bool>("zapper.canShutdown");
	_name = name;
}

Zapper::~Zapper()
{
	assert(!_initialized);
	delete _mgr;
}

//	Initialize
bool Zapper::initialize() {
	if (!_initialized) {
		LDEBUG( "Zapper", "Initialize" );

		//	Setup zapper title
		util::cfg::setValue( "gui.window.title", _name );

		//	Setup databse/RAM disk
		std::string db;
		if (!setupPaths( db )) {
			return false;
		}

		//	Initialize hardware
		if (!initHardware()) {
			return false;
		}

		//	Create and setup system canvas
		if (!initCanvas()) {
			return false;
		}

		//	Create plugin manager
		_mgr = new plugin::PluginManager( _sys->dispatcher() );

		//	Initialize plugins
		registerPlugins();

		//	Initialize manager
		mgr()->initialize( db );

		//	Create and register system services
		registerServices( mgr() );

		_initialized = true;
	}
	return _initialized;
}

//	Finalize
void Zapper::finalize() {
	LDEBUG( "Zapper", "Finalize: initialized=%d", _initialized );
	if (_initialized) {
		_initialized = false;

		//	Finalize plugins
		mgr()->finalize();

		//	Finalize system canvas
		_sys->finalize();
		DEL(_sys);

		//	Finalize hardware
		finHardware();
	}
}

std::string Zapper::name() const {
	return _name;
}

//	Database/RAM disk
bool Zapper::setupPaths( std::string &dbFile ) {
	fs::path userDir( getenv("HOME") );
	userDir /= "." + util::Tool::getName();
	
	//	Create if not exist
	if (!fs::exists( userDir )) {
		fs::create_directories( userDir );
	}
	else if (!fs::is_directory( userDir )) {
		LERROR( "Zapper", "User directory path is not a directory" );
		return false;
	}

	{	//	Storage
		fs::path tmp( userDir );
		tmp /= util::Tool::getName();
		dbFile = tmp.string();
	}

	{	//	Setup RAM disk
		fs::path tmp(userDir);
		tmp /= "ramdisk";
		_ramDisk = tmp.string();
	}

	LINFO( "Zapper", "Using %s for storage: dbFile=%s, ramDisk=%s",
		userDir.string().c_str(), dbFile.c_str(), _ramDisk.c_str() );

	if (!util::safeCleanDirectory( ramDisk() )) {
		LERROR( "Zapper", "Cannot clean RAM disk directory" );
		return false;
	}

	return true;
}

//	Restore to default configuration
void Zapper::resetConfig() {
	LDEBUG( "Zapper", "Reset stored configuration" );
	mgr()->resetConfig();
}

void Zapper::start() {
	LDEBUG( "Zapper", "Start" );

	//	Start plugin manager
	assert(_initialized);
	mgr()->start();

	//	Get input manager
	_input = mgr()->addRef<input::InputService>( "Input" );
	assert(_input);

	//	Start custom
	onStart();
}

void Zapper::stop() {
	LDEBUG( "Zapper", "Stop" );
	assert(_initialized);

	//	Stop custom
	onStop();

	//	Stop manager
	RELEASE_SERVICE(_input);
	mgr()->stop();
}

void Zapper::onStart() {
}

void Zapper::onStop() {
}

//	Loop methods
void Zapper::run() {
	//	Run canvas loop
	_sys->run();
}

void Zapper::exit( int ret/*=0*/ ) {
	if (_shutdown || ret > 0) {
		_exitCode = ret;
		_sys->exit();
	}
}

int Zapper::exitCode() {
	return _exitCode;
}

bool Zapper::initCanvas() {
	//	Create system canvas
	_sys = createCanvasSystem();
	if (!_sys) {
		LWARN( "Zapper", "Cannot create system canvas" );
		return false;
	}

	//	Setup system
	_sys->setNotifyCallback( boost::bind(&Zapper::onSystemEvent,this,_1) );

	//	Initialize
	if (!_sys->initialize()) {
		return false;
	}

	return true;
}

canvas::System *Zapper::createCanvasSystem() const {
	return canvas::System::create();	
}

//	Services registration
void Zapper::registerPlugins() {
	{	//	Add OTA plugin, if necesary
		bool useOTA = util::cfg::getValue<bool>("zapper.useOTA");
		LINFO( "Zapper", "Use OTA=%d", useOTA );
		if (useOTA) {
			addPlugin( new update::UpdatePlugin( ramDisk() ) );
		}
	}
}

void Zapper::registerServices( plugin::PluginManager *mgr ) {
	{	//	Register input service
		input::InputService *input = new input::InputService();
		if (supportStandby()) {
			zapper::input::Keys keys;
			keys.push_back( util::key::power );
			keys.push_back( util::key::f12 );			
			input->reserveKeys( "zapper", boost::bind(&Zapper::onPowerKey,this,_1,_2), keys, MAX_KEY_PRIORITY );
		}
		mgr->registerService( input );
	}
	
	mgr->registerService( new control::Service() );
	mgr->registerService( createDisplay() );

	{	//	Register mount service
		MountService *srv = createMount();
		const std::string &path = util::cfg::getValue<std::string>("zapper.imagesPath");
		if (!path.empty()) {
			srv->addPath( path, true );
		}
		mgr->registerService( srv );
	}

	//	Create audio mixer
	if (haveMixer()) {
		mgr->registerService( new audio::Service( createMixer() ) );
	}
	else {
		mgr->registerService( new audio::Service() );		
	}
	
	{	//	Create master tuner
		tuner::Provider *prov = createProvider();
		prov->setDispatcher( _sys->dispatcher() );
		mgr->registerService( new Tuner( 0, prov, createServiceFilter(), ramDisk() ) );
	}

	//	Create player service
	mgr->registerService( new player::PlayerService( "Tuner0", createPlayer() ) );

	{	//	Create channel manager
		bool useEPG = util::cfg::getValue<bool>("zapper.useEPG");
		LINFO( "Zapper", "Use EPG=%d", useEPG );		
		mgr->registerService( new channel::ChannelManager( useEPG ) );
	}

	//	Create channel player
	mgr->registerService( new channel::ChannelService() );

	{	//	Create Time service
		const std::string &country = util::cfg::getValue<std::string>("zapper.countryCode");		
		mgr->registerService(
			new plugin::GenericService(
				"Time",
				new tuner::TimeExtension(country)
			)
		);
	}
}

//	Aux hardare initialization
bool Zapper::initHardware() {
	return true;
}

void Zapper::finHardware() {
}

//	Stand by
bool Zapper::supportStandby() const {
	return false;
}

void Zapper::onPowerKey( util::key::type /*key*/, bool isUp ) {
	if (isUp) {
		if (!mgr()->inStandby()) {
			standby();
		}
		else {
			wakeup();
		}
	}
}

void Zapper::standby() {
	mgr()->standby();
	doStandby();
}

void Zapper::wakeup() {
	doWakeup();
	mgr()->wakeup();
}

void Zapper::doStandby() {
	LINFO( "Zapper", "Standby!!!" );
}

void Zapper::doWakeup() {
	LINFO( "Zapper", "Wakeup!!!" );
}

//	Instance creation
tuner::Provider *Zapper::createProvider() const {
	tuner::Provider *provider=tuner::Provider::create();
	if (!provider) {
		throw std::runtime_error("Couldn't create provider");
	}
	return provider;
}

bool Zapper::haveMixer() const {
	return false;
}

audio::Mixer *Zapper::createMixer() const {
	return new audio::Mixer();
}

display::DisplayService *Zapper::createDisplay() const {
	return new display::DisplayService(_sys);
}

player::Player *Zapper::createPlayer() const {
	return new player::impl::Player( sys() );
}

ServiceFilter *Zapper::createServiceFilter() const {
	return new BasicServiceFilter();
}

MountService *Zapper::createMount() const {
	return new MountService();
}

//	Plugins
void Zapper::addPlugin( plugin::Plugin *plugin ) {
	LDEBUG( "Zapper", "Add plugin: name=%s", plugin->name().c_str() );
	mgr()->add( plugin );
}

plugin::PluginManager *Zapper::mgr() const {
	return _mgr;
}

void Zapper::onSystemEvent( canvas::event::Notify *notify ) {
	switch (notify->type) {
		case canvas::event::type::key: {
			if (notify->keyEvt.key != util::key::null) {
				_input->dispatchKey( notify->keyEvt.key, notify->keyEvt.isUp );
			}
			break;
		}
		case canvas::event::type::button: {
			_input->dispatchButton( &notify->buttonEvt );
			break;
		}
		case canvas::event::type::shutdown: {
			this->exit();
			break;
		}			
		default: {
			LWARN( "Zapper", "Canvas notification ignored: type=%d", notify->type );
		}
	}
}

//	Canvas
canvas::System *Zapper::sys() const {
	return _sys;
}

const std::string &Zapper::ramDisk() const {
	assert(!_ramDisk.empty());
	return _ramDisk;
}

int Zapper::getMinVersion() const {
	return ZAPPER_OTA_VERSION;
}

std::string Zapper::getVersion() const {
	char version [15];
	snprintf( version, sizeof(version), "%03d.%03d.%03d",
		util::cfg::getValue<int>("zapper.otaMakerId"),
		util::cfg::getValue<int>("zapper.otaModelId"),
		util::cfg::getValue<int>("zapper.otaVersion") );
	return version;
}

std::string Zapper::getPlatformVersion() const {
	return util::cfg::getValue<std::string>("zapper.platformVersion");
}

}

